/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AdcConfigWidget.h"
#include "ModularDeviceStatus.h"
#include "ui_AdcConfigWidget.h"

#include <QDialog>
#include <QDialogButtonBox>
#include <optional>

#include "base/gui_utils.h"
#include "base/Globals.h"
#include "AdcSystem/DominoSettings.h"
#include "WaveRec/AdcBuildParams.h"

AdcConfigWidget::AdcConfigWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::AdcConfigWidget)
{
    ui->setupUi(this);
    if(!Globals::getDebugMode()) {
        ui->spinBoxPattern->hide();
        ui->checkBoxPattern->hide();
    }
    ui->groupBoxRunLogic->hide();
    ui->horizontalLayout->setAlignment(Qt::AlignLeft);
}

AdcConfigWidget::~AdcConfigWidget()
{
    delete ui;
}

void AdcConfigWidget::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void AdcConfigWidget::setConfig(const ConfigKey &, const BaseDeviceAppConfig &_config)
{
    // no need to check for modifications here, see ConfigDispatcher
    if(config == _config)
        return;
    auto devSet = config.getDeviceIndexEnabledMap().getEnabled();
    bool devSetChanged = devSet != _config.getDeviceIndexEnabledMap().getEnabled();
    config = _config;

    if(config.knownConfigs.isEmpty()){
        restoreFormValues();
    }
    if(devSetChanged) {
        for(auto i : deviceStatus.keys()) {
            if(!devSet.contains(i))
                deviceStatus.remove(i);
        }
        updateFirVisiblity();
    }

    updateLatMwSpinBox();
    updateRunLogicVisiblity();
    setupChanged();
}

static bool needReset(const QMap<DeviceIndex, ModularDeviceStatus> &stMap, const ModularDeviceStatus &st)
{
    const auto &index = st.getDeviceIndex();
    if (!stMap.contains(index))
        return true;
    return stMap[index].sdb != st.sdb;
}

void AdcConfigWidget::deviceStatusUpdated(const ModularDeviceStatus &status)
{
    const auto &index = status.getDeviceIndex();
    bool performReset = needReset(deviceStatus, status);

    deviceStatus[index] = status;
    if(performReset) {
        updateFirVisiblity();
    }
}

void AdcConfigWidget::setShowRunLogic(bool show)
{
    showRunLogic = show;
    updateRunLogicVisiblity();
}

void AdcConfigWidget::setupChanged()
{
    config.unify_config();
    emit configChanged(config);
}

bool AdcConfigWidget::updateDeviceBuildParams(DeviceIndex index, const AdcBuildParams &bp)
{
    bool ret = !adcBuildParams.contains(index);
    if (!ret) {
        ret |= (adcBuildParams[index] != bp);
    }
    if (ret) {
        adcBuildParams[index] = bp;
    }
    return ret;
}

void AdcConfigWidget::adcBuildParamsUpdated(DeviceIndex index, const AdcBuildParams &bp)
{
    bool deviceMwStatusUpdated = updateDeviceBuildParams(index, bp);
    if (deviceMwStatusUpdated) {
        updateLatMwSpinBox();
    }
}

void AdcConfigWidget::restoreFormValues()
{
    RecursiveWidgetSignalBlocker blocker(this);

    updateLatMwSpinBox();
}


TrigCtrlStatWidget *AdcConfigWidget::getTrigCtrlStatWidget() const
{
    return ui->trigCtrlStatWidget;
}

WaveBlcWidget *AdcConfigWidget::getWaveBlcWidget() const
{
    return ui->waveBlcWidget;
}

FirFilterWidget *AdcConfigWidget::getFirFilterWidget() const
{
    return ui->firFilterWidget;
}

RunLogicWidget *AdcConfigWidget::getRunLogicWidget() const
{
    return ui->runLogicWidget;
}

//--readout window
void AdcConfigWidget::on_spinBoxSize_valueChanged(int newval)
{
    config.defaultSetup.waveRec.readCellNumber = newval;
    setupChanged();
    emit updateSparseMarkers();
}

void AdcConfigWidget::on_spinBoxLatency_valueChanged(int newval)
{
    config.defaultSetup.tqdcCore.latency = newval;
    setupChanged();
}

//--trigger
void AdcConfigWidget::setTriggerSourceEnabled(TrigCtrlStatConfig::TrigSource trigSource, bool state)
{
    // if(!config.knownConfigs.contains(selectedDevice)) return;
    //auto &settings = config.knownConfigs[selectedDevice];
    //settings.trigCtrlStat.setEnabled(TrigCtrlStatConfig::TrigSourcePeriodic, checked);
    config.defaultSetup.trigCtrlStat.setEnabled(trigSource, state);
    setupChanged();
}

void AdcConfigWidget::updateLatMwSpinBox() const
{
    RecursiveWidgetSignalBlocker blocker(ui->groupBox_ReadoutWindow);

    std::optional<AdcBuildParams> minBp;
    for(auto i : config.getDeviceIndexEnabledMap().getEnabled()) {
        auto it = adcBuildParams.constFind(i);
        if(it==adcBuildParams.cend())
            continue;
        if(minBp) {
            minBp->adcMaxLat = std::min(minBp->adcMaxLat, it->adcMaxLat);
            minBp->adcMaxMatchWin = std::min(minBp->adcMaxMatchWin, it->adcMaxMatchWin);
        } else
            minBp = *it;
    }
    ui->spinBoxLatency->setEnabled(minBp.has_value());
    ui->spinBoxSize->setEnabled(minBp.has_value());

    const auto &defCfg = config.defaultSetup;
    if(!minBp)
        minBp = {.adcMaxLat=defCfg.tqdcCore.latency,
                 .adcMaxMatchWin=defCfg.waveRec.readCellNumber};

    if(minBp) {
        ui->spinBoxLatency->setMinimum(0);
        ui->spinBoxLatency->setMinimum(-511);
        ui->spinBoxSize->setMinimum(4);

        ui->spinBoxLatency->setMaximum(minBp->adcMaxLat);
        ui->spinBoxSize->setMaximum(minBp->adcMaxMatchWin);
    }
    ui->spinBoxSize->setValue(defCfg.waveRec.readCellNumber);
    ui->spinBoxLatency->setValue(defCfg.tqdcCore.latency);
}

void AdcConfigWidget::updateFirVisiblity() const
{
    bool hasFir = false;
    auto devSet = config.getDeviceIndexEnabledMap().getEnabled();
    for(auto i : devSet) {
        auto it=deviceStatus.constFind(i);
        if(it != deviceStatus.cend()) {
            if(it->firFilterStatus) {
                hasFir = true;
            }
        }
    }
    ui->groupBoxFirFilter->setVisible(hasFir);
}

void AdcConfigWidget::updateRunLogicVisiblity() const
{
    ui->groupBoxRunLogic->setVisible(showRunLogic ||
                                     config.defaultSetup.runLogic.isAnyLimit());
}

void AdcConfigWidget::on_spinBoxPattern_valueChanged(int arg1)
{
    config.defaultSetup.adcSystemCore.pattern = arg1;
    setupChanged();
}

void AdcConfigWidget::on_checkBoxPattern_toggled(bool checked)
{
    config.defaultSetup.adcSystemCore.usePattern = checked;
    ui->spinBoxPattern->setEnabled(checked);
    setupChanged();
}
