#include "HrbCore.h"

#include <QElapsedTimer>
#include <QThread>
#include <QTimer>

#include "HrbMainWindow.h"
#include "base/app/AppVersion.h"
#include "base/DeviceId.h"
#include "fsm/MStreamDeviceMgrFsm.h"
#include "HrbSignalRouter.h"

#include "config/HrbConfigDb.h"
#include "hardware/HrbDeviceGlobal.h"
#include "hardware/TtbDeviceGlobal.h"

#include "base/ConfigIndexName.h"
#include "base/DeviceIndexEnabled.h"
#include "base/threadQuit.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "remote-control-server/RemoteControlServer.h"
#include "remote-control-server/EvNumChecker.h"

namespace {
const int METRICS_SEND_PERIOD_MS = 1000;
}

HrbCore::HrbCore(QObject *parent)
    : QObject(parent)
    , fsm(new MStreamDeviceMgrFsm(this))
    , remoteControlServer(new RemoteControlServer(this, 0))
    , configDb(new HrbConfigDb(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , evNumChecker(new EvNumChecker(this))
{
    qRegisterMetaType<ModularDeviceConfig>();
    qRegisterMetaType<HrbAppConfig>();
    qRegisterMetaType<HrbTriggerSetup>();

    qRegisterMetaType<HrbStatus>();
    qRegisterMetaType<TtbStatus>();

    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<LldpInfoCache>();

//    qRegisterMetaType<DevFwRevMap>();

    auto signalRouter = &HrbSignalRouter::instance();
    signalRouter->setDeviceDispatcher(deviceDispatcher);
    signalRouter->setRemoteControlServer(remoteControlServer);
    signalRouter->setEvNumChecker(evNumChecker);
    signalRouter->setConfigDb(configDb);
    signalRouter->setCore(this);
    signalRouter->setFsm(fsm);
    signalRouter->connectAll();

    initProgDescr();
    updateMetricNamespace();

    QTimer * sendMetricsTimer = new QTimer(this);
    connect(sendMetricsTimer, &QTimer::timeout, this, &HrbCore::sendMetrics);
    sendMetricsTimer->start(METRICS_SEND_PERIOD_MS);
}

HrbCore::~HrbCore()
{
    hrbDevices.clear();
    ttbDevices.clear();
}

void HrbCore::setConfig(const ConfigKey &config_key, const HrbAppConfig &newConfig)
{
    Q_UNUSED(config_key);
    if(config == newConfig)
        return;
    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;

    DeviceIndexEnabledMap diem;
    const auto toDel = oldSet-newSet;

    for (const auto &index : toDel) {
        if (index.getDevId()==DEVICE_ID_HRB6ASD) {
            hrbDevices.remove(index);
        } else {
            ttbDevices.remove(index);
        }
    }


    const DeviceIndexSet toAdd = newSet-oldSet;
    for (const auto &index : toAdd)
        addDevice(index);

    auto enabledHrbDevSet = newConfig.getDeviceIndexEnabledMap(DeviceFilter::HRB).getEnabled();
    evNumChecker->set_dev_serials(enabledHrbDevSet);
    fsm->setMStreamDevSet(enabledHrbDevSet);
    updateProgDescrOptions();
    emit allowRemoteControl(config.allowRemoteControl);
}

void HrbCore::configSelectorChanged(const ConfigIndexName &conf)
{
    if(progDescr.getProgramIndex() != conf.getIndex()) {
        emit closeProgram(progDescr);
        progDescr.uuid = QUuid::createUuid();
        progDescr.setProgramIndex(conf.getIndex());
        emit updateProgramDescription(progDescr);
    }

    updateMetricNamespace(conf);
}

void HrbCore::performLocalStart()
{
    if(fsm->getState() != FsmState::Ready) {
        qWarning() << __PRETTY_FUNCTION__ << "wrong state:" << fsm->getState();
        return;
    }
    fsm->setRequestState(FsmState::Configure);
}

void HrbCore::performLocalStop()
{
    if(fsm->getState() != FsmState::Run) {
        qWarning() << __PRETTY_FUNCTION__ << "wrong state:" << fsm->getState();
        return;
    }
    fsm->setRequestState(FsmState::Stopping);
}

void HrbCore::remoteStartRequest(const RcRunId &runId_, const QVariant &startEvNum)
{
    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
        return;
    }

    runId = runId_;
    emit preparingRun(runId);
    fsm->setRequestState(FsmState::Configure);
}

void HrbCore::remoteStopRequest(const RcRunId &runId_, const QVariant &stopEvNum_)
{
    bool success=true;
    if(runId.runIndex != runId_.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); "
                             "remote control runIndex (%2)")
                     .arg(runId.runIndex)
                     .arg(runId_.runIndex);
        success = false;
    }

    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
    }

    //todo add PrepareStopping state
    stopEvNum = stopEvNum_;
    fsm->setRequestState(fsm->getState() == FsmState::Run ? FsmState::Stopping
                                                          : FsmState::Error);
}

void HrbCore::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = progDescr.updateProgramInterface(
                ProgramInterfaceRemoteControl, 0,
                status.tcpPort, status.listening, !status.portInUse,
                status.peerHost, status.peerPort);
    if(changed)
        emit updateProgramDescription(progDescr);

    if(status.portInUse) {
        QString ready4run_str = "Ready for run";
        for(auto &globDev : hrbDevices) {
            if(!globDev->getDevStatus().connected){
                ready4run_str = "Not ready for run";
            }
        }
        for(auto &globDev : ttbDevices) {
            if(!globDev->getDevStatus().connected){
                ready4run_str = "Not ready for run";
            }
        }
//        emit sendRemoteState(ready4run_str);
    } else if(lastRcStatus.portInUse) {
        qWarning() << "Remote control disconnected.";

        if(fsm->getState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            fsm->setRequestState(fsm->getState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }
    lastRcStatus = status;
}

void HrbCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    switch (state) {
    case FsmState::Ready :
    {
        if(oldState == FsmState::Stopping)
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
        break;
    }
    case FsmState::Configure :
    {
        if(++mstreamSeqNum < 100) mstreamSeqNum = 100;
        fsm->setMStreamSeqNum(mstreamSeqNum);
        emit mstream_reset_all(mstreamSeqNum);
        fsm->setRequestState(FsmState::Starting);
        break;
    }
    case FsmState::Starting :
    {
        config.setRunState(true);
        emit configChanged(config);
        fsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Run :
    {
        qInfo() << QString("Start completed successfully");
        break;
    }
    case FsmState::Stopping :
    {
        config.setRunState(false);
        emit configChanged(config);
        fsm->setRequestState(FsmState::Ready);
        break;
    }
    case FsmState::Error :
    {
        if(config.isRun()) {
            config.setRunState(false);
            emit configChanged(config);
        }
        break;
    }
    default:;
    }
}

void HrbCore::addDevice(const DeviceIndex &index)
{
    try {
        if(index.getDevId() == DEVICE_ID_HRB6ASD) {
            auto *devGlob = new HrbDeviceGlobal(index, this);
            hrbDevices.insert(index, QSharedPointer<HrbDeviceGlobal>(devGlob));
            HrbSignalRouter::instance().connect_DeviceGlobal(devGlob);
        } else {
            auto *devGlob = new TtbDeviceGlobal(index, this);
            ttbDevices.insert(index, QSharedPointer<TtbDeviceGlobal>(devGlob));
            HrbSignalRouter::instance().connect_DeviceGlobal(devGlob);
        }
    } catch (std::exception &e){
        qCritical() << QString("[%1] Error on connect to device: %2")
                       .arg(index.getDeviceIdName()).arg(e.what());
    }
}

void HrbCore::deviceDisconnected(const DeviceIndex &index)
{
    if(config.isRun()) {
        const auto &set = config.getDeviceIndexEnabledMap().getEnabled();
        if(set.contains(index))
            sendRunError(runId, "Device disconnected");
    }
}

void HrbCore::devConnectedStateUpdated(const DeviceIndex &index, bool connected)
{
    if(!config.getDeviceIndexEnabledMap().getEnabled().contains(index))
        return;
    const QString &msg = QString("[info] [%1] %2")
            .arg(index.getIdent())
            .arg(connected ? "connected" : "disconnected");
    (connected ? qInfo() : qWarning()) << msg;

    if(config.isRun() && connected) {
        sendRunError(runId, msg);
    }
}

void HrbCore::onRunWarning(const QString &msg)
{
    if(fsm->getState() == FsmState::Configure
            || fsm->getState() == FsmState::Starting
            || fsm->getState() == FsmState::Run
            || fsm->getState() == FsmState::Stopping) {
        emit sendRunWarning(runId, msg);
    }
}

void HrbCore::sendMetrics() const
{
    for(auto dev : hrbDevices) {
        auto status = dev->getDevStatus();
        const auto &metricStatus = dev->getLastMetricDevStatus();
        if(!metricStatus.getDeviceIndex().isNull())
            status -= metricStatus;
        const QMap<QString, QString> tags({{"serial", status.getSerialStr()}});
        const auto temps = status.getTemps();
        if (!temps.empty())
            MetricSender::send(QString("hrb.temp"), temps.first(), tags);
        MetricSender::send(QString("hrb.count.events"), status.globEvNum, tags);
        MetricSender::send(QString("hrb.time.idle"), status.statistic.idleTimer*HRB_CLK_SEC, tags);
        MetricSender::send(QString("hrb.time.busy"),
                           status.globEvNum ? status.statistic.busyTimer*HRB_CLK_SEC/status.globEvNum : 0, tags);
        MetricSender::send(QString("hrb.time.busyMStream"),
                           status.globEvNum ? status.statistic.busyMStreamTimer*HRB_CLK_SEC/status.globEvNum : 0, tags);
        dev->updateLastMetricDevStatus();
    }
    for(auto dev : ttbDevices) {
        auto status = dev->getDevStatus();
        const auto &metricStatus = dev->getLastMetricDevStatus();
        if(!metricStatus.getDeviceIndex().isNull())
            status -= metricStatus;
        const QMap<QString, QString> tags({{"serial", status.getSerialStr()}});
        const auto &temps = status.getTemps();
        if (!temps.empty())
            MetricSender::send(QString("ttb.temp"), temps.first(), tags);
        MetricSender::send(QString("ttb.count.events"), status.globEvNum, tags);
        MetricSender::send(QString("ttb.time.spill"), status.statistic.spillCnt*TTB_CLK_SEC, tags);
        MetricSender::send(QString("ttb.time.busy"),
                           status.globEvNum ? status.statistic.ttbBusyCnt*TTB_CLK_SEC/status.globEvNum : 0, tags);
        for(int ch=0; ch<TTB_CH_COUNT; ++ch) {
            if(!(status.linkUp & (1<<ch)))
                continue;
            QMap<QString, QString> tags2(
                {{"serial", status.getSerialStr()}, {"channel", QString::number(ch)}});
            MetricSender::send(QString("ttb.time.busyCh"),
                               status.globEvNum ? status.statistic.chBusyCnt[ch]*TTB_CLK_SEC/status.globEvNum : 0, tags2);
        }
        dev->updateLastMetricDevStatus();
    }
}

void HrbCore::initProgDescr()
{
    progDescr.setClientIndex(configDb->getClientIndex());

    progDescr.ver_date = getAppDateStr();
    progDescr.ver_hash = getAppVersion();

    emit updateProgramDescription(progDescr);
}

void HrbCore::updateProgDescrOptions()
{
    QStringList devList;
    for(const auto &dg : hrbDevices)
        devList << dg->getDeviceIndex().toString();
    for(const auto &dg : ttbDevices)
        devList << dg->getDeviceIndex().toString();
    progDescr.options["DevList"] = devList.join(";");
    progDescr.options["run"] = QVariant(config.isRun()).toString();

    emit updateProgramDescription(progDescr);
}

void HrbCore::updateMetricNamespace(const ConfigIndexName conf) const
{
    MetricSender::setBaseTag("index", conf.getIndex().toString());
}
