/*
**    Copyright 2013-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "UtcTimestampModule.h"
#include <cmath>
#include <QDebug>

#include "TTBRegs.h"

QList<AFI_SDB_ID> UtcTimestampModule::getSupportedSdbIds()
{
    return {};
}

void UtcTimestampModule::post_init()
{
    if (!isAvailable()) {
        // no-SDB compatibility
        bus.setAddrRange({0x0300, 0x000F});
    }
}

double UtcTimestampModule::timestampFrequencyHz() const
{
    return 125e6;
}

void UtcTimestampModule::setClock()
{
    const int TS_COUNT = 4;
    struct timeval tv0, tv;
    gettimeofday(&tv0, NULL);
    for(int i=0; i<TS_COUNT; i++) {
        gettimeofday(&tv, NULL);
        qint64 ts = (qint64)timestampFrequencyHz() * tv.tv_sec + (qint64)timestampFrequencyHz()/1e6 * tv.tv_usec;
        writeClock(ts);
    }
    double dt = (tv.tv_sec-tv0.tv_sec) + 1e-6*(tv.tv_usec-tv0.tv_usec);
    double ts_roundtrip = dt/TS_COUNT;
    qDebug() << QString("Timestamp write delay %1 s").arg(ts_roundtrip);
}

void UtcTimestampModule::writeClock(quint64 value)
{
    regWrite64(REG64_TTB_UTC_SET, value);
    ts_write_ts = value;
}

quint64 UtcTimestampModule::readClock()
{
    const int read_count = 4;
    std::vector<qint64> host_ts_v, hardware_ts_v;
    for(int i=0; i<read_count; i++) {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        qint64 t = (qint64)timestampFrequencyHz() * tv.tv_sec + (qint64)timestampFrequencyHz()/1e6 * tv.tv_usec;
        host_ts_v.push_back(t);
        quint64 ts;
        regRead64(REG64_TTB_UTC_TS_LOCAL, &ts);
//        qDebug() << "regRead64:" << hex << REG_TTB_UTC_LOCAL_TS << ts;
        hardware_ts_v.push_back(ts);
    }
    qint64 min_delta = hardware_ts_v[0] - host_ts_v[0];
    int min_index = 0;
    for(int i=0; i<read_count; i++) {
        qint64 delta = hardware_ts_v[i] - host_ts_v[i];
        if (delta < min_delta) {
            min_delta = delta;
            min_index = i;
        }
    }
    qint64 host_ts = host_ts_v[min_index];
    qint64 hardware_ts  = hardware_ts_v[min_index];
    if (hardware_ts == 0 || hardware_ts == -1) {
        qWarning() << QString("Timestamp not available (%1)").arg(hardware_ts);
        return 0;
    }
    double ts_step = 1.0/timestampFrequencyHz();
//    logger.info(QString("Timestamp period %1, frequency %2").arg(ts_step).arg(TIMESTAMP_FREQUENCY()));
//    logger.info(QString("Local Timestamp: %1 s").arg(ts_step*(host_ts), 0, 'f', 6));
//    logger.info(QString("ADC   Timestamp: %1 s").arg(ts_step*(hardware_ts), 0, 'f', 6));
    double delta_ts = ts_step * (hardware_ts - host_ts);
    double ts_runtime = ts_step*(host_ts - ts_write_ts);
//    double delta_freq = delta_ts / ts_runtime;
    if (ts_runtime > 1.0) {
        if (std::fabs(delta_ts) > 10) {
//            qWarning() << QString("Timestamp delta %1 s, period %2 s, delta_freq %3").arg(delta_ts).arg(ts_runtime).arg(delta_freq);
        } else {
//            qDebug() << QString("Timestamp delta %1 s, period %2 s, delta_freq %3").arg(delta_ts).arg(ts_runtime).arg(delta_freq);
        }
    }
    return hardware_ts;
}
