#include "Viewer.h"
#include "MongoViewer.h"
#include "JsonViewer.h"
#include "mongo/RootConfigUtil.h"
#include "base/ConfigDocument.h"

Viewer::Viewer() :
    mongo(new MongoViewer()),
    json(new JsonViewer())
{}

Viewer::~Viewer()
{
    delete mongo;
    delete json;
}

void Viewer::update_hosts()
{
    clear_all();
    hosts = mongo->get_all_hosts(config);
}

QList<ConfigDocument> Viewer::aggregate()
{
    if(config.location == MongoViewerConfig::LOCAL)
        return json->aggregate(config);
    else
        return mongo->aggregate(config);
}

void Viewer::setConfig(const MongoViewerConfig &newConfig)
{
    config = newConfig;
}

AbstractDbViewer *Viewer::getViewer()
{
    switch (config.location) {
    case MongoViewerConfig::MONGO:
        return mongo;
    case MongoViewerConfig::LOCAL:
        return json;
    }
    Q_UNREACHABLE();
}

AbstractDB *Viewer::getDb()
{
    switch (config.location) {
    case MongoViewerConfig::MONGO:
        return mongo;
    case MongoViewerConfig::LOCAL:
        return json;
    }
    Q_UNREACHABLE();
}


void Viewer::update_objects()
{
    objects = getViewer()->agregate_json(config);
    if(config.type == MongoViewerConfig::SAVED){
        cs_objects = getViewer()->agregate_cs_json(config);
    }
    else
        ri_objects = getViewer()->agregate_ri_json(config);
}

void Viewer::update_filters()
{
    all_program_types.clear();
    all_program_indexes.clear();
    all_run_indexes.clear();
    all_run_numbers.clear();

    auto viewer = getViewer();
//    all_program_types = viewer->get_program_types(config);
//    all_program_indexes = viewer->get_program_indexes(config);

    if(config.type == MongoViewerConfig::SAVED){
        QList<ConfigSelector> list = viewer->get_config_selectors(config);

        for(auto i=list.cbegin(); i!=list.cend(); ++i){
            const auto &cs = QVariant::fromValue(*i).value<ConfigSelector>();
            all_program_types.append(cs.getType().toString());
            all_program_indexes.append(cs.getIndex().toString());
        }
    }
    if(config.type == MongoViewerConfig::HISTORY){
         QList<RcRunId> list_run_id = viewer->get_rc_run_ids(config);
         for(auto i=list_run_id.cbegin(); i!=list_run_id.cend(); ++i){
             const auto &runId = QVariant::fromValue(*i).value<RcRunId>();
            all_run_indexes.append(runId.runIndex);
            all_run_numbers.append(runId.runNumber);
         }
         all_program_types = viewer->get_rc_program_type(config);
    }

    all_program_types = remove_duplicates<QString>(all_program_types);
    all_program_indexes = remove_duplicates<QString>(all_program_indexes);
    all_run_indexes = remove_duplicates<QString>(all_run_indexes);
    all_run_numbers = remove_duplicates<int>(all_run_numbers);
}

void Viewer::set_object(int i)
{
    selected_obj = objects[i];
    if(config.type == MongoViewerConfig::SAVED) {
        selected_obj = cs_objects.values().value(i);
    } else
        selected_obj = ri_objects.values().value(i);
}

void Viewer::delete_config()
{
//    ConfigSelector cs = RootConfigUtil::getConfigSelector(selected_obj);
//    ConfigSelector cs = selected_obj;
//    int i = getDb()->delete_config(cs);
    //    qInfo() << QString("Removed %1 config: %2").arg(i).arg(cs.toString());
}

void Viewer::clear_all()
{
    hosts.clear();
    all_program_types.clear();
    all_program_indexes.clear();
    all_run_indexes.clear();
    objects.clear();
    cs_objects.clear();
    ri_objects.clear();
    selected_obj =  QJsonObject();
}

template<typename T>
QList<T> Viewer::remove_duplicates(QList<T> &list)
{
    QSet<T> set(list.begin(), list.end());
    QList<T> str_list(set.begin(), set.end());
    return str_list;
}
