#include "Tlu16SfpModel.h"

#include "Tlu16SfpColumns.h"

#include <QColor>
#include <QTimer>

#include "base/DeviceId.h"

namespace {
inline Qt::CheckState boolToCheck(bool v)
{ return v ? Qt::Checked : Qt::Unchecked; }

Tlu16SfpColumns getColumnByIndex(const QModelIndex &index)
{
    return static_cast<Tlu16SfpColumns>(index.column());
}

QVector<bool> Tlu16SfpTriggerConfig::* getConfigBoolVectorByCol(Tlu16SfpColumns col) {
    switch(col) {
    case Tlu16SfpColumns::COL_TOF_EN: return &Tlu16SfpTriggerConfig::tofChEn;
    case Tlu16SfpColumns::COL_LUM_EN: return &Tlu16SfpTriggerConfig::lumChEn;
    default: return nullptr;
    }
}

static QMap<DeviceIndex, int> buildNchMap(const BaseDeviceAppConfig &config)
{
    QMap<DeviceIndex, int> nch;
    for(auto it = config.knownConfigs.cbegin(); it!=config.knownConfigs.cend(); ++it) {
        if(it.key().getDevId() != DEVICE_ID_TLU16SFP)
            continue;
        if(!it->enabled)
            continue;
        auto v = std::max(it->tlu16SfpTrigger.tofChEn.size(), it->tlu16SfpTrigger.lumChEn.size());
        v = std::max(it->tluDecDes.chReset.size(), v);
        nch.insert(it.key(), v);
    }
    return nch;
}

static bool needReset(const BaseDeviceAppConfig &config, const BaseDeviceAppConfig &newConfig)
{
    const auto &oldNch = buildNchMap(config);
    const auto &newNch = buildNchMap(newConfig);
    if (oldNch != newNch)
        return true;

    return false;
}

static bool hasChanges(const ModularDeviceConfig &c1, const ModularDeviceConfig &c2)
{
    if(c1.tlu16SfpTrigger.tofChEn != c1.tlu16SfpTrigger.tofChEn)
        return true;
    if(c1.tlu16SfpTrigger.lumChEn != c1.tlu16SfpTrigger.lumChEn)
        return true;
    return false;
}

static bool hasChanges(const BaseDeviceAppConfig &c1, const BaseDeviceAppConfig &c2)
{
    for (auto it1=c1.knownConfigs.cbegin(); it1!=c1.knownConfigs.cend(); ++it1) {
        auto it2 = c2.knownConfigs.constFind(it1.key());
        if(it2 == c2.knownConfigs.cend())
            return true;
        if (hasChanges(*it1, *it2))
            return true;
    }
    return false;
}
}

Tlu16SfpModel::Tlu16SfpModel(QObject *parent)
    : QAbstractItemModel(parent)
    , delayedSendConfigTimer(new QTimer(this))
{
    delayedSendConfigTimer->setInterval(100);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &Tlu16SfpModel::postCheckStateChange);
}

QVariant Tlu16SfpModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::ToolTipRole:
        return toolTipData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::ForegroundRole:
        return foregroundData(index);
    case Qt::UserRole:
        return limitsData(index);
    }
    return QVariant();
}

bool Tlu16SfpModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;

    const auto item = getItemByIndex(index);
    auto devCfgIt = config.knownConfigs.find(item->dev);
    if(devCfgIt == config.knownConfigs.end())
        return false;

    switch (role) {
    case Qt::CheckStateRole:
    {
        Q_ASSERT(value!=Qt::PartiallyChecked);
        if(item->ch == -1) {
            auto rows = rowCount(index);
            bool ok = true;
            for(int r=0; r<rows; ++r){
                ok &= setData(this->index(r,index.column(),index), value, role);
            }
            return ok;
        } else {
            Q_ASSERT(!item->dev.isNull());
            auto col = getColumnByIndex(index);
            if(col == Tlu16SfpColumns::COL_RESET) {
                auto &vec = devCfgIt->tluDecDes.chReset;
                if(vec.size() <= item->ch)
                    return false;
                vec[item->ch] = (value==Qt::Checked) ? true : false;
            } else if(auto ptr = getConfigBoolVectorByCol(col)) {
                auto &trigCfg = devCfgIt->tlu16SfpTrigger;
                auto &vec = (trigCfg.*ptr);
                if(vec.size() <= item->ch)
                    return false;
                vec[item->ch] = (value==Qt::Checked) ? true : false;
            } else
                return false;

            enqueueConfigSend();
        }
        return true;
    }
    case Qt::EditRole:
    {
        const auto item = getItemByIndex(index);
        auto col = getColumnByIndex(index);
        if(col != Tlu16SfpColumns::COL_TOF_SUM_THR || item->ch != -1)
            return false;

        devCfgIt->tlu16SfpTrigger.tofSumThr = value.toInt();
        enqueueConfigSend();
        emit dataChanged(index, index, {Qt::DisplayRole, Qt::EditRole});
        return true;
    }
    }

    return false;
}

Qt::ItemFlags Tlu16SfpModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = getColumnByIndex(index);
    switch (col) {
    case Tlu16SfpColumns::COL_TOF_SUM_THR:
        f |= Qt::ItemIsEditable | Qt::ItemIsEnabled;
    case Tlu16SfpColumns::COL_RESET:
    case Tlu16SfpColumns::COL_TOF_EN:
    case Tlu16SfpColumns::COL_LUM_EN:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    default:;
    }
    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant Tlu16SfpModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return {};
    if(role != Qt::DisplayRole)
        return {};
    switch (static_cast<Tlu16SfpColumns>(section)) {
    case Tlu16SfpColumns::COL_NAME: return "Name";
    case Tlu16SfpColumns::COL_RESET:   return "Reset";
    case Tlu16SfpColumns::COL_TOF_SUM_THR: return "TOF sum thr";
    case Tlu16SfpColumns::COL_TOF_EN: return "TOF en";
    case Tlu16SfpColumns::COL_LUM_EN: return "Lum en";

    case Tlu16SfpColumns::COL_STATUS: return "Status";
    case Tlu16SfpColumns::COL_BITSLIP: return "BitSlip";
    case Tlu16SfpColumns::COL_IDELAY: return "iDelay";
    case Tlu16SfpColumns::COL_RX_CODE_ERR_CNT: return "Rx code err cnt";
    default:;
    }
    return {};
}

QModelIndex Tlu16SfpModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const auto *parentItem = parent.isValid() ?
        static_cast<Tlu16SfpModelItem*>(parent.internalPointer()) :
        &rootItem;
    Tlu16SfpModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex Tlu16SfpModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const auto *childItem = static_cast<Tlu16SfpModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    Tlu16SfpModelItem *parentItem = childItem->parent;
    if(parentItem == &rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int Tlu16SfpModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<Tlu16SfpModelItem*>(parent.internalPointer()) :
        &rootItem;
    return parentItem->childCount();
}

int Tlu16SfpModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return static_cast<int>(Tlu16SfpColumns::COL_TOTAL);
}

void Tlu16SfpModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}

int Tlu16SfpModel::getDevRow(const DeviceIndex &di) const
{
    return -1;
}

void Tlu16SfpModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if (config == newConfig)
        return;

    bool performReset = needReset(config, newConfig);
    bool hasAnyChanges = performReset ? true : hasChanges(config, newConfig);

    config = newConfig;
    if (performReset) {
        doReset();
    } else {
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void Tlu16SfpModel::setTlu16SfpTrigBp(const DeviceIndex &index,
                                      const Tlu16SfpTriggerBuildParams &bp)
{
    if(index.getDevId() != DEVICE_ID_TLU16SFP)
        return;
    auto it = buildParamsTrig.find(index);
    if(it != buildParamsTrig.end() || *it == bp)
        return;
    buildParamsTrig[index] = bp;

    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;

    bool modified = false;
    auto &tofChEn = devIt->tlu16SfpTrigger.tofChEn;
    if(tofChEn.size() != bp.chNum) {
        while(tofChEn.size() < bp.chNum)
            tofChEn.push_back(Tlu16SfpTriggerConfig::DEF_CH_EN);
        tofChEn.resize(bp.chNum);
        modified = true;
    }
    auto &lumChEn = devIt->tlu16SfpTrigger.lumChEn;
    if(devIt->tlu16SfpTrigger.lumChEn.size() != bp.chNum) {
        while(lumChEn.size() < bp.chNum)
            lumChEn.push_back(Tlu16SfpTriggerConfig::DEF_CH_EN);
        devIt->tlu16SfpTrigger.lumChEn.resize(bp.chNum);
        modified = true;
    }

    if(modified) {
        doReset();
        postCheckStateChange();
    }
}

void Tlu16SfpModel::setTluDesDecBp(const DeviceIndex &index,
                                        const TluDecDesBuildParams &bp)
{
    if(index.getDevId() != DEVICE_ID_TLU16SFP)
        return;
    auto it = buildParamsDecDes.find(index);
    if(it != buildParamsDecDes.end() || *it == bp)
        return;
    buildParamsDecDes[index] = bp;

    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;
    auto curSize = devIt->tluDecDes.chReset.size();
    if(curSize == bp.chNum)
        return;

    devIt->tluDecDes.chReset.resize(bp.chNum);

    doReset();
    postCheckStateChange();
}

void Tlu16SfpModel::setTluDecDesStatus(const DeviceIndex &index, const TluDecDesStatus &status)
{
    if(index.getDevId() != DEVICE_ID_TLU16SFP)
        return;
    if(!config.knownConfigs.contains(index))
        return;
    devStatus[index] = status;

    sendDataChanged({Qt::DisplayRole});
}

void Tlu16SfpModel::resetCounters()
{
    devStatusOffset = devStatus;
    sendDataChanged();
}

void Tlu16SfpModel::onDeviceConnected(const DeviceIndex &devIndex)
{
    devStatusOffset.remove(devIndex);
    sendDataChanged();
}

void Tlu16SfpModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    emit configChanged(this->config);
}

void Tlu16SfpModel::enqueueConfigSend()
{
    delayedSendConfigTimer->start();
}

void Tlu16SfpModel::doReset()
{
    beginResetModel();
    rootItem.children.clear();
    const auto &nchMap = buildNchMap(config);

    if (!nchMap.isEmpty()) {
        auto allItem = rootItem.createChild();

        for (auto devIt = nchMap.cbegin(); devIt != nchMap.cend(); ++devIt) {
            auto devItem = allItem->createChild();
            devItem->dev = devIt.key();

            for (int ch = 0; ch<*devIt; ch++) {
                auto chItem = devItem->createChild();
                chItem->dev = devIt.key();
                chItem->ch = ch;
            }
        }
    }
    endResetModel();
}

void Tlu16SfpModel::sendDataChanged(QVector<int> roles, const QModelIndex parent)
{
    auto rows = rowCount(parent);

    if(!rows)
        return;

    auto top = index(0,0, parent);
    auto last = index(rows-1, columnCount(top)-1, parent);
    emit dataChanged(top, last, roles);

    for(auto r=0; r<rows; ++r) {
        auto chIndex = index(r, 0, parent);
        sendDataChanged(roles, chIndex);
    }
}

QVariant Tlu16SfpModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case Tlu16SfpColumns::COL_NAME:
        if(item->dev.isNull())
            return "All";
        if(item->ch == -1)
            return item->dev.getSerialStr();
        return QString::number(item->ch);
    case Tlu16SfpColumns::COL_STATUS:
        if(auto stPtr = getChStatusPtr(item)) {
            return stPtr->isStatusOk() ? "Ok" :
                QString("%1").arg(stPtr->status, 3, 2, QChar('0'));
        }
        break;
    case Tlu16SfpColumns::COL_BITSLIP:
        if(auto stPtr = getChStatusPtr(item))
            return stPtr->bitSlip;
        break;
    case Tlu16SfpColumns::COL_IDELAY:
        if(auto stPtr = getChStatusPtr(item))
            return stPtr->iDelay;
        break;
    case Tlu16SfpColumns::COL_RX_CODE_ERR_CNT:
    {
        auto p = getCounterPair(item);
        if(!p.first)
            return {};
        return QString("%L1").arg(*p.first - p.second.value_or(0));
    }
    default:;
    }

    return {};
}

QVariant Tlu16SfpModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto &col = getColumnByIndex(index);
    if(col != Tlu16SfpColumns::COL_TOF_SUM_THR)
        return {};

    const auto &deviceIndex = item->dev;
    if (deviceIndex.isNull())
        return {};

    const auto &devModCfg = config.knownConfigs.value(deviceIndex).tlu16SfpTrigger;
    if (item->ch != -1)
        return {};

    return devModCfg.tofSumThr;
}

QVariant Tlu16SfpModel::toolTipData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case Tlu16SfpColumns::COL_STATUS:
        if(auto stPtr = getChStatusPtr(item))
            return stPtr->getStatusStringList().join('\n');
        break;
    case Tlu16SfpColumns::COL_RX_CODE_ERR_CNT:
    {
        auto p = getCounterPair(item);
        if(!p.first)
            return {};
        auto ret = QString("%L1").arg(*p.first);
        if(p.second)
            ret += QString(" - %L2").arg(*p.second);
        return ret;
    }
    default:;
    }

    return {};
}

QVariant Tlu16SfpModel::checkStateData(const QModelIndex &index) const
{
    const auto &col = getColumnByIndex(index);
    auto ptr = getConfigBoolVectorByCol(col);
    if(!ptr && col != Tlu16SfpColumns::COL_RESET)
        return {};

    return getRecursiveState(index);
}

QVariant Tlu16SfpModel::foregroundData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case Tlu16SfpColumns::COL_STATUS:
        if(auto stPtr = getChStatusPtr(item))
            if(!stPtr->isStatusOk())
                return QColor(Qt::red);
        break;
    default:;
    }

    return {};
}

QVariant Tlu16SfpModel::limitsData(const QModelIndex &index) const
{
    return std::numeric_limits<quint16>::max();
}

Tlu16SfpModelItem *Tlu16SfpModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<Tlu16SfpModelItem*>(index.internalPointer());
}

QVariant Tlu16SfpModel::getRecursiveState(const QModelIndex &mIndex) const
{
    auto rows = rowCount(mIndex);
    if(rows) {
        QVariant allSt;
        for(int r=0; r<rows; ++r) {
            auto st = getRecursiveState(index(r, mIndex.column(), mIndex));
            if(st.isNull())
                continue;
            if(allSt.isNull())
                allSt = st;
            else if(allSt != st)
                return Qt::PartiallyChecked;
        }
        return allSt;
    } else {
        auto col = getColumnByIndex(mIndex);
        auto ptr = getConfigBoolVectorByCol(col);
        auto item = getItemByIndex(mIndex);
        auto devIt = config.knownConfigs.constFind(item->dev);
        if(devIt == config.knownConfigs.cend())
            return {};
        if(ptr) {
            const auto &chVector = devIt->tlu16SfpTrigger.*ptr;
            if(chVector.size() <= item->ch)
                return {};
            return boolToCheck(chVector.value(item->ch));
        } else {
            const auto &chCfg = devIt->tluDecDes.chReset;
            if(chCfg.size() <= item->ch)
                return {};
            return boolToCheck(chCfg.value(item->ch));
        }
    }
}

const TluDecDesChStatus *Tlu16SfpModel::getChStatusPtr(Tlu16SfpModelItem *mItem, bool offset) const
{
    if(mItem->ch == -1)
        return nullptr;
    const auto &statMap = offset ? devStatusOffset : devStatus;
    auto it = statMap.constFind(mItem->dev);
    if(it == statMap.cend())
        return nullptr;
    if(it->chSt.size() <= mItem->ch)
        return nullptr;

    return &it->chSt[mItem->ch];
}

QPair<std::optional<quint32>, std::optional<quint32> > Tlu16SfpModel::getCounterPair(Tlu16SfpModelItem *item) const
{
    QPair<std::optional<quint32>, std::optional<quint32> > ret;
    if(auto stPtr = getChStatusPtr(item)) {
        ret.first = stPtr->rxCodeErrCnt;
        if(auto stOffPtr = getChStatusPtr(item, true))
            ret.second = stOffPtr->rxCodeErrCnt;
    }
    return ret;
}
