#include "TluTtlIoWidget.h"

#include <QGridLayout>
#include <QComboBox>
#include <QLabel>
#include <QSizePolicy>

#include <optional>

#include "base/Globals.h"

TluTtlIoWidget::TluTtlIoWidget(QWidget *parent)
    : QWidget(parent)
    , gridLayout(new QGridLayout(this))
{
    setObjectName(__func__);
}

TluTtlIoWidget::~TluTtlIoWidget()
{
}

void TluTtlIoWidget::setTtlIoStartupCfg(const DeviceIndex &index, const TluTtlIoControlConfig &cfg)
{
    if(isSkippedDevice(index))
        return;
    auto it = startupConfigs.find(index);
    if(it != startupConfigs.end() && *it == cfg)
        return;
    startupConfigs[index] = cfg;
    unifyConfigs();
}

void TluTtlIoWidget::onFsmStateChange(const FsmState &state)
{
    bool en = state==FsmState::Ready || state==FsmState::Error
        || Globals::getDebugMode();
    setEnabled(en);
}

void TluTtlIoWidget::setConfig(const ConfigKey &, const BaseDeviceAppConfig &_config)
{
    if(config == _config)
        return;

    bool hasChanges = false;
    auto it2 = _config.knownConfigs.cbegin();
    for(auto it1=config.knownConfigs.cbegin(); it1!=config.knownConfigs.cend(); ++it1) {
        if(it2==_config.knownConfigs.cend()) {
            hasChanges = true;
            break;
        }
        if(it1.key() != it2.key() ||
            (!isSkippedDevice(it1.key()) && it1->tluTtlIo != it2->tluTtlIo)) {
            hasChanges = true;
            break;
        }
        ++it2;
    }
    if(it2!=_config.knownConfigs.cend()) {
        hasChanges = true;
    }

    config = _config;
    if(hasChanges) {
        unifyConfigs();
        refreshGui();
    }
}

void TluTtlIoWidget::ttlItemChanged(int index)
{
    for(auto w : singleTtlWidgets) {
        if(w->cb != sender())
            continue;
        Q_ASSERT(curTtlCfg.ttlCfg.size()>w->id);
        auto &ttlCfg = curTtlCfg.ttlCfg[w->id];
        auto cbData = w->cb->itemData(index).toUInt();
        if(ttlCfg.all == cbData)
            return;
        ttlCfg.all = cbData;
        break;
    }

    applyCfgChange();
}

void TluTtlIoWidget::addTtl(reg_tlu_ttl_io_t r)
{
    auto id = singleTtlWidgets.size();

    auto w = new SingleTtlWidget(id, this);
    singleTtlWidgets.append(w);
    gridLayout->addWidget(w->lab, id, 0, 1, 1);
    gridLayout->addWidget(w->cb, id, 1, 1, 1);

    w->setRegConfig(r);
    connect(w->cb, SIGNAL(currentIndexChanged(int)), this, SLOT(ttlItemChanged(int)));
}

void TluTtlIoWidget::removeTtl()
{
    auto w = singleTtlWidgets.takeLast();
    gridLayout->removeWidget(w->lab);
    gridLayout->removeWidget(w->cb);
    delete w;
}

int TluTtlIoWidget::getChNum() const
{
    for(auto cfg : startupConfigs)
        if(int s = cfg.ttlCfg.size())
            return s;
    return 0;
}

void TluTtlIoWidget::unifyConfigs()
{
    std::optional<TluTtlIoControlConfig> startupTtlCfg;
    for(auto c : startupConfigs) {
        if(c.ttlCfg.size()) {
            startupTtlCfg = c;
            break;
        }
    }

    std::optional<TluTtlIoControlConfig> newTtlCfg;
    for(auto it=config.knownConfigs.begin(); it!=config.knownConfigs.end(); ++it) {
        if(isSkippedDevice(it.key()))
            continue;
        if(it->tluTtlIo.ttlCfg.size()) {
            newTtlCfg = it->tluTtlIo;
            break;
        }
    }
    if(newTtlCfg) {
        if(startupTtlCfg) {
            while(newTtlCfg->ttlCfg.size() < startupTtlCfg->ttlCfg.size()) {
                newTtlCfg->ttlCfg.push_back(startupTtlCfg->ttlCfg[newTtlCfg->ttlCfg.size()]);
            }
        }
    } else {
        if(startupTtlCfg)
            newTtlCfg = startupTtlCfg;
        else
            return;
    }

    bool modified = false;
    for(auto it=config.knownConfigs.begin(); it!=config.knownConfigs.end(); ++it) {
        if(isSkippedDevice(it.key()))
            continue;
        if(it->tluTtlIo != *newTtlCfg) {
            it->tluTtlIo = *newTtlCfg;
            modified = true;
        }
    }

    if(modified) {
        refreshGui();
        emit configChanged(config);
    }
}

void TluTtlIoWidget::refreshGui()
{
    TluTtlIoControlConfig newTtlCfg;
    for(auto it=config.knownConfigs.begin(); it!=config.knownConfigs.end(); ++it) {
        if(isSkippedDevice(it.key()))
            continue;
        if(it->tluTtlIo.ttlCfg.size()) {
            newTtlCfg = it->tluTtlIo;
            break;
        }
    }
    if(curTtlCfg == newTtlCfg)
        return;
    curTtlCfg = newTtlCfg;

    const auto &cnt = curTtlCfg.ttlCfg.size();
    // Remove extra
    while(singleTtlWidgets.size()>cnt)
        removeTtl();

    // Update existing
    for(int i=0; i<singleTtlWidgets.size(); ++i)
        singleTtlWidgets[i]->setRegConfig(curTtlCfg.ttlCfg[i]);

    // Add new
    for(int i=singleTtlWidgets.size(); i<cnt; ++i)
        addTtl(curTtlCfg.ttlCfg[i]);
}

void TluTtlIoWidget::applyCfgChange()
{
    bool modified = false;
    for(auto it=config.knownConfigs.begin(); it!=config.knownConfigs.end(); ++it) {
        if(isSkippedDevice(it.key()))
            continue;
        if(it->tluTtlIo != curTtlCfg) {
            it->tluTtlIo = curTtlCfg;
            modified = true;
        }
    }
    if(modified)
        emit configChanged(config);
}

bool TluTtlIoWidget::isSkippedDevice(const DeviceIndex &i) const
{
    return i.getDevId() != deviceIdFilter;
}

SingleTtlWidget::SingleTtlWidget(int id, QWidget *parent)
    : QWidget(parent)
    , id(id)
    , lab(new QLabel(this))
    , cb(new QComboBox(this))
{
    lab->setText(QString("LEMO #%1").arg(id+1));
    lab->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Preferred);

    QSizePolicy sp = cb->sizePolicy();
    sp.setHorizontalPolicy(QSizePolicy::Fixed);
    cb->setSizePolicy(sp);
    QStringList cbToolTip;
    cbToolTip << "Inputs:";
    for(auto i :TluTtlIoControlConfig::getInputs()) {
        reg_tlu_ttl_io_t r;
        r.setInput(i);
        cb->addItem(TluTtlIoControlConfig::getName(i).prepend("In: "), r.all);
        cbToolTip << QString("%1 - %2").arg(TluTtlIoControlConfig::getName(i))
                         .arg(TluTtlIoControlConfig::getPrettyName(i));
    }
    cb->insertSeparator(cb->count());
    cbToolTip << "" << "Outputs:";
    for(auto i :TluTtlIoControlConfig::getOutputs()) {
        reg_tlu_ttl_io_t r;
        r.setOutput(i);
        cb->addItem(TluTtlIoControlConfig::getName(i).prepend("Out: "), r.all);
        cbToolTip << QString("%1 - %2").arg(TluTtlIoControlConfig::getName(i))
                         .arg(TluTtlIoControlConfig::getPrettyName(i));
    }
    cb->setToolTip(cbToolTip.join('\n'));
}

SingleTtlWidget::~SingleTtlWidget()
{
    delete cb;
    delete lab;
}

void SingleTtlWidget::setRegConfig(const reg_tlu_ttl_io_t &regCfg)
{
    const auto &targ = QVariant::fromValue(regCfg.all);
    int id = cb->findData(targ);
    if(id!=-1)
        cb->setCurrentIndex(id);
    else {
        for(int i=0; i<cb->count(); ++i) {
            qInfo() << i << cb->itemText(i) << cb->itemData(i);
        }
    }
}
