#include "MscStatusPublisher.h"

#include "MscSignalRouter.h"
#include "config/MscConfigDispatcher.h"

#include "base/Globals.h"
#include "message-bus/MessagePublisher.h"
#include "message-bus/impl/AbstractPublisher.h"
#include <optional>

#include <QDateTime>
#include <QDebug>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QTimer>

namespace {
constexpr const char k_state[] = "state";
constexpr const char k_fsm[] = "fsm";

static const int GUARD_TIMEOUT = 1000;


QJsonObject histToJsonObject(const MscDevHists &devHists, const QMap<int, QString> &chTitles)
{
    QJsonObject histsObj;
    // point
    constexpr const char *k_x = "x";
    constexpr const char *k_y = "y";
    // hist
    constexpr const char *k_hist_plot = "plot";
    constexpr const char *k_hist_time = "time";
    constexpr const char *k_title = "title";
    // MscHists
    constexpr const char *k_msc_hists = "msc_hists";

    for(auto hIt=devHists.hists.cbegin(); hIt!=devHists.hists.cend(); ++hIt) {
        QJsonObject histObj;
        {
            QJsonArray wf;
            for(auto p : hIt->plot) {
                wf.append(QJsonObject{{k_x,p.x}, {k_y,p.y}});
            }
            histObj[k_hist_plot] = wf;

        }
        {
            QJsonArray wf;
            for(auto p : hIt->time) {
                wf.append(QJsonObject{{k_x,p.x}, {k_y,p.y}});
            }
            histObj[k_hist_time] = wf;
        }
        {
            auto t = chTitles[hIt.key()];
            if(!t.isEmpty())
                histObj[k_title] = t;
        }

        histsObj[QString::number(hIt.key())] = histObj;
    }

    QJsonObject ret;
    ret[k_msc_hists] = histsObj;
    ret["cur_time"] = QJsonValue::fromVariant(QDateTime::currentDateTime().toTime_t());

    return ret;
}
}

struct MscHistPublishStatus
{
    QElapsedTimer lastFlushTimer;
    std::optional<MscDevHists> pendingHist;
};

class MscStatusPublisherPrivate
{
public:
    QMap<DeviceIndex, MscHistPublishStatus> devHists;
};

MscStatusPublisher::MscStatusPublisher(QObject *parent)
    : QObject(parent)
    , flushTimer(new QTimer(this))
    , d_ptr(new MscStatusPublisherPrivate())
{
    setObjectName(__func__);
    auto &router = MscSignalRouter::instance();
    router.setStatusPublisher(this);

    connect(flushTimer, &QTimer::timeout, this, &MscStatusPublisher::flushHists);
    flushTimer->setSingleShot(true);
    MscConfigDispatcher::instance().subscribeConfig(this);
}

void MscStatusPublisher::configSelectorChanged(const ConfigIndexName &conf)
{
    config_index = conf;
}

void MscStatusPublisher::pub(const QString &key, const QJsonDocument &json)
{
    MessagePublisher::instance().pub(key, json.toJson(QJsonDocument::Compact));
}

void MscStatusPublisher::onDeviceMgrFsmStateChanged(const FsmState &state)
{
    QString key = QStringList{k_state, config_index.getIndex().getIndex() ,
                              k_fsm, "main"}.join("/");
    const auto &json = QJsonDocument(QVariant::fromValue(state).toJsonObject());
    pub(key, json);
}

void MscStatusPublisher::histUpdated(const DeviceIndex &index, const MscDevHists &h)
{
    Q_D(MscStatusPublisher);
    d->devHists[index].pendingHist = h;

    if(!flushTimer->isActive())
        flushTimer->start(0);
}

void MscStatusPublisher::flushHists()
{
    Q_D(MscStatusPublisher);

    QElapsedTimer earliestTimer;
    for(auto devIt=d->devHists.begin(); devIt!=d->devHists.end(); ++devIt) {
        if(!devIt->pendingHist.has_value())
            continue;

        if(devIt->lastFlushTimer.isValid()) {
            auto v = devIt->lastFlushTimer.elapsed();
            if(v<GUARD_TIMEOUT) {
                if(!earliestTimer.isValid() || devIt->lastFlushTimer.msecsTo(earliestTimer)<0)
                    earliestTimer = devIt->lastFlushTimer;
                continue;
            }
        }

        flushHist(devIt.key());
    }
    if(earliestTimer.isValid()) {
        flushTimer->start(GUARD_TIMEOUT-earliestTimer.elapsed());
    }
}

void MscStatusPublisher::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    cfg = newConfig;
}

void MscStatusPublisher::flushHist(const DeviceIndex &index)
{
    Q_D(MscStatusPublisher);

    auto &devSt = d->devHists[index];
    auto &h = devSt.pendingHist;
    if(!h.has_value())
        return;

    QString key = QStringList{Globals::getDaqModule().toString(), index.getSerialStr()}.join("/");

    QMap<int, QString> chTitles;
    const auto &histCfg = cfg.findDevice(index).mscCore.hist;
    for(auto ch : histCfg.plot.keys()) {
        auto name = histCfg.plot[ch].name;
        if(!name.isEmpty()) {
            chTitles[ch] = name;
        }
    }
    auto json = histToJsonObject(h.value(), chTitles);
    pub(key, QJsonDocument(json));
    h.reset();
    devSt.lastFlushTimer.start();
}
