/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef RCCOREAPP_H
#define RCCOREAPP_H

#include <QObject>
#include <QPointer>

#include "AbstractConfigurable.h"
#include "CoreTypes.h"
#include "RcConfig.h"
#include "fsm/FsmState.h"
#include "TrigStatus.h"
#include "pnp-server/PNPMessage.h"
#include "base/ProgramIndex.h"
#include "run-control-client-manager/RcRunInfo.h"
#include "remote-control-server/RcProgramState.h"

class ConfigIndexName;
class ConfigSelector;
class DeviceDispatcher;
class DeviceIndex;
class DeviceIndexEnabledMap;
class DiscoverDialog;
class EvNumChecker;
class LldpDiscovery;
class QHostAddress;
class QLabel;
class QTimer;
class RcClientManager;
class RcConfigDb;
class RcConfigDispatcher;
class RcCoreApp;
class RcFsm;
class RcSignalRouter;
class RemoteControlServer;
struct RemoteControlServerStatus;
struct RunDocument;
class SlowControlServer;
class TrigCore;
class TrigDevice;
class TrigDeviceProxy;
struct TrigStatus;
class VmeTrigsController;
class VmeTrigsCore;
class StatusPublisher;

class RcCoreApp : public AbstractConfigurable
{
    Q_OBJECT
public:
    explicit RcCoreApp(QObject *parent = nullptr);
    virtual ~RcCoreApp() override;
    RcConfigDispatcher *getConfigDispatcher() { return rcConfigDispatcher; }
    DeviceDispatcher *getRcDeviceDispatcher();
    TrigCore *getTrigCore() { return trigCore; }
    VmeTrigsCore *getVmeTrigsCore() { return vmeTrigsCore; }

signals:
    void rcRunIdChanged(const RcRunId &runId) const;
    void runStart(const RcRunId &runId, const QVariant &startEvNum) const;
    void closeGlobTrig() const;
    void openGlobTrig() const;
    void prepareStop() const;
    void runStop(const QVariant &stopEvNum={}, bool force = false) const;
    void slowControlStateUpdated(QString, QString) const;
    void clientManagerRunStateChanged_(const FsmState &); //proxy
    void vmeTrigsRunStateChanged_(const FsmState &); //proxy
    void trigManagerRunStateChanged_(const FsmState &); //proxy
    void sendRemoteState(const QString &) const;
    void sendVmeFlag(bool) const;
    void sendBackPressure(const RcRunId &, bool) const;
    void updateProgramDescription(const ProgramDescription &) const;
    void closeProgram(const ProgramDescription &) const;
    void sendRunError(const RcRunId &, QString) const;
    void allowRemoteControl(bool);
    void sendRunDocumentToRc(const RunDocument &data) const;
    void sendRunDocumentToClients(const RunDocument &data) const;
    void sendSlaveMessage(const MessageTreeItem &m);
    void write_history_config(const RcRunId &, const QJsonObject &, bool megamaster_flg) const;
    void switch_to_config(const ConfigIndexName &conf) const;
    void printFsmInput() const;
    void myRcProgramStateChanged(const RcProgramState &) const;
    void allRcProgramStateChanged(const RcProgramState &) const;

public slots:
    void quit();
    void configSelectorChanged(const ConfigIndexName &conf);
    void setConfig(const RcConfig &newConfig) override;

    void onRcFsmStateChange(const FsmState &state, const FsmState &oldState);
    void setTrigState(bool state);
    void trigManagerRunStateChanged(const FsmState &state);
    void clientManagerRunStateChanged(const FsmState &state);
    void clientRcProgramStateChanged(const RcProgramState &state);
    void vmeTrigsRunStateChanged(const FsmState &state);

    void onBackPressure(const RcRunId &runId, bool backPressure);
    void onVmeFlagUpdated(bool hasVme);

    void config_transfer_from_clients(const ClientIndex &id, const QJsonObject &json);

    bool perfomeLocalStart();
    void perfomeLocalStop();
    void onRemoteStartRequest(const RcRunId &runId, const QVariant &startEvNumber);
    void onRemoteStopRequest(const RcRunId &runId, const QVariant &stopEvNumber);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);
    void trigStatusUpdated(const TrigStatus &status);
    void onSlaveErrorMessageReceived(const MessageTreeItem &m);
    void slaveMessageReceived(const MessageTreeItem &m);
    void onRunWarning(const RcRunId &runId, QString warnStr);
    void onRunError(const RcRunId &runId, QString warnStr);

private slots:
    void config_timeout();

private:
    bool performStart();
    void perfomeStop(bool isError);
    void actionOnStateRun();
    void actionPostRun();
    void incrementRunNumber();
    VmeTrigsController *vmeTrigsController();
    void checkCanSendRunDocument();
    void sendRunDocument();
    inline bool isRemoteConnected() const { return remoteConnected; }
    void logRcFsmStateChange(const FsmState &previousState, const FsmState &state);
    void refreshProgramDescription();
    void sendMyRcProgramState() const;

    QMap<QString, QString> createProgramDescriptionOptions();
    static QString createRemoteState(const RcClientConfigMap &clients,
                                     const QMap<DeviceIndex,
                                     VmeTrigConfig> &vmeTrigs);
    void updateRemoteState();
    SlaveType mapSenderToSlaveType(QObject *sender) const;
    bool isInRun() const;
    ClientIndexSet getMissingConfigClients() const;
    void setMegaMasterFlag(bool mm);

    RcConfig config;
    RcConfigDb * const configDb;
    ProgramIndex localProgramIndex;
    RcConfigDispatcher * const rcConfigDispatcher;
    TrigCore * const trigCore;
    RcClientManager * const clientCore;
    VmeTrigsCore * const vmeTrigsCore;
    SlowControlServer * const slowControl;
    RemoteControlServer * const remoteControl;
    bool remoteConnected = false;
    ProgramDescription progDesc;
    EvNumChecker * const evNumChecker;
    QTimer * const config_timer;
    QMap<ClientIndex, QJsonObject> clients_config;
    bool megamaster_flg = false;
    RcFsm * const fsm;
    TrigStatus lastTrigStatus;
    StatusPublisher * const statusPublisher;
    // RcSlaveStatus slaveStatus;
    bool hasVme = false;
    RcRunInfo rcRunInfo;
    RcProgramState rcProgramState; // mine view of tree
};

#endif // RCCOREAPP_H
