/*
**    Copyright 2014-2015 Ivan Filippov
**    Copyright 2014-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcMainWindow.h"
#include "ui_RcMainWindow.h"

#include "RcConfigDispatcher.h"
#include "RcCoreApp.h"
#include "RcSignalRouter.h"
#include "SlowControlDialog.h"
#include "TrigCore.h"
#include "TrigDeviceController.h"
#include "VmeTrigsController.h"
#include "VmeTrigsCore.h"
#include "base/config/ConfigSelector.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "DlnpCalibrationTrigger/DlnpCalibrationTriggerDialog.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "remote-control-server/RemoteControlServerStatus.h"
#include "run-control-client-manager/RunStates.h"
#include "widgets/DeviceStatusDialog.h"
#include "widgets/LogViewWidget.h"

#include <QDateTime>
#include <QDebug>
#include <QDesktopWidget>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QHostAddress>
#include <QInputDialog>
#include <QJsonDocument>
#include <QJsonObject>
#include <QLabel>
#include <QMessageBox>
#include <QSettings>
#include <QShortcut>
#include <QStatusBar>
#include <QStyle>
#include <QTabBar>
#include <QThread>
#include <QTimer>

Q_DECLARE_METATYPE(Qt::GlobalColor)

static const char *styleLargeFont = "font-size: 18pt";

RcMainWindow::RcMainWindow(QWidget *parent) :
    BaseMainWindow(parent),
    ui(new Ui::RcMainWindow),
    labelFsmStateRc(new QLabel(this)),
    labelFsmStateTrig(new QLabel(this)),
    labelFsmStateClient(new QLabel(this)),
    labelFsmStateVme(new QLabel(this)),
    labelDevState(new QLabel(this)),
    labelSlowControl(new QLabel(this)),
    labelDevice(new QLabel(this)),
    labelFirmware(new QLabel(this)),
    labelTemp(new QLabel(this)),
    discoverDialog(new DiscoverDialog(this)),
    slowControlDialog(new SlowControlDialog(this)),
    deviceStatusDialog(new DeviceStatusDialog(this)),
    dlnpCalibrationTriggerDialog(new DlnpCalibrationTriggerDialog(this)),
    rcCoreApp(new RcCoreApp(this))
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);

    ui->tabWidget->addTab(new LogViewWidget(ui->tabWidget), "Log");
    ui->tabWidget->setCurrentIndex(0); // always start with main controls view

    discoverDialog->setObjectName("MainDiscoverDialog");

    qRegisterMetaType<RUN_STATES>();
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<TrigStatus>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<RcTrigDeviceConfig>();

    // setup VmeTrigsWidget
    auto vmeTrigsWidget = ui->vmeTrigsWidget;
    auto vmeTrigsCore = rcCoreApp->getVmeTrigsCore();
    auto vmeTrigsController = vmeTrigsCore->getVmeTrigsController();
    auto vmeDeviceDispatcher = vmeTrigsCore->getDeviceDispatcher();
    vmeTrigsWidget->setController(vmeTrigsController);
    vmeTrigsWidget->setDeviceDispatcher(vmeDeviceDispatcher);

    auto rcConfigDispatcher = rcCoreApp->getConfigDispatcher();
    ui->trigMainForm->setConfigDispatcher(rcConfigDispatcher);

    auto trigCore = rcCoreApp->getTrigCore();
    auto trigDeviceController = trigCore->getTrigDeviceController();
    auto trigDeviceDispatcher = trigCore->getDeviceDispatcher();
    trigDeviceDispatcher->subscribeDevices(discoverDialog);
    trigDeviceDispatcher->subscribeDevices(this);

    LldpDiscovery::instance().subscribe(discoverDialog, trigDeviceController->getDeiveIdFilter());

    connect(rcCoreApp, &RcCoreApp::sendSlaveMessage,
            this, &RcMainWindow::slaveMessageReceived);

    init();
}

void RcMainWindow::closeEvent(QCloseEvent *event)
{
    discoverDialog->close();

    rcCoreApp->quit();

    emit destroyAll();
    BaseMainWindow::closeEvent(event);
}

RcMainWindow::~RcMainWindow()
{
    delete ui->trigMainForm;
    delete ui->clientManagerWidget;
    delete ui->vmeTrigsWidget;
    delete ui;
}

void RcMainWindow::init()
{
    ui->groupBoxWarn->hide();
    ui->labelWarTrigDev->hide();
    ui->labelWarClients->hide();
    ui->labelWarTmwr->hide();
    ui->labelWarnTrigDev->hide();
    ui->labelWarnClients->hide();
    ui->labelWarnTmwr->hide();
    ui->labelWarnTrigDev->setStyleSheet("QLabel { color : orange;}");
    ui->labelWarnClients->setStyleSheet("QLabel { color : orange;}");
    ui->labelWarnTmwr->setStyleSheet("QLabel { color : orange;}");

    ui->mainToolBar->hide();

    labelFsmStateRc->setMinimumWidth(150);
    labelFsmStateRc->setStyleSheet("font-weight: bold;");
    labelFsmStateClient->setMinimumWidth(150);
    labelFsmStateTrig->setMinimumWidth(150);
    labelFsmStateVme->setMinimumWidth(150);

    statusBar()->addWidget(labelFsmStateRc);
    statusBar()->addWidget(labelFsmStateTrig);
    statusBar()->addWidget(labelFsmStateClient);
    statusBar()->addWidget(labelFsmStateVme);
    statusBar()->addPermanentWidget(labelSlowControl);
    statusBar()->addPermanentWidget(labelDevState);
    statusBar()->addPermanentWidget(labelDevice);
    statusBar()->addPermanentWidget(labelTemp);
    statusBar()->addPermanentWidget(labelFirmware);

//    ui->groupBoxRunControl->setStyleSheet();
    ui->pushButtonRestartAll->setStyleSheet(styleLargeFont);
    ui->pushButtonStopAll->setStyleSheet(styleLargeFont);
    ui->pushButtonRestartAll->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButtonStopAll->setIcon(style()->standardIcon(QStyle::SP_MediaStop));
//  labelState->setStyleSheet("QLabel {color: red}");

    configuration_manager = new ConfigurationManager(this);

    connect(ui->pushButtonRestartAll, &QAbstractButton::clicked,
            this, &RcMainWindow::perfomeLocalStart,
            Qt::UniqueConnection);
    connect(ui->pushButtonStopAll, &QAbstractButton::clicked,
            this, &RcMainWindow::perfomeLocalStop,
            Qt::UniqueConnection);
    connect(configuration_manager.data(), &ConfigurationManager::switch_to_config,
            this, &RcMainWindow::configSelectorChanged,
            Qt::UniqueConnection);

    auto &router = RcSignalRouter::instance();
    router.addClientManagerWidget(ui->clientManagerWidget);
    router.addTrigMainForm(ui->trigMainForm);
    router.addMainWindow(this);
    router.addConfigurationManager(configuration_manager);
    router.addSlowControlDialog(slowControlDialog);
    router.setCalibrationTriggerDialog(dlnpCalibrationTriggerDialog);
    router.connectAll();

    configuration_manager->setDefaultConfig(RcConfig());

    loadSettings();

//    if(config.rcTrigConfig.deviceIndex) {
//        quint32 ipAddr = 0;
//        if(!config.rcTrigConfig.hostStr.isEmpty()){
//            QHostAddress host(config.rcTrigConfig.hostStr);
//            if(!host.isNull())
//                ipAddr = host.toIPv4Address();
//        }
//        DeviceDescription dd(config.rcTrigConfig.rcDeviceId,
//                             config.rcTrigConfig.deviceIndex,
//                             ipAddr);
//        connectToTrig(dd);
//    }
}

void RcMainWindow::setConfig(const RcConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;
    // update GUI elements that depend on config
    updateStatStopButton();
    loadSettings();
}

void RcMainWindow::setDeviceList(const DeviceIndexEnabledMap &devices)
{
    updateDeviceLabel();
}

static QString map_fsm_to_ui_style(const FsmState &state)
{
    QString color;
    switch (state) {
    case FsmState::Init:;
    case FsmState::Error:
        color = "red";
        break;
    case FsmState::Ready:;
        break;
    case FsmState::Run:;
        color = "lightgreen";
        break;
    case FsmState::Starting:;
    case FsmState::Stopping:
        color = "orange";
        break;
    default:
    {}
    }
    return QString("QLabel { %1 font-size: 14pt}")
            .arg(color.isNull() ? "" : "background-color:" + color + ";");
};

void RcMainWindow::logRcFsmStateChange(const FsmState &previousState, const FsmState &state)
{
    auto RunIdStr = QString(rcRunId);
    auto previousStateStr = fsmStateToString(previousState);

    ui->labelRunState->setStyleSheet(map_fsm_to_ui_style(state));

    if (state == FsmState::Starting) {
        ui->labelRunState->setText(QString("Starting Run %1...").arg(RunIdStr));
    }

    if (state == FsmState::Stopping) {
        ui->labelRunState->setText(QString("Stopping Run %1...").arg(RunIdStr));
    }

    if (state == FsmState::Run) {
        QString txt = QString("Run started: %1").arg(RunIdStr);
        ui->labelRunState->setText(txt);
    }
    if (state == FsmState::Ready) {
        ui->labelRunState->setText(QString("Previous Run %1").arg(RunIdStr));
    }
    if (state == FsmState::Error) {
        switch (previousState) {
        case FsmState::Starting:
        case FsmState::Stopping:
        case FsmState::Run: {
            QString txt = QString("%1 '%2' failed").arg(previousStateStr, RunIdStr);
            ui->labelRunState->setText(txt);
            break;
        }
        default:;
        }
    }
}

void RcMainWindow::onRcFsmStateChange(const FsmState &state, const FsmState &previousState)
{
    logRcFsmStateChange(curRcFsmState, state);
    if(state == FsmState::Starting) {
    // run has been started
        ui->trigMainForm->clearStats();
    }
    curRcFsmState = state;

    labelFsmStateRc->setText("RcFsm: " + fsmStateToString(state));
    updateStatStopButton();
    updateGuiControlBlock();
}

void RcMainWindow::trigManagerRunStateChanged(const FsmState &state)
{
    subFsmStateChanged(TAB_TRIG_DEV, state);
}

void RcMainWindow::clientManagerRunStateChanged(const FsmState &state)
{
    subFsmStateChanged(TAB_CLIENTS, state);
}

void RcMainWindow::vmeTrigsRunStateChanged(const FsmState &state)
{
    subFsmStateChanged(TAB_VME_TRIGS, state);
}

void RcMainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = underRemoteCtrl != status.portInUse;
    underRemoteCtrl = status.portInUse;
    if(changed)
        updateStatStopButton();
}

void RcMainWindow::setRcRunId(const RcRunId &id)
{
    rcRunId = id;
}

void RcMainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void RcMainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    const ConfigSelector cs(DaqModule(DaqModule::Type::RunControl), conf);
    setWindowTitle(cs.toWindowTitle());

    // Initialize MetricSender
    MetricSender::setBaseTag("index", cs.getIndex().toString());
}

void RcMainWindow::clearRunWarning()
{
    ui->labelWarClients->hide();
    ui->labelWarnClients->clear();
    ui->labelWarnClients->hide();

    ui->labelWarTrigDev->hide();
    ui->labelWarnTrigDev->hide();
    ui->labelWarnTrigDev->clear();

    ui->labelWarTmwr->hide();
    ui->labelWarnTmwr->hide();
    ui->labelWarnTmwr->clear();
    updateGroupBoxWarn();
}

bool RcMainWindow::isInRun() const
{
    return curRcFsmState != FsmState::Error && curRcFsmState != FsmState::Ready;
}

void RcMainWindow::onSlowCntStatusUpdated(quint16 tcpPort, bool listening, bool portInUse)
{
    QString txt = "SlowControl: ";

    if(listening)
        if(portInUse)
            txt.append("connected");
        else
            txt.append("disconnected");
    else
        txt.append("disabled");

    labelSlowControl->setStyleSheet(QString("QLabel { color : %1; }")
                                    .arg(listening ? (portInUse?"green":"red") : "gray"));

    if(tcpPort)
        labelSlowControl->setToolTip(QString("Listening on %1 port").arg(tcpPort));
    labelSlowControl->setText(txt);
}

void RcMainWindow::subFsmStateChanged(int tabIndex, const FsmState &state)
{
    const bool isOk = state!=FsmState::Error;
    setTabState(state!=FsmState::Error, tabIndex);
    QLabel *label = nullptr;
    QString prefix;
    switch (tabIndex) {
    case TAB_TRIG_DEV:
        label = labelFsmStateTrig;
        prefix = "TrigFsm";
        break;
    case TAB_CLIENTS:
        label = labelFsmStateClient;
        prefix = "ClientFsm";
        break;
    case TAB_VME_TRIGS:
        label = labelFsmStateVme;
        prefix = "VmeTrigFsm";
        break;
    default:;
    }
    if(label) {
        label->setText(prefix + ": " + fsmStateToString(state));
        auto palette = label->palette();
        palette.setColor(label->foregroundRole(), isOk ? QColor() : Qt::red);
        label->setPalette(palette);
    }
}

void RcMainWindow::setTabState(bool isOk, int tabIndex)
{
    QTabBar *bar = ui->tabWidget->tabBar();
    const auto curTabCol = bar->tabTextColor(tabIndex);
    if(curTabCol.isValid() == isOk)
        bar->setTabTextColor(tabIndex, isOk ? QColor() : Qt::red);
}

void RcMainWindow::updateGuiControlBlock()
{
    bool block = curRcFsmState == FsmState::Starting ||
            curRcFsmState == FsmState::Stopping;
    const bool enabled = !block;
    ui->menuOptions->setEnabled(enabled);
    bool isReadyOrErr = !isInRun();
    ui->actionFindDevice->setEnabled(isReadyOrErr);
    ui->actionSet_local_Run_number->setEnabled(isReadyOrErr);
    ui->actionCheck_LVDS_parity->setEnabled(isReadyOrErr);

// it is independent
//    ui->actionSlow_control->setEnabled(true);
}

/*
bool RcMainWindow::waitVmeTrigsStarted(const QElapsedTimer &startTime)
{
    bool result = true;
    // Get started state from Vme Modules
    while (! startTime.hasExpired(START_TIMEOUT)) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        if (vmeTrigsController()->get_runId()==config.pc.runId &&
                vmeTrigsController()->getRunState() != RS_STARTING )
            // Break on error or started
            break;
    }
    if (vmeTrigsController()->getRunState() != RS_STARTED){
        qCritical() << "Some vme trigger failed to reset EvNumber";
        result = false;
    }
    return result;
}
*/
/*
bool RcMainWindow::waitClientsStarted(const QElapsedTimer &startTime)
{
    bool result = true;
    // Get started state from Tcp Clients (sub-programms)
    while (! startTime.hasExpired(START_TIMEOUT)) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        if(ui->clientManagerWidget->get_run_id()==config.pc.runId &&
                ui->clientManagerWidget->getRunState() != RS_STARTING)
            // Break on error or started
            break;
    }
    auto sttt = ui->clientManagerWidget->getRunState();
    if(sttt != RS_STARTED){
        qDebug() << "Some client failed to start";
        result = false;
    }
    return result;
}
*/


void RcMainWindow::updateDeviceLabel()
{
    auto deviceIndex = config.rcTrigConfig.index;
    bool devicePresent = ! deviceIndex.isNull();
    if (devicePresent) {
        QString devLabelStr;
        const QString serialStr = deviceIndex.getSerialStr();
        devLabelStr = QString("%1")
            .arg(deviceIndex.getIdent());
//        if(config.rcTrigConfig.hostStr.isEmpty())
//            devLabelStr.append(QString(", IP %3").arg(config.rcTrigConfig.hostStr));
//        else
//            devLabelStr.append(", no IP");
        labelDevice->setText(devLabelStr);
    } else {
        labelDevice->clear();
        labelDevState->setText("Device status: unknown");
        updateDeviceStatusLabel(TrigStatus());
    }
}

void RcMainWindow::updateDeviceStatusLabel(const TrigStatus &status)
{
    labelTemp->setText(TempsInfo(status.getTemps()).brief());
    labelFirmware->setText(QString("f/w:%1").arg(status.getFwStr()));
    labelDevice->setStyleSheet(status.conn.isReady() ? "" : "background-color : red");
}

void RcMainWindow::trigStatusUpdated(const TrigStatus &status)
{
    deviceStatusDialog->setIndex(status.getDeviceIndex()); // single device mode
    deviceStatusDialog->setStatus(status);

    updateDeviceStatusLabel(status);

    static bool oldStatusReady = false;
    if (status.conn.isReady() != oldStatusReady) {
        oldStatusReady = status.conn.isReady();
        updateDeviceLabel();
    }
}

void RcMainWindow::slaveMessageReceived(const MessageTreeItem &m)
{
    if (!m.isEmpty() && (!config.pc.runEnable || rcRunId != m.runId) )
        return;

    QLabel *lab=nullptr;
    QLabel *label=nullptr;
    switch (m.type()) {
    case SlaveType::TcpClient:
        lab = ui->labelWarClients;
        label = ui->labelWarnClients;
        break;
    case SlaveType::Trig:
        lab = ui->labelWarTrigDev;
        label = ui->labelWarnTrigDev;
        break;
    case SlaveType::VmeTrig:
        lab = ui->labelWarTmwr;
        label = ui->labelWarnTmwr;
        break;
    default:
        return;
    }

    QString warnStr = m.text();
    const int WARN_SIZE_LIMIT= 90;
    if(warnStr.size()>WARN_SIZE_LIMIT) {
        warnStr = warnStr.left(WARN_SIZE_LIMIT-3)+"...";
    }
    Q_ASSERT(label);
    Q_ASSERT(lab);
    if (label && lab) {
        label->setText(warnStr);
        label->setHidden(warnStr.isEmpty());
        lab->setHidden(label->isHidden());
    }

    updateGroupBoxWarn();
}

void RcMainWindow::loadSettings()
{
    updateDeviceLabel();

//    ui->trigMainForm->init(&config);
//    ui->vmeTrigsWidget->init(&config);

    bool oldState1 = ui->actionAllow_Remote_Control->blockSignals(true);
    ui->actionAllow_Remote_Control->setChecked(config.pc.allowRemoteControl);
    ui->actionAllow_Remote_Control->blockSignals(oldState1);
    const QSettings settings;
    auto enableStatusDump = settings.value("enableStatusDump", false).toBool();
    ui->actionEnableStatusDump->setChecked(enableStatusDump);
    emit setStatusDumpEnable(enableStatusDump);
    auto statusDumpDirName = settings.value("statusDumpDirName").toString();
    emit setStatusDumpDirName(statusDumpDirName);

//    updateProgramDescription();
}

void RcMainWindow::updateStatStopButton()
{
    bool enableStart = curRcFsmState == FsmState::Ready;
    bool enableStop = curRcFsmState == FsmState::Run || curRcFsmState == FsmState::WaitSpillEnd;
//    bool isRunning = config.pc.runEnable;
    ui->pushButtonRestartAll->setEnabled(enableStart && !underRemoteCtrl);
    ui->pushButtonStopAll->setEnabled(enableStop && !underRemoteCtrl);
}

void RcMainWindow::updateGroupBoxWarn()
{
    QStringList warnStrList;
    bool hidden = true;
    if(!ui->labelWarnTrigDev->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnTrigDev->text();
    }
    if(!ui->labelWarnClients->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnClients->text();
    }
    if(!ui->labelWarnTmwr->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnTmwr->text();
    }

    ui->groupBoxWarn->setHidden(hidden);
    if(!hidden)
        emit sendRunWarn(rcRunId, warnStrList.join("; "));
}

void RcMainWindow::updateRemoteBlocking()
{
    // TODO
    //    ui->trigMainForm->updateRemoteBlocking(remoteConnected());
}

void RcMainWindow::on_actionFindDevice_triggered()
{
    discoverDialog->show();
}

void RcMainWindow::on_actionAllow_Remote_Control_toggled(bool val)
{
    config.pc.allowRemoteControl = val;
    emit configChanged(config);
}

void RcMainWindow::on_actionEnableStatusDump_toggled(bool val)
{
    QSettings settings;
    settings.setValue("enableStatusDump", val);
    emit configChanged(config);
    emit setStatusDumpEnable(val);
}

void RcMainWindow::on_actionSetStatusDumpDirectory_triggered()
{
    QSettings settings;
    auto statusDumpDirName = settings.value("statusDumpDirName").toString();
    QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"),
                                                    statusDumpDirName,
                                                    QFileDialog::ShowDirsOnly
                                                    | QFileDialog::DontResolveSymlinks);
    if (dir.isNull())
        return;
    settings.setValue("statusDumpDirName", dir);
    emit configChanged(config);
    emit setStatusDumpDirName(dir);
}

void RcMainWindow::on_tabWidget_currentChanged(int index)
{
    if(index == TAB_LOGS)
        setTabState(true, TAB_LOGS);
}

void RcMainWindow::on_actionPrint_Fsm_input_triggered()
{
    emit printFsmInput();
}

void RcMainWindow::on_actionSet_local_Run_number_triggered()
{
    bool ok;
    int val = QInputDialog::getInt(this, "Set local Run number...", "Previous Run number",
                         config.pc.localRunNumber, 0, 2147483647, 1, &ok, Qt::Dialog);
    if(ok) {
        config.pc.localRunNumber = val;
        logRcFsmStateChange(FsmState::Init, curRcFsmState);
        emit configChanged(config);
    }
}

void RcMainWindow::on_actionSlow_control_triggered()
{
    slowControlDialog->show();
}

// FIXME: duplicated code
void RcMainWindow::on_actionDevice_status_triggered()
{
    deviceStatusDialog->show();
}

void RcMainWindow::on_actionCalibration_trigger_triggered()
{
    dlnpCalibrationTriggerDialog->show();
}

