/*
**    Copyright 2014-2015 Ivan Filippov
**    Copyright 2014-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigDevice.h"

#include <cmath>

#include <QDateTime>
#include <QDebug>
#include <QFile>
#include <QHostAddress>
#include <QLoggingCategory>
#include <QTextStream>
#include <QTimer>

#include "DlnpCalibrationTrigger/DlnpCalibrationTriggerModule.h"
#include "LegacyDeviceModule.h"
#include "RunControlRegs.h"
#include "MscHist/MscHistModule.h"
#include "SpillControl/SpillControlModule.h"
#include "WrEpState.h"
#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "regio/RegOpStream.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc)
Q_LOGGING_CATEGORY(lc, "afi.rc.device")
}

namespace {
const int TRIG_STATUS_POLL_TIME_MS_MIN = 20;
const bool DEBUG_INFO = false;
}

TrigDevice::TrigDevice(const DeviceIndex &deviceIndex, QObject *parent)
    : ModularMregDevice(deviceIndex, status, parent)
    , trigStatusTimer(new QTimer(this))
    , status(getDeviceIndex())
    , statusPollTimeMs(TRIG_STATUS_POLL_TIME_MS_DEFAULT)
    , legacyDevice (createModule<LegacyDeviceModule>())
{
    setObjectName(QString("Trig_%1").arg(deviceIndex.getSerialStr()));
    qRegisterMetaType<FirmwareBuildParams>();

    regCtrl.all = 0;
    connect(trigStatusTimer, &QTimer::timeout,
            this, &TrigDevice::trigStatusPoll);
    trigStatusTimer->setSingleShot(true);
    connect(this, &TrigDevice::setStatusDumpEnable,
            &tsw, &TrigStatisticsWriter::setStatusDumpEnable);
    connect(this, &TrigDevice::setStatusDumpDirName,
            &tsw, &TrigStatisticsWriter::setStatusDumpDirName);
}

std::optional<FirmwareBuildParams> TrigDevice::readFirmwareBuildParams()
{
    std::array<quint32, 4> val;
    quint16 nch;
    RegOpVector r;
    r.RegRead32(REG32_RC_BUILD_PARAM_0, &val[0]);
    r.RegRead32(REG32_RC_BUILD_PARAM_1, &val[1]);
    r.RegRead32(REG32_RC_BUILD_PARAM_2, &val[2]);
    r.RegRead32(REG32_RC_BUILD_PARAM_3, &val[3]);
    r.RegRead16(REG_RC_CHANNEL_CNT, &nch);
    if (!legacyDevice->regOpExecRebased(r)) {
        qWarning(lc).noquote() << "Error reading FirmwareBuildParams";
        return {};
    }

    FirmwareBuildParams fbp;
    fbp.ver = val.at(0) & 0xFF;
    if (fbp.ver == 0) { // before r.25460
        fbp.edge_rollover = false;
        fbp.tih_log = false;
        fbp.multihist = false;
        fbp.nch = nch;
        fbp.ntrig = getTrigSrcCount();
        fbp.clk_hz = 62500000;
        fbp.spill_plot_ram_addr_bits = 13;
        fbp.trig_hist_ram_addr_bits = 13;
        fbp.level_cnt_bits = 48;
        fbp.edge_cntr_bits = 32;
    }
    if (fbp.ver == 0x81) { // v 1.0
        fbp.edge_rollover = val.at(0) & 0x100;
        fbp.tih_log = val.at(0) & 0x400;
        fbp.multihist = val.at(0) & 0x800;
        fbp.nch = val.at(1) & 0xFF;
        fbp.ntrig = (val.at(1) >> 8) & 0xFF;
        fbp.spill_plot_ram_addr_bits = val.at(2) & 0xFF;
        fbp.trig_hist_ram_addr_bits = (val.at(2) >> 8) & 0xFF;
        fbp.level_cnt_bits = (val.at(2) >> 16) & 0xFF;
        fbp.edge_cntr_bits = (val.at(2) >> 24) & 0xFF;
        fbp.clk_hz = val.at(3);
        if (fbp.tih_log)
            if (fbp.trig_hist_ram_addr_bits > 10)
                fbp.trig_hist_ram_addr_bits = 10;
    }
    if (histMod && histMod->getBuildParam()) {
        const MscHistBuildParams &hbp = histMod->getBuildParam().value();
        fbp.spill_plot_ram_addr_bits = hbp.ramBitsNumHistPlot;
        fbp.spill_plot_ram_start_addr = hbp.ramBaseAddrPlotHist;
        fbp.trig_hist_ram_start_addr = hbp.ramBaseAddrLogHist;
        fbp.trig_hist_ram_addr_bits = hbp.ramBitsNumLogHist;
    }

    if (! fbp.isValid()) {
        qWarning(lc).noquote() << "Invalid FirmwareBuildParams";
        return {};
    }
    return fbp;
}

bool TrigDevice::clearStats()
{
    RegOpVector r;
    regCtrl.bit.clear = 1;
    r.RegWrite16(REG_RC_CTRL, regCtrl.all);
    regCtrl.bit.clear = 0;
    r.RegWrite16(REG_RC_CTRL, regCtrl.all);
    return regOpExec(r);
}

void TrigDevice::resetAll()
{
    if (!isConnected()) return;
    const quint16 val = 0;

    RegOpVector r;
    r.RegWrite16(REG_CSR_CTRL, val | REG_CSR_CTRL_BIT_SOFTRESET);
    r.RegWrite16(REG_CSR_CTRL, val);
    regOpExec(r);

    status.conn.setUnconfigured();
    writeSetup(lastConfig, lastSetupSeqNum);
}

void TrigDevice::preRunReset(const QVariant &startEvNum)
{
    evNumReset(startEvNum);
    clearStats();
}

bool TrigDevice::writeTrigSource(const TrigSourceConfig &trig)
{
    RegOpVector r;
    r.RegWrite16(REG_RC_TRIG_SRC, trig.trigSrc);
    r.RegWrite16(REG_RC_XOFF_MASK, trig.xOffMask);
    return regOpExec(r);
}

bool TrigDevice::writeTrigCtrlStatConfig(const TrigCtrlStatConfig &trig)
{
    RegOpVector r;
    r << setTrigTimerFreqCmd(trig.trigTimerFreq);
    r << setTrigRandomFreqCmd(trig.randomPulserFreq);
    r << setTrigRandomDeadTimeCmd(trig.programmedDeadTime);
    return regOpExec(r);
}

void TrigDevice::enforceConfigConstraints(ModularDeviceConfig *config)
{
    ModularMregDevice::enforceConfigConstraints(config);
}

void TrigDevice::createModules()
{
    ModularMregDevice::createModules();
    histMod = createSingleModule<MscHistModule>();
}

bool TrigDevice::writeSetup_internal(const RcTrigDeviceConfig &lastConfig, const RcTrigDeviceConfig &cfg, bool force)
{
    auto spillMod = getFirstModule<SpillControlModule>();
    if (spillMod) {
        if (!spillMod->writeConfig(cfg.spillControl))
            return false;
    }
    if (auto mod = getFirstModule<DlnpCalibrationTriggerModule>()) {
        if (!mod->writeConfig(cfg.dlnpCalibrationTrigger))
            return false;
    }
    bool ok = true;
    if (force || lastConfig.trigCtrlStat != cfg.trigCtrlStat) {
        ok &= writeTrigCtrlStatConfig(cfg.trigCtrlStat);
    }
    if (force || lastConfig.trig != cfg.trig) {
        ok &= writeTrigSource(cfg.trig);
    }
    RegOpVector r;
    if (force || cfg.spillPlot.spillPlotStep != lastConfig.spillPlot.spillPlotStep) {
        if(histMod && histMod->getBuildParam() && histMod->getBuildParam()->plotHistNum) {
            MscHistConfig hCfg;
            auto &hChCfg = hCfg.plot[0];
            double stepInSec = (1*cfg.spillPlot.spillPlotStep) / getClockFreqHz();
            hChCfg.plotSlice = stepInSec*1e6;
            ok &= histMod->writeConfig(hCfg);
        }
        r.RegWrite32(REG32_RC_SPILLPLOT_STEP, cfg.spillPlot.spillPlotStep);
    }
    if (force || lastConfig.spillPlot.plotSrc != cfg.spillPlot.plotSrc) {
        r.RegWrite16(REG_RC_SPILLPLOT_SOURCE, cfg.spillPlot.plotSrc);
    }
    if (force || lastConfig.trig.trigOpen == cfg.trig.trigOpen
        || lastConfig.spillControl.softSpill == cfg.spillControl.softSpill) {
        r << setControlStateCmd(cfg.trig.trigOpen, cfg.spillControl.softSpill);
    }
    ok &= regOpExec(r);
    return ok;
}

void TrigDevice::writeSetup(const RcTrigDeviceConfig &cfg, const ConfigKey &setupSeqNum)
{
    bool writeOk = true;
    if (isConnected()) {
        bool force = ! status.conn.isReady();
        writeOk = writeSetup_internal(lastConfig, cfg, force);
        lastConfig = cfg;
        lastSetupSeqNum = setupSeqNum;
        if (writeOk)
            status.conn.setReady();
    } else {
        lastConfig = cfg;
        lastSetupSeqNum = setupSeqNum;
        status.conn.setDisconnected();
    }

    emit deviceConfigWritten(getDeviceIndex(), status.conn.isReady() ? lastSetupSeqNum : ConfigKey());
    if (status.conn.isReady()) {
        qDebug(lc).noquote() << QString("[%1] Wrote device setup seq #")
                                    .arg(getDeviceIndex().getSerialStr())
                             << setupSeqNum;
    } else {
        QStringList errors;
        if (!isConnected())
            errors += "not connected";
        if (!writeOk)
            errors += "write error";
        qWarning(lc).noquote() << QString("[%1] Failed to write device setup seq #")
                                      .arg(getDeviceIndex().getSerialStr())
                               << setupSeqNum
                               << ": " << errors.join("; ");
    }
}

RegOpVector TrigDevice::setControlStateCmd(bool trigOpen, bool softSpill)
{
    RegOpVector r;
    regCtrl.bit.run = trigOpen ? 1 : 0;
    regCtrl.bit.softspill = softSpill ? 1 : 0;
    r.RegWrite16(REG_RC_CTRL, regCtrl.all);
    return r;
}

RegOpVector TrigDevice::setTrigTimerFreqCmd(double freq)
{
    double periodInClocks = TrigDevice::getClockFreqHz() / freq;
    if (periodInClocks < 1) periodInClocks = 1;
    if (periodInClocks > quint32(-1)) periodInClocks = quint32(-1);
    quint32 regValue = periodInClocks;
    RegOpVector r;
    r.RegWrite32(REG32_RC_TRIG_PERIODICAL, regValue);
    return r;
}

RegOpVector TrigDevice::setTrigRandomFreqCmd(double freq)
{
    double period = 1.0 * freq * (quint32)(-1) / TrigDevice::getClockFreqHz();
    if (period < 1) period = 1;
    if (period > quint32(-1)) period = quint32(-1);
    quint32 regValue = period;
    RegOpVector r;
    r.RegWrite32(REG32_RC_TRIG_RANDOM, regValue);
    return r;
}

RegOpVector TrigDevice::setTrigRandomDeadTimeCmd(double seconds)
{
    double ticks = seconds * TrigDevice::getClockFreqHz();
    if (ticks < 0) ticks = 0;
    if (ticks > quint32(-1)) ticks = quint32(-1);
    quint32 regValue = ticks;
    RegOpVector r;
    r.RegWrite32(REG32_RC_TRIG_PDEAD, regValue);
    return r;
}

RegOpStream &operator>>(RegOpStream &in, TrigStatCounters &status)
{
    std::array<quint64, REG64_CNT_BLOCK_SIZE> xOffHiTimeoutCnt;
    std::array<quint64, REG64_CNT_BLOCK_SIZE> xOffLowTimeoutCnt;
    std::array<quint64, REG64_CNT_BLOCK_SIZE> xOffEventCnt;

    in >> xOffHiTimeoutCnt;
    in >> xOffLowTimeoutCnt;
    in >> xOffEventCnt;

    if (!in.isValid())
        return in;

    for (size_t i=0; i<xOffHiTimeoutCnt.size(); i++)
        status.xOffHiTimeoutCnt[i] = xOffHiTimeoutCnt[i];
    for (size_t i=0; i<xOffLowTimeoutCnt.size(); i++)
        status.xOffLowTimeoutCnt[i] = xOffLowTimeoutCnt[i];
    for (size_t i=0; i<xOffEventCnt.size(); i++)
        status.xOffEventCnt[i] = xOffEventCnt[i];

    return in;
}

mlink::RegOpVector TrigDevice::regOpGetCoreStatus(TrigCoreStatus *result) const
{
    RegOpVector r;
    r.RegRead64(REG64_RC_GLOB_EV_NUM, &result->globEvNum);
    r.RegRead16(REG_RC_TRIG_SRC, &result->trigSrc);
    r.RegRead16(REG_RC_TRIG_STATE, &result->curTrigState);
    r.RegRead16(REG_RC_XOFF_STATE, &result->curXOffState);
    r.RegRead16(REG_RC_XT_STATE, &result->curXOffTimeoutState);
    r.RegRead32(REG32_RC_SPILLPLOT_ADDR, &result->spillPlotAddr);
    return r;
}

RegOpStream &operator>>(RegOpStream &in, TrigCoreStatus &status)
{
    TrigCoreStatus s;
    in >> s.globEvNum;
    in >> s.trigSrc;
    in >> s.curTrigState;
    in >> s.curXOffState;
    in >> s.curXOffTimeoutState;
    in >> s.spillPlotAddr;
    if (in.isValid())
        status = s;
    return in;
}

std::optional<TrigCoreStatus> TrigDevice::readTrigCoreStatus()
{
    TrigCoreStatus result;
    auto r = regOpGetCoreStatus(&result);
    if (!regOpExec(r)) {
        qWarning(lc).noquote() << "Error reading TrigCoreStatus";
        return {};
    }
    return result;
}

mlink::RegOpVector TrigDevice::regOpGetWrStatus(TrigWrStatus *result) const
{
    RegOpVector r;
    r.RegRead16(REG_WR_STATUS, &result->Status.code.all);
    r.RegRead32(REG32_WR_UPTIME, &result->Uptime);
    r.RegRead64(REG64_WR_TAI_CODE, &result->curTai.code.all);
    r.RegRead32(REG32_WR_LINK_LOSS_COUNT, &result->errors.LinkLostCount);
    r.RegRead32(REG32_WR_SYNC_LOSS_COUNT, &result->errors.SyncLostCount);
    r.RegRead32(REG32_WR_CODE_ERROR_COUNT, &result->errors.CodeErrorCount);
    r.RegRead32(REG32_WR_OTHER_ERROR_COUNT, &result->errors.OtherErrorCount);
    r.RegRead64(REG64_WR_TAI_TS_1, &result->timestamps[0].code.all);
    r.RegRead64(REG64_WR_TAI_TS_2, &result->timestamps[1].code.all);
    r.RegRead64(REG64_WR_TAI_TS_3, &result->timestamps[2].code.all);
    r.RegRead64(REG64_WR_TAI_TS_4, &result->timestamps[3].code.all);
    return r;
}

RegOpStream &operator>>(RegOpStream &in, TrigWrStatus &status)
{
    TrigWrStatus wr;

    in >> wr.Status.code.all;
    in >> wr.Uptime;
    quint64 cur_tai_raw;
    in >> cur_tai_raw;
    wr.curTai = WrTaiCode(cur_tai_raw);
    in >> wr.errors.LinkLostCount;
    in >> wr.errors.SyncLostCount;
    in >> wr.errors.CodeErrorCount;
    in >> wr.errors.OtherErrorCount;
    std::array<quint64, 4> ts_raw;
    in >> ts_raw;
    for (int i=0; i<4; i++)
        wr.timestamps[i] = WrTaiCode(ts_raw.at(i));

    if (in.isValid())
        status = wr;
    return in;
}

std::optional<TrigWrStatus> TrigDevice::readTrigWrStatus()
{
    TrigWrStatus result;
    auto r = regOpGetWrStatus(&result);
    if (!regOpExec(r)) {
        qWarning(lc).noquote() << "Error reading TrigWrStatus";
        return {};
    }
    return result;
}

std::optional<RcCounters> TrigDevice::readRcCounters()
{
/*
        /// Selective readout
        QVector<quint16> timeCntIndex;
        for (quint16 j=0; j <= 13 && j <= fbp.nch; j++)
            timeCntIndex << j;
        timeCntIndex << 14 << 15;
        for(quint16 j : qAsConst(timeCntIndex)) {
            for (quint16 i=0; i<4; i++) {
                quint16 index = i*2*REG64_CNT_BLOCK_SIZE + j;
                v << RegRead64(REG64_CNT_BASE + 4*index);
            }
        }
        QVector<quint16> edgeCntIndex;
        edgeCntIndex << 0 << 14 << 15;
        for(quint16 i : qAsConst(edgeCntIndex)) {
            for (quint16 j=0; j<4; j++) {
                quint16 index = (j*2+1)*REG64_CNT_BLOCK_SIZE + i;
                v << RegRead64(REG64_CNT_BASE + 4*index);
            }
        }
*/
    RcCounters rcCounters;
    for (int bank = 0; bank < REG64_CNT_BLOCK_COUNT; bank++) {
        RegOpVector r;
        const auto numCounters = REG64_CNT_BLOCK_SIZE;
        std::array<quint64, numCounters> counters;
        for (quint16 i=0; i<numCounters; i++)
            r.RegRead64(REG64_CNT_BASE + 4 * (i + bank * REG64_CNT_BLOCK_SIZE), &counters[i]);
        if (!regOpExec(r)) {
            qWarning(lc).noquote() << "Error reading RcCounters bank" << bank;
            return {};
        }
        for (size_t i=0; i<counters.size(); i++) {
            rcCounters.cnt[bank][i] = counters[i];
        }
    }
    return rcCounters;
}

std::optional<TrigStatCounters> TrigDevice::readStatCounters()
{
    RegOpVector r;
    /*

        for(quint16 ch=0; ch<fbp.nch; ++ch) {
            v << RegRead64(REG64_CNT_XOFF_HI_BASE + 4*ch);
            v << RegRead64(REG64_CNT_XOFF_LOW_BASE + 4*ch);
            v << RegRead64(REG64_CNT_XOFF_EVENTS_BASE + 4*ch);
        }

    v << RegRead64(REG64_CNT_XOFF_EVENTS_BASE + 4*14);
    v << RegRead64(REG64_CNT_XOFF_EVENTS_BASE + 4*15);
    */

    const auto &rcCountersOpt = readRcCounters();
    if (!rcCountersOpt.has_value())
        return{};

    TrigStatCounters status;
    status.rc_counters = rcCountersOpt.value();

    // xOffHiTimeoutCnt
    for (quint16 i=0; i<REG64_CNT_BLOCK_SIZE; i++)
        r.RegRead64(REG64_CNT_XOFF_HI_BASE + 4 * i);

    // xOffLoTimeoutCnt
    for (quint16 i=0; i<REG64_CNT_BLOCK_SIZE; i++)
        r.RegRead64(REG64_CNT_XOFF_LOW_BASE + 4 * i);

    // xOffEventCnt (need only 14, 15)
    for (quint16 i=0; i<REG64_CNT_BLOCK_SIZE; i++)
        r.RegRead64(REG64_CNT_XOFF_EVENTS_BASE + 4 * i);

    if (!regOpExec(r)) {
        qWarning(lc).noquote() << "Error reading TrigStatCounters";
        return {};
    }
    return status;
    /*
            if (regOp.addr >= REG64_CNT_BASE && regOp.addr < REG64_CNT_BASE + 4 * REG64_CNT_TOTAL_NUM) {
                if(((regOp.addr-REG64_CNT_BASE) % 4) == 0) {
                    const int index =  (regOp.addr - REG64_CNT_BASE) / 4; // index of 64-bit register
                    const int i = index / REG64_CNT_BLOCK_SIZE; // ch-Number shift
                    const int j = index % REG64_CNT_BLOCK_SIZE; // counter type
                    status.rc_counters.cnt[i][j] |= get64(regOpIt, &ok);
                    if(ok) continue; else break;
                }
            }
            if (regOp.addr >= REG64_CNT_XOFF_HI_BASE && regOp.addr < REG64_CNT_XOFF_HI_BASE + 4 * REG64_CNT_BLOCK_SIZE) {
                int byte = regOp.addr % 4;
                int ch =  (regOp.addr - REG64_CNT_XOFF_HI_BASE) / 4;
                status.xOffHiTimeoutCnt[ch] |= (quint64(regOp.data)) << 16*byte;
            }
            if (regOp.addr >= REG64_CNT_XOFF_LOW_BASE && regOp.addr < REG64_CNT_XOFF_LOW_BASE + 4 * REG64_CNT_BLOCK_SIZE) {
                int byte = regOp.addr % 4;
                int ch =  (regOp.addr - REG64_CNT_XOFF_LOW_BASE) / 4;
                status.xOffLowTimeoutCnt[ch] |= (quint64(regOp.data)) << 16*byte;
            }
            if (regOp.addr >= REG64_CNT_XOFF_EVENTS_BASE && regOp.addr < REG64_CNT_XOFF_EVENTS_BASE + 4 * REG64_CNT_BLOCK_SIZE) {
                int byte = regOp.addr % 4;
                int ch =  (regOp.addr - REG64_CNT_XOFF_EVENTS_BASE) / 4;
                status.xOffEventCnt[ch] |= (quint64(regOp.data)) << 16*byte;
            }
    */
}

bool TrigDevice::updateCounters()
{
    RegOpVector r;
    r.RegWrite16(REG_CSR_CTRL, 0); // unlock counters
    r.RegWrite16(REG_CSR_CTRL, REG_CSR_CTRL_BIT_LOCK_COUNTERS); // lock counters
    if (!regOpExec(r)) {
        qWarning(lc).noquote() << "Error updating countes";
        return {};
    }
    return true;
}

bool TrigDevice::readTrigStatus()
{
    status.nch = fbp.nch;

    if (!updateCounters())
        return {};

    const auto &statCounters = readStatCounters();
    if (!statCounters.has_value())
        return {};
    status.statCounters = statCounters.value();

    const auto &coreStatus = readTrigCoreStatus();
    if (!coreStatus.has_value())
        return {};
    status.coreStatus = coreStatus.value();

    const auto &wrStatus = readTrigWrStatus();
    if (!wrStatus.has_value())
        return {};
    status.wr = wrStatus.value();

    status.histStatus.reset();
    if (histMod) {
        MscHistStatus histStatus;
        if (!histMod->readStatus(&histStatus))
            return {};
        status.histStatus = histStatus;
    }

    debugPrintStatus(status);

    const auto &spd = readSpillPlotRam();
    if (!spd.has_value())
        return {};
    const auto &thd = readTriggerTimeHistogramRam();
    if (!thd.has_value())
        return {};

    TrigSpillStatistics data{.status=status, .spd=spd.value(), .thd=thd.value()};
    analyzeDeviceStatistics(data);
    // tsw.writeStatisticsDump(data);
    emit updateStatus(status);
    return true;
}

void TrigDevice::debugPrintStatus(const TrigStatus &status)
{
    if (false) {
        qDebug(lc) << "--- trigStatusPoll ---";
        qDebug(lc) << "WR state: " <<  status.wr.Status.toString();
        qDebug(lc) << "WR Uptime: " <<  status.wr.Uptime;
        qDebug(lc) << "WR Link Loss  Count: " <<  status.wr.errors.LinkLostCount;
        qDebug(lc) << "WR Sync Loss  Count: " <<  status.wr.errors.SyncLostCount;
        qDebug(lc) << "WR Code Error Count: " <<  status.wr.errors.CodeErrorCount;
        qDebug(lc) << "WR Other Error Count: " <<  status.wr.errors.OtherErrorCount;
        qDebug(lc) << "WR now: " + status.wr.curTai.toString();
        if (status.spillStatus) {
            const auto spill = status.spillStatus.value();
            qDebug(lc) << "NCU: " + spill.lastSpillTai().toString();
            qDebug(lc) << "Spill state: " + spill.status.toString();
            qDebug(lc) << "Spill count: " << spill.count;
            qDebug(lc) << "Spill current time: " << spill.curSpillTimeMs;
            qDebug(lc) << "Spill TTL last period: " << spill.ttlLastPeriodMs;
            qDebug(lc) << "Spill WR last period: " << spill.wrLastPeriodMs;
        }
    }
    if (false) {
        QMap<int, QString> blockName;
        blockName[0] = "Time tot";
        blockName[1] = "Edge tot";
        blockName[2] = "Time run";
        blockName[3] = "Edge run";
        blockName[4] = "Time spl";
        blockName[5] = "Edge spl";
        blockName[6] = "Time r&s";
        blockName[7] = "Edge r&s";
        qInfo() << "--- counters ---";
        for (int i : status.statCounters.rc_counters.cnt.keys()) {
            QString str = blockName[i];
            for (int j : status.statCounters.rc_counters.cnt[i].keys())
//                    if (status.cnt[i][j] != 0)
//                    str += QString(" %1").arg(1.0 * status.cnt[i][j], 7, 'g', 2);
            str += QString(" %1").arg(status.statCounters.rc_counters.cnt[i][j], 16, 16);
            qInfo() << str;
        }
    }
    if (false) {
        for (int i=0; i<16; i++) {
            qInfo() << hex << i << status.statCounters.xOffEventCnt[i];
        }
    }

}

bool TrigDevice::readStatus()
{
    bool ok = ModularMregDevice::readStatus();
    return ok;
}

void TrigDevice::trigStatusPoll()
{
    if(!status.conn.isReady())
        resetAll();

    if (!isConnected())
        return;
    try {
        readTrigStatus();
    } catch (std::exception &e){
        qWarning(lc) << "Exception in readStatus";
    }
    trigStatusTimer->start(statusPollTimeMs);
}

void TrigDevice::analyzeDeviceStatistics(const TrigSpillStatistics &data)
{
    const TrigStatus &newStatus = data.status;
    const SpillPlotData &newSpd = data.spd;
    const TriggerHistogramData &newThd = data.thd;
    bool endOfReadout = false;
    if (trigSpillStatistics.status.spillStatus && newStatus.spillStatus) {
        auto newPhase = newStatus.spillStatus->status.reg.bit.phase;
        auto oldPhase = trigSpillStatistics.status.spillStatus->status.reg.bit.phase;
        // graphs are cleared on beginning of phase 2
        if (oldPhase == 4) {
            if (newPhase < 2 || newPhase == 5) {
                // flush existing graphs
                trigSpillStatistics.spd = newSpd;
                trigSpillStatistics.thd = newThd;
                endOfReadout = true;
            }
            if (newPhase == 4) {
                if (trigSpillStatistics.status.spillStatus->count != newStatus.spillStatus->count) {
                    endOfReadout = true; // spill incremented
                }
            } else {
                endOfReadout = true;
            }
        }
    }
    if (endOfReadout) {
        emit trigSpillStatisticsUpdated(trigSpillStatistics.status.getDeviceIndex(), trigSpillStatistics);
    }
    trigSpillStatistics.status = newStatus;
    trigSpillStatistics.spd = newSpd;
    trigSpillStatistics.thd = newThd;
}

void TrigDevice::evNumReset(const QVariant &startEvNum)
{
    quint64 v = startEvNum.isValid() ? startEvNum.toULongLong() : 0;
    legacyDevice->regWrite64(REG64_RC_GLOB_EV_NUM_LOAD, v-1);
}

std::optional<quint64> TrigDevice::readGlobEvNum()
{
    updateCounters();
    std::optional<RcCounters> cnts = readRcCounters();
    quint64 globEvNum;
    RegOpVector r;
    r.RegRead64(REG64_RC_GLOB_EV_NUM, &globEvNum);
    if (!regOpExec(r) || !cnts || !cnts->trigAccepted())
        return {};
    return globEvNum;
}

void TrigDevice::evNumRequest(int checkIndex, bool local)
{
    const auto &opt = readGlobEvNum();
    if (!opt.has_value()) {
        qWarning(lc) << "Error reading GlobEvNum";
        return;
    }
    auto globEvNum = opt.value();

    if (DEBUG_INFO)
        qInfo() << "TrigDevice::evNumRequest:" << checkIndex << globEvNum;

    if (local)
        emit topEvNumResponse(checkIndex, globEvNum);
    else
        emit evNumResponse(checkIndex, getDeviceIndex(), globEvNum);
}

std::optional<SpillPlotData> TrigDevice::readSpillPlotRam()
{
    if (!isConnected()) return {};
//    qDebug(lc) << regRead32(REG32_RC_SPILLPLOT_ADDR);
    if (!status.spillStatus)
        return {};
    const auto &spill = status.spillStatus.value();

    quint32 spillPlotAddr = status.coreStatus.spillPlotAddr;
    if(status.histStatus && !status.histStatus->chSt.isEmpty()) {
        const MscHistChStatus &chSt = status.histStatus->chSt[0];
        spillPlotAddr = chSt.plot_current_ram_address;
    }
    if(lastSpdInf && lastSpdInf->first == spill.count && lastSpdInf->second > spillPlotAddr) {
        // Skip incorrect spill hist reading; wait next
        return {};
    }
    lastSpdInf = {spill.count, spillPlotAddr};
    SpillPlotData shd;
    shd.spillCount = spill.count;
    shd.spillCurrentTimeMs = spill.curSpillTimeMs;
    shd.startOffset = 1.0e-3 * (lastConfig.spillControl.durPhase1);
    shd.stepSec = 1.0 * lastConfig.spillPlot.spillPlotStep / getClockFreqHz();
    quint32 readSize = std::min(spillPlotAddr, (quint32)spillPlotSize());
    std::vector<quint32> ram;
    if (!memReadBlk(fbp.spill_plot_ram_start_addr, readSize, &ram))
        return {};
    shd.data = QVector<quint32>(ram.cbegin(), ram.cend());
    emit spillPlotUpdated(shd);
    return shd;
}

std::optional<TriggerHistogramData> TrigDevice::readTriggerTimeHistogramRam()
{
    if (!isConnected()) return {};
    if (!status.spillStatus)
        return {};
    const auto &spill = status.spillStatus.value();
    TriggerHistogramData thd;
    thd.spillCount = spill.count;
    thd.spillCurrentTimeMs = spill.curSpillTimeMs;
    quint32 histSize = triggerHistogramSize();
    int hist_channels = fbp.multihist ? 16 : 1;
    for (int ch=0; ch<hist_channels; ch++) {
        if (ch >= fbp.nch && ch < 15) // unused channels, except last
            continue;
        bool xOff = lastConfig.trig.xOffMask&(1<<ch);
        if (ch < fbp.nch && !xOff) // disabled channel
            continue;
        std::vector<quint32> ram;
        quint32 startAddr = fbp.trig_hist_ram_start_addr + ch * histSize;
        if (!memReadBlk(startAddr, histSize, &ram) || ram.size() != histSize) {
            thd.histograms.remove(ch);
            continue;
        }
        TriggerHistogramChannelData thcd;
        thcd.data = QVector<quint32>(ram.cbegin(), ram.cend());

        thcd.spillCount = thd.spillCount; // alias
        thcd.spillCurrentTimeMs = thd.spillCurrentTimeMs; // alias
        thcd.channel = ch;
        QString descr = lastConfig.channelDescr[ch];
        if (descr.isEmpty())
            descr = QString::number(ch);
        if (ch == 15)
            descr = "Total";
        thcd.channelDescr = descr;
        thcd.overflowCount = thcd.data.last();
        thcd.data.pop_back();
        thcd.px.resize(thcd.data.size());
        if (fbp.tih_log) {
            const double tickSec = 1.0 / getClockFreqHz();
            for (int i=0; i<thcd.data.size(); i++) {
                const double scaler_2 = 54.0; // FIXME
                double x = exp(i/scaler_2) * tickSec * 1e6; // μs
                thcd.px[i] = x;// i*stepSec;
            }
        } else {
            const double stepSec = 1.0 * 1e-6; // lastConfig.spill.trigTimeStep / getClockFreqHz();
            for (int i=0; i<thcd.data.size(); i++) {
                double x = i * stepSec * 1e6; // μs
                thcd.px[i] = x;
            }
            thcd.px[0] = thcd.px[1]; // log scale workaround
        }
        thd.histograms[ch] = thcd;
    }
    emit triggerHistogramUpdated(thd);
    return thd;
}

quint8 TrigDevice::getTrigSrcCount() const
{
    quint8 trigSrcCnt = 1;
    switch (getDeviceId()) {
    case DEVICE_ID_U40VE_RC:
        trigSrcCnt = 1;
        break;
    case DEVICE_ID_UT24VE:
        trigSrcCnt = 4;
        break;
    default:
        trigSrcCnt = 1;
    }
    return trigSrcCnt;
}

void TrigDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();
    trigStatusTimer->start(0);
    status.conn.setConnected();

    const auto &opResult = readFirmwareBuildParams();
    if (opResult.has_value()) {
        fbp = opResult.value();
    }
    quint16 tmp;
    legacyDevice->regRead16(REG_RC_CHANNEL_CNT, &tmp);
    if (fbp.nch == 0 && tmp <= 16)
        fbp.nch = tmp;
    emit firmwareBuildParamsChanged(fbp);
    emit trigSrcCntUpdated(getTrigSrcCount());
    if(!status.conn.isReady())
        resetAll();
}

void TrigDevice::on_deviceDisconnected()
{
    ModularMregDevice::on_deviceDisconnected();
    status.setDisconnected();
    emit updateStatus(status);
    emit deviceConfigWritten(getDeviceIndex(), ConfigKey());
}

void TrigDevice::setStatusPollTimeMs(int ms)
{
    if (ms < TRIG_STATUS_POLL_TIME_MS_MIN)
        ms = TRIG_STATUS_POLL_TIME_MS_MIN;
    statusPollTimeMs = ms;
}
