/*
**    Copyright 2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigHistView.h"
#include "ui_TrigHistView.h"

#include "FirmwareBuildParams.h"
#include "RcConfig.h"
#include "RcConfigDispatcher.h"
#include "TrigDevice.h"
#include "base/gui_utils.h"
#include "rc_gui_utils.h"

TrigHistView::TrigHistView(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::TrigHistView),
    config(new RcConfig)
{
    ui->setupUi(this);

    connect(this, &TrigHistView::spillPlotUpdated,
            ui->spillPlot, &SpillPlot::updateSpillPlot);
    connect(this, &TrigHistView::triggerHistogramUpdated,
            ui->triggerHistogram, &TriggerHistogram::updateHistogram);
    connect(this, &TrigHistView::plotChUpdated,
            ui->triggerHistogram, &TriggerHistogram::setPlotCh);

    initHistWidget();
}

TrigHistView::~TrigHistView()
{
    delete ui;
}

void TrigHistView::setConfigDispatcher(RcConfigDispatcher *p)
{
    if (configDispatcher) {
        disconnect(configDispatcher);
    }
    configDispatcher = p;
    if (! configDispatcher)
        return;
    configDispatcher->subscribe(this);
}

void TrigHistView::setConfig(const RcConfig &newConfig)
{
    *config = newConfig;
    restoreGuiFromConfig();
}

void TrigHistView::initHistWidget()
{
    int i = 0;
    ui->comboBoxSpillStep->insertItem(i++, "100 ms", 100e-3);
    ui->comboBoxSpillStep->insertItem(i++, "50 ms", 50e-3);
    ui->comboBoxSpillStep->insertItem(i++, "20 ms", 20e-3);
    ui->comboBoxSpillStep->insertItem(i++, "10 ms", 10e-3);
    ui->comboBoxSpillStep->insertItem(i++, "5 ms", 5e-3);
    ui->comboBoxSpillStep->insertItem(i++, "2 ms", 2e-3);
    ui->comboBoxSpillStep->insertItem(i++, "1 ms", 1e-3);
    ui->comboBoxSpillStep->insertItem(i++, "500 us", 500e-6);
    ui->comboBoxSpillStep->insertItem(i++, "200 us", 200e-6);
    ui->comboBoxSpillStep->insertItem(i++, "100 us", 100e-6);
    ui->comboBoxSpillStep->insertItem(i++, "50 us", 50e-6);
    ui->comboBoxSpillStep->insertItem(i++, "20 us", 20e-6);
    ui->comboBoxSpillStep->insertItem(i++, "10 us", 10e-6);
//    connect(ui->comboBoxSpillStep, SIGNAL(activated(int)),
//            this, SLOT(on_comboBoxSpillStep_activated(int)));

    ui->comboBoxHistSelect->insertItem(0, "All channels", -1);
    for (int i=0; i<16; i++)
        ui->comboBoxHistSelect->insertItem(i+1, QString("%1").arg(i), i);
}

void TrigHistView::restoreGuiFromConfig()
{
    MultiSignalBlocker msb(children());
    {
        auto value = config->rcTrigConfig.spillPlot.spillPlotStep / TrigDevice::getClockFreqHz();
        auto index = findComboBoxNearestItem(*ui->comboBoxSpillStep, value);
        QSignalBlocker(ui->comboBoxSpillStep);
        ui->comboBoxSpillStep->setCurrentIndex(index);
    }
    {
        const QSettings settings;
        bool ok = false;
        auto value = settings.value("timeIntervalHistSelect").toInt(&ok);
        if (ok) {
            auto index = findComboBoxNearestItem(*ui->comboBoxHistSelect, value);
            QSignalBlocker(ui->comboBoxHistSelect);
            ui->comboBoxHistSelect->setCurrentIndex(index);
            emit plotChUpdated(value);
        }
    }
    {
        auto index = ui->comboBoxPlotType->findData(config->rcTrigConfig.spillPlot.plotSrc);
        if(index != -1)
            ui->comboBoxPlotType->setCurrentIndex(index);
    }
    // updatePlotTypeList(0);
}

void TrigHistView::on_comboBoxSpillStep_activated(int index)
{
//    config->rcTrigConfig.spill.spillPlotStep = index;
    bool ok = false;
    double value = ui->comboBoxSpillStep->itemData(index).toDouble(&ok);
    if (!ok)
        return;
    config->rcTrigConfig.spillPlot.spillPlotStep = value * TrigDevice::getClockFreqHz();
    emit configChanged(*config);
}

void TrigHistView::on_comboBoxHistSelect_activated(int index)
{
//    config->rcTrigConfig.spill.trigTimeStep = index;
    bool ok = false;
    double value = ui->comboBoxHistSelect->itemData(index).toDouble(&ok);
    if (!ok)
        return;
    emit plotChUpdated(value);
    QSettings settings;
    settings.setValue("timeIntervalHistSelect", value);
    emit configChanged(*config);
}

void TrigHistView::setFirmwareBuildParams(const FirmwareBuildParams &fbp)
{
    updatePlotTypeList(fbp.ntrig);
}

void TrigHistView::updatePlotTypeList(quint8 trigSrcCnt)
{
    MultiSignalBlocker msb(children());

    QComboBox * const cb = ui->comboBoxPlotType;
    if(cb->count() == (trigSrcCnt+2)) return;
    cb->clear();
    cb->setEnabled(trigSrcCnt);

    if(trigSrcCnt==0) {
        cb->insertItem(0, "External TTL", 0);
    } else {
        for(int i=0; i<trigSrcCnt; ++i)
            cb->insertItem(i, QString("External TTL %1").arg(i+1), i);
    }

    cb->insertItem(14, "Available", 14);
    cb->insertItem(15, "Accepted", 15);

    auto index = cb->findData(config->rcTrigConfig.spillPlot.plotSrc);
    if(index != -1)
        cb->setCurrentIndex(index);
}

void TrigHistView::on_comboBoxPlotType_currentIndexChanged(int)
{
    bool ok;
    quint8 srcIndex = ui->comboBoxPlotType->currentData().toUInt(&ok);
    if(ok) {
        config->rcTrigConfig.spillPlot.plotSrc = srcIndex;
        emit configChanged(*config);
    }
}

