#include "TrcSignalRouter.h"

#include <QObject>

#include "RcCountersModel.h"
#include "TrcConfigDispatcher.h"
#include "TrcCore.h"
#include "TrcDevice.h"
#include "TrcDeviceGlobal.h"
#include "base/DeviceId.h"
#include "configuration-manager/ConfigurationManager.h"
#include "fsm/AdcDeviceMgrFsm.h"
#include "mldiscover/DeviceDispatcher.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "pnp-server/PnpDiscovery.h"
#include "remote-control-server/RemoteControlServer.h"


#ifdef QT_WIDGETS_LIB
#include "device-discover/DiscoverDialog.h"
#include "configuration-manager/ConfigurationManager.h"
#include "m-stream-output/MStreamStatisticDialog.h"

#include "TrcMainWindow.h"
#include "TrcDeviceChModel.h"
#endif

static constexpr const auto device_filter_list = {
    DEVICE_ID_TRC
};

TrcSignalRouter::TrcSignalRouter()
{
    setDeviceIdFilter(device_filter_list);
    setConfigDispatcher(&TrcConfigDispatcher::instance());
}

void TrcSignalRouter::setCore(TrcCore *p)
{
    core = p;
    setReadoutAppCore(p);
}

void TrcSignalRouter::connect_ConfigurationDispatcher() const
{
    ReadoutAppSignalRouter::connect_ConfigurationDispatcher();

    auto trcConfigDispatcher = &TrcConfigDispatcher::instance();

    if (deviceDispatcher)
        deviceDispatcher->subscribeDevices(trcConfigDispatcher);

    if(core) {
        trcConfigDispatcher->subscribeConfig(core);
        trcConfigDispatcher->subscribeCS(core);
    }
    if(configDb) {
        trcConfigDispatcher->subscribeConfig(configDb);
        trcConfigDispatcher->subscribeCS(configDb);
    }
    if(configurationManager)
        trcConfigDispatcher->subscribeCS(configurationManager);

#ifdef QT_WIDGETS_LIB
    for(auto m : rcCountersModels) {
        connect(trcConfigDispatcher, &TrcConfigDispatcher::configChanged,
                m, &RcCountersModel::setConfig);
    }
#endif
}

#ifdef QT_WIDGETS_LIB
void TrcSignalRouter::setMainWindow(TrcMainWindow *p)
{
    // set mainWindow must be before core
    mainWindow = p;
    setReadoutAppMainWindow(p);
}

void TrcSignalRouter::connect_MainWindow() const
{
    ReadoutAppSignalRouter::connect_MainWindow();

    if(!core)
        return;

    if(!mainWindow)
        return;

    LldpDiscovery::subscribe(mainWindow, device_filter_list);

    connect(mainWindow, &TrcMainWindow::performLocalStart,
            core, &TrcCore::performLocalStart);

    connect(mainWindow, &TrcMainWindow::performLocalStop,
            core, &TrcCore::performLocalStop);

    LldpDiscovery::subscribe(discoverDialog, device_filter_list);
    deviceDispatcher->subscribeDevices(discoverDialog);

    Q_ASSERT(deviceMgrFsm);
    connect(deviceMgrFsm, &BaseFsm::stateChanged,
            mainWindow, &TrcMainWindow::onFsmStateChange);

    connect(mainWindow, &TrcMainWindow::printFsmInputs,
            deviceMgrFsm, &BaseFsm::printFsmInput);
}
#endif


void TrcSignalRouter::connect_RemoteControlServer() const
{
    ReadoutAppSignalRouter::connect_RemoteControlServer();

    if(!remoteControlServer)
        return;

    if (core) {
        connect(remoteControlServer, &RemoteControlServer::startRequest,
                core, &TrcCore::remoteStartRequest);
        connect(remoteControlServer, &RemoteControlServer::stopRequest,
                core, &TrcCore::remoteStopRequest);
        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                core, &TrcCore::onRemoteStatusUpdated);

        connect(core, &TrcCore::sendRunError,
                remoteControlServer,
                QOverload<const RcRunId &, QString>::of(&RemoteControlServer::onRunError));

        connect(core, &TrcCore::sendRunWarning,
                remoteControlServer, &RemoteControlServer::onRunWarning);
    }
}

void TrcSignalRouter::connect_FSM() const
{
    ReadoutAppSignalRouter::connect_FSM();

    if(!deviceMgrFsm)
        return;

    if(remoteControlServer){
        connect(deviceMgrFsm, &BaseFsm::stateChanged,
                remoteControlServer, &RemoteControlServer::onFsmStateChange);
    }

    connect(&TrcConfigDispatcher::instance(), &TrcConfigDispatcher::configKeyChanged,
            deviceMgrFsm, &AdcDeviceMgrFsm::setMainSetupKey);
}

void TrcSignalRouter::connect_TrcDeviceGlobal(const TrcDeviceGlobal *devGlob)
{
    if(!devGlob)
        return;

    auto dev = devGlob->getDevice<TrcDevice>();

    connect(dev, &TrcDevice::statusUpdated,
            core, &TrcCore::deviceStatusUpdated);

    connect(dev, &TrcDevice::statusUpdated,
            devGlob, &TrcDeviceGlobal::deviceStatusUpdated);

    connect(core, &TrcCore::preRunReset,
            dev, &TrcDevice::preRunReset);

    connect(dev, &TrcDevice::deviceDisconnected,
            core, &TrcCore::deviceDisconnected);

    connect(core, &TrcCore::mstreamResetAll,
            devGlob, &TrcDeviceGlobal::resetMStream);

#ifdef QT_WIDGETS_LIB
    if(mainWindow){
        connect(dev, &TrcDevice::statusUpdated,
                mainWindow, &TrcMainWindow::deviceStatusUpdated);

        connect(mainWindow, &TrcMainWindow::softReset,
                dev, &TrcDevice::resetAll);

        connect(mainWindow, &TrcMainWindow::softClear,
                dev, &TrcDevice::softClear);

        connect(dev, &TrcDevice::deviceConnected,
                mainWindow, &TrcMainWindow::deviceConnected);

        connect(dev, &TrcDevice::deviceDisconnected,
                mainWindow, &TrcMainWindow::deviceDisconnected);

        connect(devGlob, &TrcDeviceGlobal::stat_updated,
                mainWindow, &TrcMainWindow::update_mstream_data_rate);

        connect(mainWindow, &TrcMainWindow::mstream_reset_all,
                devGlob, &TrcDeviceGlobal::resetMStream);
    }
    if(mStreamStatisticDialog) {
        connect(devGlob, &TrcDeviceGlobal::stat_updated,
                mStreamStatisticDialog, &MStreamStatisticDialog::stat_updated);
        connect(devGlob, &TrcDeviceGlobal::stat_from_receiver_updated,
                mStreamStatisticDialog, &MStreamStatisticDialog::stat_from_receiver_updated);
    }
    for(auto m : rcCountersModels) {
        connect(dev, &TrcDevice::countersUpdated,
                m, &RcCountersModel::deviceStatusUpdated);
    }
    if(deviceChModel) {
        connect(dev, &TrcDevice::countersUpdated,
                deviceChModel, &TrcDeviceChModel::deviceStatusUpdated);
    }
#endif

    connect(devGlob, &TrcDeviceGlobal::closeProgram,
            core, &TrcCore::closeProgram);

    PnpDiscovery::subscribeMyProgram(devGlob);

    deviceDispatcher->subscribeHostAddressUpdates(devGlob->getDeviceIndex(),
                                                  dev);
    auto trcConfigDispatcher = &TrcConfigDispatcher::instance();
    trcConfigDispatcher->forwardDeviceConfig(dev);

    connect(dev, &TrcDevice::deviceInfoUpdated,
            trcConfigDispatcher, &TrcConfigDispatcher::deviceInfoUpdated);
}
