/*
**    Copyright 2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ThreadLatencyMonitorPrivate.h"

#include <QDebug>
#include <QLoggingCategory>
#include <QThread>
#include <QTimer>

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_threads_latency)
Q_LOGGING_CATEGORY(lc_threads_latency, "afi.threads.latency")
}

ThreadLatencyMonitorPrivate::ThreadLatencyMonitorPrivate(QObject *parent)
    : QObject(parent)
{
    QTimer *timer1 = new QTimer(this);
    connect(timer1, &QTimer::timeout, this, &ThreadLatencyMonitorPrivate::checkLoopLatency);
    timer1->start(10);

    QTimer *timer2 = new QTimer(this);
    connect(timer2, &QTimer::timeout, this, &ThreadLatencyMonitorPrivate::periodicJob);
    timer2->start(5000);
}

void ThreadLatencyMonitorPrivate::checkLoopLatency()
{
    QDateTime now = QDateTime::currentDateTime();
    if (loopTime.isValid()) {
        double dt = 0.001 * loopTime.msecsTo(now);
        if (dt > 0.2) {
            //            QString str = QString("%1 thread latency %2").arg(thread()->name()).arg(dt, 0, 'f', 3);
            //            qDebug() << str;
        }
        if (values.size() < 1000)
            values.push_back(dt);
    }
    loopTime = now;
}

void ThreadLatencyMonitorPrivate::reset()
{
    values.clear();
}

void ThreadLatencyMonitorPrivate::periodicJob()
{
    printStatistics();
    reset();
}

void ThreadLatencyMonitorPrivate::printStatistics() const
{
    if (values.isEmpty())
        return;
    double vMean = 0;
    auto vMin = values[0];
    auto vMax = values[0];
    for (const auto &v : qAsConst(values)) {
        vMean += v;
        if (v > vMax)
            vMax = v;
        if (v < vMin)
            vMin = v;
    }
    vMean /= values.size();
    QString str = QString("%1 thread latency: min %2, mean %3, max %4")
                      .arg(thread()->objectName())
                      .arg(vMin, 0, 'f', 3)
                      .arg(vMean, 0, 'f', 3)
                      .arg(vMax, 0, 'f', 3);
    qDebug(lc_threads_latency).noquote() << str;
}
