/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "BoardGpioModule.h"

#include "BoardGpio/BoardGpioConfig.h"
#include "BoardGpioStatus.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {
enum {
    REG32_ID = 0x0,
    REG_LOCK  = 0x3,
    REG_SIZE  = 0x4,
    REG_CAPS_TRI = 0x40,
    REG_CAPS_IN  = 0x50,
    REG_CAPS_OUT = 0x60,
    REG_TRI = 0x80,
    REG_IN  = 0x90,
    REG_OUT = 0xA0,
};
const quint16 REG_LOCK_BIT_LOCK = 1;
} // namespace

BoardGpioModule::BoardGpioModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
}

QList<AFI_SDB_ID> BoardGpioModule::getSupportedSdbIds()
{
    return {AFI_SDB_BOARD_GPIO};
}

std::optional<BoardGpioStatus> BoardGpioModule::readStatus()
{
    BoardGpioStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}

bool BoardGpioModule::readBuildParams()
{
    RegOpVector r;
    r.RegRead32(REG32_ID, &status.id);
    r.RegRead16(REG_SIZE, &status.size);
    if (!regOpExecRebased(r))
        return {};
    return true;
}

bool BoardGpioModule::readStatus(BoardGpioStatus *result)
{
    if (!status.isValid())
        if (!readBuildParams())
            return {};
    if (!status.isValid())
        return {};
    RegOpVector r;
    r.RegWrite16(REG_LOCK, 0);
    r.RegWrite16(REG_LOCK, REG_LOCK_BIT_LOCK);

    int wordcount = (status.size + 15) / 16;
//    if (dumb_read) {
//        r.RegRead64(REG_CAPS_TRI, &status.caps_tri);
//        r.RegRead64(REG_CAPS_IN, &status.caps_in);
//        r.RegRead64(REG_CAPS_OUT, &status.caps_out);
//        r.RegRead64(REG_TRI, &status.tri);
//        r.RegRead64(REG_IN, &status.in);
//        r.RegRead64(REG_OUT, &status.out);
//    }
    std::vector<quint16> caps_tri(wordcount);
    std::vector<quint16> caps_in(wordcount);
    std::vector<quint16> caps_out(wordcount);
    std::vector<quint16> tri(wordcount);
    std::vector<quint16> in(wordcount);
    std::vector<quint16> out(wordcount);
    for (int i=0; i<wordcount; i++) {
        r.RegRead16(REG_CAPS_TRI + i, &caps_tri[i]);
        r.RegRead16(REG_CAPS_IN + i, &caps_in[i]);
        r.RegRead16(REG_CAPS_OUT + i, &caps_out[i]);
        r.RegRead16(REG_TRI + i, &tri[i]);
        r.RegRead16(REG_IN + i, &in[i]);
        r.RegRead16(REG_OUT + i, &out[i]);
    }

    r.RegWrite16(REG_LOCK, 0);
    if (!regOpExecRebased(r))
        return {};
    status.caps_tri = 0;
    status.caps_in = 0;
    status.caps_out = 0;
    status.tri = 0;
    status.in = 0;
    status.out = 0;
    for (int i=0; i<wordcount; i++) {
        status.caps_tri |= (quint64)(caps_tri[i]) << (i*16);
        status.caps_in |= (quint64)(caps_in[i]) << (i*16);
        status.caps_out |= (quint64)(caps_out[i]) << (i*16);
        status.tri |= (quint64)(tri[i]) << (i*16);
        status.in |= (quint64)(in[i]) << (i*16);
        status.out |= (quint64)(out[i]) << (i*16);
    }

    if (result)
        *result = status;
    return true;
}

bool BoardGpioModule::writeConfig(const BoardGpioConfig &newConfig)
{
    if (!status.isValid())
        if (!readBuildParams())
            return {};
    if (!status.isValid())
        return {};
    int wordcount = (status.size + 15) / 16;
    std::vector<quint16> tri(wordcount);
    std::vector<quint16> out(wordcount);
    for (auto it = newConfig.channels.begin(); it != newConfig.channels.end(); ++it) {
        auto ch = it.key();
        auto channelConfig = it.value();
        tri[ch/16] |= channelConfig.tri ? (1 << (ch % 16)) : 0;
        out[ch/16] |= channelConfig.out ? (1 << (ch % 16)) : 0;
    }
    RegOpVector r;
    r.RegWrite16(REG_LOCK, REG_LOCK_BIT_LOCK);
    for (int i=0; i<wordcount; i++) {
        r.RegWrite16(REG_TRI + i, tri[i]);
        r.RegWrite16(REG_OUT + i, out[i]);
    }
    r.RegWrite16(REG_LOCK, 0);
    return regOpExecRebased(r);
}
