#include "MscCycleCountersModule.h"

#include "MscCycleStatus.h"

#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

enum {
    REG_CTRL                = 0x000,
    REG_HIST_COUNT          = 0x001, // Reg 0x1 - Number of channels, RO
    REG64_CYCLE_STATE       = 0x004, // Reg 0x4 - Current state of cycles, 64-bits, RO
    REG_HIST_CYCLE_BASE     = 0x010, // Regs 0x10~0x1F - Histogram #0 registers group
    REG_HIST_CYCLE_SHIFT    = 0x010, // Regs 0x10~0x1F - Histogram #0 registers group
};

enum {
    REG_CTRL_BIT_RESET   = 1 << 0,
    REG_CTRL_BIT_LOCK    = 1 << 1,
};

static const int MAX_HIST_COUNT = 64;

QList<AFI_SDB_ID> MscCycleCountersModule::getSupportedSdbIds()
{
    return {AFI_SDB_MSC_CYCLE_COUNTERS};
}

void MscCycleCountersModule::post_init()
{
    BaseDeviceModule::post_init();
    regRead16(REG_HIST_COUNT, &numOfHists);
    // limit munber of hists to available address space
    auto regCount = lastAddr() - baseAddr() + 1;
    auto limit = (regCount - REG_HIST_CYCLE_BASE) / REG_HIST_CYCLE_SHIFT;
    std::clamp(limit, 0, MAX_HIST_COUNT);
    if (numOfHists > limit)
        numOfHists = limit;

    prevGateStart.resize(numOfHists);
}

RegOpVector MscCycleCountersModule::getStatusRegOp(MscCycleStatus *result) const
{
    RegOpVector r;
    r.RegWrite16(REG_CTRL, REG_CTRL_BIT_LOCK); // lock counters
    r.RegRead64(REG64_CYCLE_STATE, &result->curGateState);
    result->gates.resize(numOfHists);
    for (quint16 g = 0; g < numOfHists; ++g) {
        auto &gate = result->gates[g];
        quint16 base = REG_HIST_CYCLE_BASE + REG_HIST_CYCLE_SHIFT * g;
        r.RegRead32(base + 0, &gate.activeGateCnt);
        r.RegRead32(base + 2, &gate.fullGateCnt);
        r.RegRead64(base + 4, &gate.lastGateStartRaw);
        r.RegRead64(base + 8, &gate.lastGateEndRaw);
        r.RegRead32(base + 12, &gate.gateCnt);
    }
    r.RegWrite16(REG_CTRL, 0); // unlock counters
    return r;
}

bool MscCycleCountersModule::readStatus(MscCycleStatus *result)
{
    prevGateStart.resize(numOfHists);

    auto r = getStatusRegOp(&curStatus);
    if (!regOpExecRebased(r))
        return {};

    // update s/w part of status
    for (quint16 i = 0; i < numOfHists; ++i) {
        auto &g = curStatus.gates[i];
        auto &prevStart = prevGateStart[i];
        if(prevStart.gateCnt+1 == g.gateCnt)
            g.lastGatePeriod = g.lastGateStart().toNs() - prevStart.gateStart.toNs();
        else if(prevStart.gateCnt != g.gateCnt)
            g.lastGatePeriod.reset();

        if(!curStatus.isInGate(i)) {
            g.lastGateWidth = g.lastGateEnd().toNs() - g.lastGateStart().toNs();
        } else {
            if(prevStart.gateCnt+1 == g.gateCnt)
                g.lastGateWidth = g.lastGateEnd().toNs() - prevStart.gateStart.toNs();
            else if(prevStart.gateCnt != g.gateCnt)
                g.lastGateWidth.reset();
        }
        prevStart.gateCnt = g.gateCnt;
        prevStart.gateStart = g.lastGateStart();
    }
    if (result)
        *result = curStatus;
    return true;
}

std::optional<MscCycleStatus> MscCycleCountersModule::readStatus()
{
    MscCycleStatus tmp;
    if (!readStatus(&tmp))
        return {};
    return tmp;
}
