#include "Tlu40LvdsTofTriggerModule.h"

#include "DeviceModuleCommand.h"
#include "base/DeviceIndex.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {
enum {
    REG32_ID        = 0x00, // Identification cef86b29, 32 bits, RO static
    REG_CTRL        = 0x02, // Control, R/W // [0] - softreset
    REG_LOCK        = 0x03, // Lock, R/W // [0] - 1: lock, 0: continuous update
    REG_CH_NUM      = 0x04, // Number of channels for one TOF module, RO static
    REG_TRIG_CH_NUM = 0x05, // Number of trigger channels for one TOF module, RO static
    REG_NUM_TOF_MOD = 0x06, // Number of TOF modules, RO static
    REG_CLK_FREQ    = 0x07, // Clock frequency, kHz, RO static

    REG_MOD_BASE    = 0x10, // Base module register region
    REG_MOD_SHIFT   = 0x10, // Base module register region

    REG32_CH_INPUT_CNT_BASE     = 0x100, //100~13Fh - 1st TOF module input counters, 32 bits per counter, RO
    REG32_CH_INPUT_CNT_SHIFT    = 0x040, //140~17Fh - 2nd TOF module input counters, 32 bits per counter, RO
};

enum {
    REG64_MOD_TRIG_CH_EN = 0x00, // - 1st TOF module trigger channels enable, 64 bits, R/W
    REG_MOD_SHAPER_VAL = 0x04, // - Shaper value for 1st TOF module signals, R/W
    REG64_MOD_CUR_INPUT_STATE = 0x08, // - 1st TOF module channels current input state, 64 bits, RO
};
}

QList<AFI_SDB_ID> Tlu40LvdsTofTriggerModule::getSupportedSdbIds()
{
    return {AFI_SDB_TLU40LVDS_TOF_Trigger};
}

bool Tlu40LvdsTofTriggerModule::writeConfig(const Tlu40LvdsTofTriggerConfig &config)
{
    if(!buildParams)
        return {};

    RegOpVector r;
    for(auto mod=0; mod<buildParams->modCnt; ++mod) {
        auto modCfg = config.modConfig.value(mod);
        quint16 base = REG_MOD_BASE + mod*REG_MOD_SHIFT;
        quint64 chEn = 0;
        for(auto ch=0; ch<buildParams->chCntPerMod; ++ch)
            if(modCfg.chEn.value(ch, false))
                chEn |= 1<<ch;

        r.RegWrite16(base+REG_MOD_SHAPER_VAL, modCfg.shaperVal);
        r.RegWrite64(base+REG64_MOD_TRIG_CH_EN, chEn);
    }
    return regOpExecRebased(r);
}

std::optional<Tlu40LvdsTofTriggerStatus> Tlu40LvdsTofTriggerModule::readStatus()
{
    if(!buildParams)
        return {};

    Tlu40LvdsTofTriggerStatus st;
    st.modSt.resize(buildParams->modCnt);

    RegOpVector r;
    r.RegWrite16(REG_LOCK, 0x1);
    for(auto mod=0; mod<buildParams->modCnt; ++mod) {
        quint16 base = REG_MOD_BASE + mod*REG_MOD_SHIFT;
        quint16 cntBase = REG32_CH_INPUT_CNT_BASE + mod*REG32_CH_INPUT_CNT_SHIFT;

        auto &modSt = st.modSt[mod];
        modSt.chInputCnt.resize(buildParams->chCntPerMod);

        r.RegRead64(base+REG64_MOD_CUR_INPUT_STATE, &modSt.inputState);
        for(auto ch=0; ch<buildParams->chCntPerMod; ++ch)
            r.RegRead32(cntBase+ch*2, &modSt.chInputCnt[ch]);
    }
    r.RegWrite16(REG_LOCK, 0x0);

    if(!regOpExecRebased(r))
        return {};
    return st;
}

void Tlu40LvdsTofTriggerModule::post_init()
{
    buildParams.reset();
    quint32 id;
    Tlu40LvdsTofTriggerBuildParams bp;

    RegOpVector r;
    r.RegRead32(REG32_ID, &id);
    r.RegRead16(REG_NUM_TOF_MOD, &bp.modCnt);
    r.RegRead32(REG_CLK_FREQ, &bp.clkFreq);
    r.RegRead16(REG_CH_NUM, &bp.chCntPerMod);
    r.RegRead16(REG_TRIG_CH_NUM, &bp.trigChCntPerMod);

    if(regOpExecRebased(r) && getSupportedSdbIds().contains(static_cast<AFI_SDB_ID>(id))) {
        if(bp.trigChCntPerMod*2 == bp.chCntPerMod) {
            buildParams = bp;
        } else {
            qCritical("[%s] '%s': failed initialize trigChCntPerMod(%d)*2 != chCntPerMod(%d)",
                      getDeviceIndex().getIdent().toStdString().c_str(),
                      name().toStdString().c_str(),
                      bp.trigChCntPerMod, bp.chCntPerMod);
        }
    }
}
