﻿#include "FeLinkTreeView.h"

#include <QHeaderView>
#include <QScrollBar>
#include <QSettings>

#include "FeLinkTreeColumns.h"
#include "FeLinkTreeModel.h"
#include "PhyResetButtonDelegate.h"

const QString EXPAND_DEVICE_MAP = "expand_device_map";

FeLinkTreeView::FeLinkTreeView(QWidget *parent)
    : QTreeView(parent)
    , phyResetPbDelegate(new PhyResetButtonDelegate(this))
{
    connect(this, &QTreeView::expanded, this, &FeLinkTreeView::expanded);
    connect(this, &QTreeView::collapsed, this, &FeLinkTreeView::collapsed);
    header()->setStretchLastSection(false);
    header()->setSectionResizeMode(QHeaderView::ResizeToContents);

    auto sp = sizePolicy();
    sp.setVerticalPolicy(QSizePolicy::Maximum);
    setSizePolicy(sp);
}

void FeLinkTreeView::setModel(QAbstractItemModel *model)
{
    auto feLinkModel = qobject_cast<const FeLinkTreeModel*>(model);
    Q_ASSERT(feLinkModel);
    if(!feLinkModel)
        return;

    QTreeView::setModel(model);
    connect(model, &QAbstractItemModel::modelReset, this, &FeLinkTreeView::onModelReset);
    connect(phyResetPbDelegate, &PhyResetButtonDelegate::phyReset,
            feLinkModel, &FeLinkTreeModel::phyReset);

    setItemDelegateForColumn(static_cast<int>(FeLinkTreeColumns::COL_RESET_PHY_BUT), phyResetPbDelegate);
}

void FeLinkTreeView::saveAppGuiSettings(QSettings &s) const
{
    s.beginGroup(EXPAND_DEVICE_MAP);
    QStringList devList;
    for(DeviceIndex index : expand_device_map.keys()){
        s.setValue(index.toString(), expand_device_map[index]);
        devList << index.toString();
    }
    s.setValue("devList", devList);
    s.endGroup();
}

void FeLinkTreeView::loadAppGuiSettings(const QSettings &s)
{
    expand_device_map.clear();
    const auto &devList = s.value(EXPAND_DEVICE_MAP+"/devList").toStringList();
    for(const QString &key : devList){
        DeviceIndex index = DeviceIndex::fromString(key);
        if(!index.isNull())
            expand_device_map[index] = s.value(EXPAND_DEVICE_MAP+"/"+key).toBool();
    }
}

QSize FeLinkTreeView::sizeHint() const
{
    auto size = QTreeView::sizeHint();
    auto m = model();
    if(!m)
        return size;

    int height = 2;
    if(auto h = header())
        if(h->isVisible())
            height += h->sizeHint().height(); // 26

    for(auto row=0; row<m->rowCount(); ++row) {
        height += recursiveRowHeight(QModelIndex()); // 30 per row
    }

    if(auto sb = horizontalScrollBar())
        height += sb->sizeHint().height(); // 14

    size.setHeight(height);
    return size;
}

void FeLinkTreeView::expanded(const QModelIndex &index)
{
   if(!getDeviceIndex(index).isNull()){
        expand_device_map[getDeviceIndex(index)] = true;
   }
}

void FeLinkTreeView::collapsed(const QModelIndex &index)
{
    if(!getDeviceIndex(index).isNull()){
        expand_device_map[getDeviceIndex(index)] = false;
    }
}

void FeLinkTreeView::onModelReset()
{
    expandTopItem();
    expandAll();
    updateGeometry();
    addDelegate({});
}

void FeLinkTreeView::addDelegate(const QModelIndex &parIndex)
{
    const auto m = model();
    for(auto c : {FeLinkTreeColumns::COL_RESET_PHY_BUT}) {
        auto col = static_cast<int>(c);
        for(int r=0; r<m->rowCount(parIndex); ++r) {
            const auto &childIndex = m->index(r, col, parIndex);
            if(!isPersistentEditorOpen(childIndex))
                openPersistentEditor(childIndex);
            addDelegate(childIndex);
        }
        resizeColumnToContents(col);
    }
}

void FeLinkTreeView::expandTopItem()
{
    auto model = qobject_cast<const FeLinkTreeModel*>(this->model());
    auto top = model->index(0,0);
    expand(top);

    for(DeviceIndex index : expand_device_map.keys()){
        auto row = model->getDevRow(index);
        if(expand_device_map[index])
            expand(model->index(row, 0, top));
        else
            collapse(model->index(row, 0, top));
    }

    for(int row=0; row<this->model()->rowCount(); ++row){
        DeviceIndex index = static_cast<FeLinkTreeModelItem*>(model->index(row, 0, top).internalPointer())->dev;
        if(!expand_device_map.contains(index))
            expand_device_map.remove(index);
     }
}

DeviceIndex FeLinkTreeView::getDeviceIndex(const QModelIndex &index) const
{
    return static_cast<FeLinkTreeModelItem*>(index.internalPointer())->dev;
}

int FeLinkTreeView::recursiveRowHeight(const QModelIndex &index) const
{
    int height = rowHeight(index);
    auto m = model();
    for(auto row=0; row<m->rowCount(index); ++row) {
        auto child = m->index(row,0,index);
        height += recursiveRowHeight(child);
    }
    return height;
}
