/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "FpgaBitFile.h"
#include <netinet/in.h>

#include <QDebug>
#include <QFile>
#include <QString>
#include <QDataStream>

#include "base/HexDump.h"

int FpgaBitFile::findBitstreamSync(const QByteArray &data)
{
    const char *fileData = data.constData();
    const int dataLen = data.size();
    const int headerSizeMax = std::min(dataLen - static_cast<int>(sizeof(uint32_t)), 512);
    for (int i = 0; i < headerSizeMax; ++i) {
        auto valSync = reinterpret_cast<const uint32_t *>(fileData + i);
        if (*valSync == BITSTREAM_SYNC) {
            return i;
        }
    }
    return -1;
}

QMap<int, QString> type1regs = {
    {0, "CRC"},
    {1, "FAR"},
    {2, "FDRI"},
    {3, "FDRO"},
    {4, "CMD"},
    {5, "CTL0"},
    {6, "MASK"},
    {7, "STAT"},
    {8, "LOUT"},
    {9, "COR0"},
    {10, "MFWR"},
    {11, "CBC"},
    {12, "IDCODE"},
    {13, "AXSS"},
    {14, "COR1"},
    {16, "WBSTAR"},
    {17, "TIMER"},
    {22, "BOOTSTS"},
    {24, "CTL1"},
    {31, "BSPI"}
};

void FpgaBitFile::dumpBitstream(const QByteArray &bitstream)
{
    auto syncPos = findBitstreamSync(bitstream);
    if (syncPos < 0 || (syncPos + 4*12) > bitstream.size())
        return;

    auto dumpSize = std::min(int(bitstream.size() - syncPos), 4*16);
    qDebug().noquote() << "bitstream dump:\n" +
            HexDump::toStringList(bitstream.constData()+syncPos, dumpSize).join("\n");

    QVector<QString> opcodeStr = {"NOP", "Read", "Write", "Reserved"};

    auto data = reinterpret_cast<const uint32_t *>(bitstream.data() + syncPos + 4);

    for (int i=0; i<dumpSize; i++) {
        uint32_t d = ntohl(data[i]);
        int type = d >> 29;
        int opcode = (d >> 27) & 0x3;
        int t1_reg = (d >> 13) & 0x1F;
        int wc = (type == 1) ? (d & 0x7FF) : (d & 0x07FFFFFF);
        QString s;
        if (type == 1) {
            s = QString("%1 %2 type-1 %3")
                    .arg(i, 2)
                    .arg(d, 8, 16, QChar('0'))
                    .arg(opcodeStr[opcode]);
            if (opcode != 0)
                s += QString(" %1 #%2")
                        .arg(type1regs[t1_reg])
                        .arg(wc);
        }
        if (type == 2) {
            s = QString("%1 %2 type-2 %3 #%4")
                                      .arg(i, 2)
                                      .arg(d, 8, 16, QChar('0'))
                                      .arg(opcodeStr[opcode])
                                      .arg(wc);
        }
        qDebug().noquote() << s;
        for (int k=1; k<=wc; k++) {
            qDebug().noquote() << QString("%1 %2 data")
                                      .arg(i+k, 2)
                                      .arg(ntohl(data[i+k]), 8, 16, QChar('0'));
            if (k > 5) break;
        }
        i+=wc;
    }
}

bool FpgaBitFile::readFile(const QString fileName)
{
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly)) {
        qWarning().noquote() << QString("Can't open file %1").arg(fileName);
        return {};
    }

    const auto &fileData = file.readAll();
    file.close();

    if (!parseBitFile(fileData))
        return {};

    auto syncPos = findBitstreamSync(bitstream);
    if (syncPos < 0) {
        qWarning().noquote() << QString("Unknown bitstream, sync not found: %1").arg(fileName);
        dumpBitstream(bitstream);
        return {};
    }
    // bitstream.remove(0, syncPos);
    auto valSync = reinterpret_cast<const uint32_t *>(bitstream.constData() + syncPos);
    Q_ASSERT(*valSync == BITSTREAM_SYNC);

    qDebug().noquote()
        << QString("Read bitstream: %1 bytes (sync at %2), %3, %4, %5")
               .arg(bitstream.size())
               .arg(syncPos)
               .arg(getProject())
               .arg(getPart())
               .arg(getDateTime().toString());

    return true;
}

bool FpgaBitFile::parseBitFile(const QByteArray &bitfileData)
{
    QByteArray bitstream;
    QMap<QString, QString> meta;
    QDataStream ds(bitfileData);
    {
        uint16_t h1 = 0;
        uint32_t h2 = 0;
        uint32_t h3 = 0;
        uint16_t h4 = 0;
        uint8_t h5 = 0;
        ds >> h1 >> h2 >> h3 >> h4 >> h5;
        if (ds.status() != QDataStream::Ok)
            goto err;
        if (h1 != 0x0009)
            goto err;
        if (h2 != 0x0FF00FF0)
            goto err;
        if (h3 != 0x0FF00FF0)
            goto err;
        if (h4 != 0x0000)
            goto err;
        if (h5 != 0x01)
            goto err;
        while (!ds.atEnd()) {
            const auto pos = ds.device()->pos();
            uint8_t code = 0;
            uint32_t len = 0;
            ds >> code;
            if (code == 'e') {
                uint32_t len32;
                ds >> len32;
                len = len32;
            } else {
                uint16_t len16;
                ds >> len16;
                len = len16;
            }
            if (ds.status() != QDataStream::Ok)
                goto err;
            if (len < 1)
                goto err;
            if (pos + len  > bitfileData.size())
                goto err;
            QByteArray buf;
            buf.resize(len);
            ds.readRawData(buf.data(), len);
            if (ds.status() != QDataStream::Ok)
                break;
            if (code >= 'a' && code <= 'd') {
                if (!buf.endsWith('\0'))
                    goto err;
            }
            switch(code) {
                case 'a': meta["project"] = buf; break;
                case 'b': meta["part"] = buf; break;
                case 'c': meta["date"] = buf; break;
                case 'd': meta["time"] = buf; break;
                case 'e': bitstream = buf; break;
                default: goto err;
            }
            // qDebug().noquote() << HexDump::toStringList(buf.constData(), std::min(buf.size(), 256)).join("\n");
        }
    }
    this->meta = meta;
    this->bitstream = bitstream;
    return true;
err:
    qWarning().noquote() << QString("Parse error at offset %1").arg(ds.device()->pos());
    return false;
}

QString FpgaBitFile::getProject() const
{
    auto list = meta["project"].split(";");
    if (list.empty())
        return {};
    return list.first();
}

QString FpgaBitFile::getPart() const
{
    return meta["part"];
}

QDateTime FpgaBitFile::getDateTime() const
{
    return QDateTime::fromString(meta["date"] + " " + meta["time"], "yyyy/MM/dd hh:mm:ss");
}
