#include "Histogramm.h"

#include <numeric>
#include <math.h>

Histogramm::Histogramm():
    Histogramm("", 0, 0, 0)
{}

Histogramm::Histogramm(QString _title, int _nBins, double _min, double _max):
    title(_title),
    n_bins(_nBins),
    min(_min),
    max(_max),
    n_entries(0)
{
    resize(n_bins, min, max);
}

bool Histogramm::isEmpty()
{
    if(title.isEmpty() && n_bins == 0 && min == 0 && max == 0)
        return true;
    else
        return false;
}

void Histogramm::reset()
{
    if(n_bins)
        memset(&bins[0], 0, static_cast<size_t>(n_bins)*sizeof(double));
    under_flow = 0;
    over_flow = 0;
    n_entries = 0;
    min_filled_val = 0;
    max_filled_val = 0;
    min_filled_bin = -1;
    max_filled_bin = -1;
}

void Histogramm::resize(int n_bins_, double min_, double max_)
{
    n_bins = n_bins_;
    min = min_;
    max = max_;
    bins.resize(n_bins);
    reset();
    if(n_bins)
        bin_step = (max-min)/n_bins;
    else
        bin_step = 0;
}

void Histogramm::fill(double v, int w)
{
    Q_ASSERT(w>0);
    if(v < min) {
        ++under_flow;
    } else if(v >= max) {
        ++over_flow;
    } else {
        int binNum = findBin(v);
        Q_ASSERT(binNum>=0 && binNum<n_bins);
        bins[binNum] += w;
        if(min_filled_bin == -1) {
            min_filled_bin = max_filled_bin = binNum;
        } else {
            min_filled_bin = std::min(min_filled_bin, binNum);
            max_filled_bin = std::max(max_filled_bin, binNum);
        }
    }
    if(n_entries) {
        min_filled_val = qMin(min_filled_val, v);
        max_filled_val = qMax(max_filled_val, v);
    } else {
        max_filled_val = min_filled_val = v;
    }
    n_entries += static_cast<uint>(w);
}

int Histogramm::findBin(double v) const
{
    if(n_bins && v >= min && v < max)
        return static_cast<int>(floor((v-min)/bin_step));
    else
        return -1;
}

uint Histogramm::getEntries() const
{
    return n_entries;
}

double Histogramm::getBinLower(int bin) const
{
    return min+bin_step*bin;
}

double Histogramm::getBinCenter(int bin) const
{
    return min+bin_step*(bin+0.5);
}

double Histogramm::getBinUpper(int bin) const
{
    return min+bin_step*(bin+1);
}

DistributionParams Histogramm::getDistribution() const
{
    DistributionParams fit;
    fit.cnt = n_entries-getUnderFlow()-getOverFlow();
    int minBin=-1;
    int maxBin=-1;
    for(int bin=0; bin<n_bins; ++bin) {
        const auto &binVal = bins[bin];
        if(!qFuzzyIsNull(binVal)) {
            fit.mean += getBinCenter(bin)*binVal;
            if(minBin==-1)
                minBin = bin;
            maxBin = bin;
        }
    }
    if(fit.cnt) {
        fit.mean /= fit.cnt;

        for(int bin=0; bin<n_bins; ++bin) {
            const auto &binVal = bins[bin];
            if(!qFuzzyIsNull(binVal)) {
                double dx = getBinCenter(bin)-fit.mean;
                fit.rms += dx*dx*binVal;
            }
        }
        fit.rms = sqrt(fit.rms/fit.cnt);
        if(minBin!=-1) {
            fit.min = getBinLower(minBin);
            fit.max = getBinUpper(maxBin);
        }
    }
    return fit;
}
