#include "ClientManagerFsm.h"

#include <QDebug>
#include <QLoggingCategory>

#include "base/DeviceIndexEnabled.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.client")
}

ClientManagerFsm::ClientManagerFsm(QObject *parent):
    BaseFsm(parent)
{
    setObjectName("ClientFSM");
    recreateInput(new ClientsFsmInput());
}

bool ClientManagerFsm::isTransitionAllowed(const FsmState &toState) const
{
    switch (getState()) {
    case FsmState::Ready:
        if(toState == FsmState::Starting) return true;
        break;
    case FsmState::Starting:
        if(toState == FsmState::Run) return allClientsInState(FsmState::Run);
        break;
    case FsmState::Run:
        if(toState == FsmState::Stopping) return true;
        break;
    case FsmState::Stopping:
        if(toState == FsmState::Ready) return false;
        break;
    default:;
    }
    return BaseFsm::isTransitionAllowed(toState);
}

void ClientManagerFsm::setClientList(const ClientIndexSet &newList,
                                 const QMap<ClientIndex, FsmState> &clientStates)
{
    if(getInput()->clientSet == newList
            && getInput()->clientStates == clientStates)
        return;
    getInput()->clientSet = newList;
    getInput()->clientStates = clientStates;
    auto dbg = qDebug(fsm) << __PRETTY_FUNCTION__;
    printInputString(dbg);
    evaluate();
}

void ClientManagerFsm::setClientState(const ClientIndex &id, const FsmState &clState)
{
    getInput()->clientStates[id] = clState;
    qDebug(fsm) << __PRETTY_FUNCTION__ << QString("id:%1 state:%2")
                   .arg(id.toString()).arg(fsmStateToString(clState));
    evaluate();
}

bool ClientManagerFsm::evaluate()
{
    if(BaseFsm::evaluate())
        return true;

    switch (getState()) {
    case FsmState::Error:
        if (allClientsInState(FsmState::Ready)) {
            setState(FsmState::Ready);
            return true;
        }
        break;
    case FsmState::Ready:
        if (!allClientsInState(FsmState::Ready)) {
            qDebug(fsm) << "Some clients not in Ready state, go to Error state";
            setState(FsmState::Error);
            return true;
        }
        break;
    case FsmState::Starting:
        if(isAnyClientInError()) {
            setState(FsmState::Error);
            return true;
        }
        break;
    case FsmState::Run:
        if (!allClientsInState(FsmState::Run)) {
            setState(FsmState::Error);
            return true;
        }
        break;
    case FsmState::Stopping:
        if (allClientsInState(FsmState::Ready)) {
            setState(FsmState::Ready);
            return true;
        }
        break;
    default:;
    }

    return false;
}

void ClientManagerFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 ClientManagerFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::Starting:
    case FsmState::Stopping:
        return 6666;
    default:
        return BaseFsm::stateTimeoutMs(state);
    }
}

bool ClientManagerFsm::allClientsInState(const FsmState &state) const
{
    for(auto id : getInput()->clientSet)
        if(getInput()->clientStates.value(id, FsmState::Init) != state)
            return false;
    return true;
}

bool ClientManagerFsm::isAnyClientInError() const
{
    for(auto id : getInput()->clientSet)
        if(getInput()->clientStates.value(id, FsmState::Init) == FsmState::Error)
            return true;
    return false;
}


QDebug operator<<(QDebug dbg, const ClientsFsmInput &in)
{
    dbg << BaseFsmInput(in);

    dbg << "clientSet:" << in.clientSet << "; ";
    dbg << "clientStatus:" << in.clientStates << "; ";

    return dbg;
}
