/*
**    Copyright 2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ZmqKvPublisher.h"

#include <QDebug>
#include <QEvent>
#include <QLoggingCategory>
#include <QTimer>
#include <QThread>
#include <QVariant>

#include <zmq_addon.hpp>

#include "base/ServiceConfigUrl.h"
#include "ZmqContext.h"
// #include "ZmqMonitor.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_zmq)
Q_LOGGING_CATEGORY(lc_zmq, "afi.zmq")
static const char *DEFAULT_URL = "tcp://localhost:5678";
static const bool DEFAULT_ENABLE = true;
static const int DEFERRED_CONNECT_TIMEOUT = 0;
}

ZmqKvPublisher::ZmqKvPublisher(const QString &serviceName, QObject *parent)
    : AbstractPublisher(parent),
    context(ZmqContext::instance())
{
    setObjectName("ZmqKvPublisher");
    ServiceConfigUrl u(serviceName, QUrl::fromUserInput(DEFAULT_URL), DEFAULT_ENABLE);
    enabled = u.isEnabled();
    url = u.getUrl();
    if (url.isEmpty())
        enabled = false;
    if (!enabled)
        return;

    // thread may change after ctor, use QueuedConnection
    auto t = new QTimer(this);
    t->setSingleShot(true);
    t->start(DEFERRED_CONNECT_TIMEOUT);
    connect(t, &QTimer::timeout, this, &ZmqKvPublisher::deferredConnect, Qt::QueuedConnection);
}

ZmqKvPublisher::~ZmqKvPublisher()
{
    deleteSocket();
}

void ZmqKvPublisher::createSocket()
{
    deleteSocket();
    try {
        sock = new zmq::socket_t(context, ZMQ_PUB);
        // new ZmqMonitor(*sock, addr, this);
    }  catch (std::exception &e) {
        qCritical(lc_zmq).nospace() << __func__ << ": " << e.what();
    }
}

void ZmqKvPublisher::deleteSocket()
{
    if (!sock)
        return;
    delete sock;
    sock = nullptr;
}

bool ZmqKvPublisher::event(QEvent* ev)
{
    if (ev->type() == QEvent::ThreadChange) {
        if (sock) {
            deleteSocket();
            QTimer::singleShot(0, this, &ZmqKvPublisher::deferredConnect);
        }
        return true;
    }
    return QObject::event(ev);
}

void ZmqKvPublisher::deferredConnect()
{
    createSocket();
    Q_ASSERT(sock);
    if (!sock)
        return;
    auto addr = url.toString().toLatin1();
    qInfo(lc_zmq).noquote().nospace() << "connecting to '" << addr << "'";
    try {
        sock->connect(addr);
    }  catch (std::exception &e) {
        qCritical().nospace() << __func__ << ": " << e.what();
    }
}

void ZmqKvPublisher::pub(const QString &key, const QByteArray &data)
{
    if(!enabled)
        return;

    if (!sock) {
        qWarning(lc_zmq).nospace() << __func__ << ": socket not initialized, message dropped";
        return;
    }
    try {
        std::array<zmq::message_t, 2> imsgs = {
            zmq::message_t(key.toStdString()),
            zmq::message_t(data.constData(), data.size())
        };
        auto ret = zmq::send_multipart(*sock, imsgs, zmq::send_flags::dontwait);
        Q_ASSERT(*ret == 2);
    }  catch (std::exception &e) {
        qCritical(lc_zmq).nospace() << "send: " << e.what();
    }
}
