//
//    Copyright 2016 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
#include "MetricSenderHelper.h"

#include <QProcessEnvironment>
#include <QUrl>
#include <QHostInfo>
#include <QCoreApplication>
#include <QRegularExpression>
#include <QDateTime>
#include <QSettings>

#include "base/Globals.h"

namespace {
// Set environment variable in form hostname:port
static const int METRICS_URL_DEFAULT_PORT = 8089;
static const char *METRICS_URL_DEFAULT_HOST = "localhost";
static const char *METRICS_URL_DEFAULT_SCHEME = "udp";
enum {
    METRIC_FORMAT_GRAPHITE = 1,
    METRIC_FORMAT_INFLUX = 2,
};
static const int METRIC_OUTPUT_FORMAT = METRIC_FORMAT_INFLUX;
}

static QString getMetricsUrl()
{
    const auto &settings = Globals::getSettings();
    QString key = Globals::getEnvironment() + "/" + "influxdb";
    if (settings->contains(key)) {
        return settings->value(key).toString();
    }
    // nothing defined
    return QString();
}

MetricSenderHelper::MetricSenderHelper(QObject *parent)
    : QObject(parent),
      baseTags(MetricSenderHelper::defaultTags()),
      enable(false)
{
    const QString &url_str = getMetricsUrl();
    if (url_str.isEmpty()) {
        qInfo() << QString("Metrics disabled: no config value");
        return;
    }
    url = QUrl::fromUserInput(url_str);
    if (!url.isValid()) {
        qWarning() << QString("Metrics disabled: '%1' is not a valid URL")
                      .arg(url_str);
        return;
    }
    if (url.scheme().compare("http") == 0)
        url.setScheme(METRICS_URL_DEFAULT_SCHEME);
    if (url.scheme().compare("udp") != 0) {
        qWarning() << QString("Metrics disabled: %1 is not a supported scheme")
                      .arg(url.scheme());
        return;
    }
    if (url.port() <= 0)
        url.setPort(METRICS_URL_DEFAULT_PORT);

    if (url.host().isEmpty()) {
        url.setHost(METRICS_URL_DEFAULT_HOST);
    }
    // resolve to IP address
    const QHostInfo &hostInfo = QHostInfo::fromName(url.host());
    const QList<QHostAddress> &addresses = hostInfo.addresses();
    if (addresses.isEmpty()) {
        qWarning() << QString("Metrics disabled: no IP address for %1")
                      .arg(url.toDisplayString());
        return;
    }
    hostAddrCached = addresses.first(); // TODO: asynchronous resolver

    enable = true;
    qInfo().noquote() << QString("[MetricSender]: Metrics destination %1")
               .arg(url.toDisplayString());
}

MetricTags MetricSenderHelper::defaultTags()
{
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    const QString appDomain = Globals::getAppDomain();
    return MetricTags(
        {
            {"ns", "daq"},
            {"domain", appDomain},
            {"host", shortHostName},
            {"app", QCoreApplication::applicationName()},
            });
}

// Tag names must have a length >= 1 and may contain any ascii characters except ;!^=
// Tag values must also have a length >= 1, they may contain any ascii characters
// except ; and the first character must not be ~

static QString formatTagGraphite(QString k, QString v)
{
    k.replace(' ', '_');
    k.replace(';', '_');
    k.replace('!', '_');
    k.replace('^', '_');
    k.replace('=', '_');
    v.replace(';', '_');
    if (v.startsWith('~'))
        v.replace(0, 1, '_');
    return k + "=" + v;
}

static QStringList formatTagsGraphite(const MetricTags &tags)
{
    QStringList sl;
    for (auto it = tags.begin(); it != tags.end(); ++it) {
        QString k = it.key();
        QString v = it.value();
        if (k.length() < 1 || v.length() < 1)
            continue;
        sl += formatTagGraphite(k, v);
    }
    return sl;
}

QString quoteNonNumeric(const QString &v)
{
    bool isNumber;
    v.toDouble(&isNumber);
    if (isNumber)
        return v;
    return '"' + v + '"';
}

static QStringList formatFields(const MetricFields &fields)
{
    QStringList sl;
    for (auto it = fields.begin(); it != fields.end(); ++it) {
        QString k = it.key();
        QString v = it.value().toString();
        if (k.length() < 1 || v.length() < 1)
            continue;
        v = quoteNonNumeric(v);
        sl += k + "=" + v;
    }
    return sl;
}

static QString quote_influx_keys(QString s)
{
    return s
        .replace(QChar(','), "\\,")
        .replace(QChar('='), "\\=")
        .replace(QChar(' '), "\\ ");
}

static QString quote_influx_meas(const QString &s)
{
    QString v(s);
    v.replace(QChar(','), "\\,").replace(QChar('"'), "\\\"").replace(QChar(' '), "\\ ");
    v = quoteNonNumeric(v);
    return v;
}

static QStringList formatTagsInflux(const MetricTags &tags)
{
    QStringList sl;
    for (auto it = tags.begin(); it != tags.end(); ++it) {
        QString k = it.key();
        QString v = it.value();
        if (k.length() < 1 || v.length() < 1)
            continue;
        sl += QString(quote_influx_keys(k) + "=" + quote_influx_keys(v));
    }
    return sl;

}

static QStringList formatFieldsInflux(const MetricFields &fields)
{
    QStringList sl;
    for (auto it = fields.begin(); it != fields.end(); ++it) {
        QString k = it.key();
        QString v = it.value().toString();
        if (k.length() < 1 || v.length() < 1)
            continue;
        sl += quote_influx_keys(k) + "=" + quote_influx_meas(v);
    }
    return sl;
}

QString MetricSenderHelper::formatMetricLine(const MetricEntity &ent)
{
    QString s;
    switch (METRIC_OUTPUT_FORMAT) {
    case METRIC_FORMAT_INFLUX: {
        // <measurement>[,<tag_key>=<tag_value>[,<tag_key>=<tag_value>]] <field_key>=<field_value>[,<field_key>=<field_value>] [<timestamp>]
        {
            QString series_name = ent.measurement;
            series_name.replace(QRegularExpression("[^A-Za-z0-9.]"), QChar('_'));
            auto tags = baseTags;
            tags.insert(ent.tags);
            QStringList sl;
            sl += series_name;
            sl += formatTagsInflux(tags);
            series_name = sl.join(",");
            s += series_name;
        }
        s += " ";
        s += formatFieldsInflux(ent.fields).join(",");
        s += " ";
        qint64 ts_ns = QDateTime::currentMSecsSinceEpoch() * 1000000ull;
        s += QString::number(ts_ns);
        s += "\n";
        break;
    }
    case METRIC_FORMAT_GRAPHITE:
    default: {
        QString series_name = ent.measurement;
        series_name.replace(QRegularExpression("[^A-Za-z0-9.]"), QChar('_'));
        auto tags = baseTags;
        tags.insert(ent.tags);
        QStringList sl;
        sl += series_name;
        sl += formatTagsGraphite(tags);
        series_name = sl.join(";");

        s += series_name;
        s += " ";
        s += QString("%2").arg(ent.fields["value"].toString());
        s += " ";
        s += QString::number(QDateTime::currentDateTimeUtc().toTime_t());
        s += "\n";
        break;
    }
    }
    return s;
}

QHostAddress MetricSenderHelper::hostAddr() const
{
    return hostAddrCached;
}

quint16 MetricSenderHelper::hostPort() const
{
    return url.port(METRICS_URL_DEFAULT_PORT);
}

QUrl MetricSenderHelper::getUrl() const
{
    return url;
}
