#include "PNPMessage.h"

#include "QStringList"
#include "QDomDocument"
#include "QDomNodeList"

#include "base/DaqModule.h"
#include "base/ProgramIndex.h"

#include "PNPServerUtil.h"

namespace {
const QString DOC_TYPE = "pnp_message";

const QString TAG_SEARCH = "discover_request";
const QString TAG_SEARCH_TARGET = "target";

const QString TAG_PROGRAM_CLOSE = "program_close";
const QString TAG_PROGRAM_ANNOUNCE = "program";
__attribute__((deprecated("will be deleted in future versions")))
const QString TAG_PROGRAM_NAME = "name";
const QString TAG_PROGRAM_TYPE = "type";
const QString TAG_PROGRAM_INDEX = "index";
const QString TAG_PROGRAM_PARENT = "parent";
const QString TAG_PROGRAM_UUID = "uuid";
const QString TAG_PROGRAM_SEQ = "seq";
const QString TAG_PROGRAM_HOST_NAME = "hostName";
const QString TAG_PROGRAM_HOST = "host";
const QString TAG_VER_HASH = "ver_hash";
const QString TAG_VER_DATE = "ver_date";

const QString TAG_OPTIONS_SECTION = "options";
const QString TAG_OPTION_SECTION = "option";
const QString TAG_OPTION_NAME = "name";
const QString TAG_OPTION_VALUE = "value";

const QString TAG_INTERFACES_SECTION = "interfaces";
const QString TAG_INTERFACE_SECTION = "interface";
//const QString TAG_INTERFACE_HOST = "host"; // deprecated
const QString TAG_INTERFACE_PORT = "port";
const QString TAG_PEER_SECTION = "peer";
const QString TAG_PEER_HOST = "h";
const QString TAG_PEER_PORT = "p";
const QString TAG_INTERFACE_TYPE = "type";
const QString TAG_INTERFACE_ENABLED = "enabled";
const QString TAG_INTERFACE_ISFREE = "isFree";
const QString TAG_INTERFACE_ID = "id";
}

PNPMessage::PNPMessage()
{}

QString PNPMessage::toString() const
{
    QString res;
    switch (type) {
    case PNPMessage_Type_Search: {
        res = QString("Discover_request for: %1").arg(getSearchString(searchTargets));
        break;
    }
    case PNPMessage_Type_Discover:
        res = QString("Program announcement: clientIndex='%1' uuid='%2'")
                .arg(progDesc.getName())
                .arg(progDesc.uuid.toString());
        for (auto it : progDesc.interfaces) {
            QStringList peerStr;
            for (auto peer : it.peers)
                peerStr << peer.toString();
            res.append(QString("\ninterface %1 %5 at %2:%3 %4")
                       .arg(it.type).arg(it.host.toString()).arg(it.port)
                       .arg(it.isFree?"free":QString("busy (%1)").arg(peerStr.join(";")))
                       .arg(it.enabled?"enabled":"disabled"));
        }
        for (auto opKey : progDesc.options.keys()) {
            res.append(QString("\n[%1]: %2").arg(opKey).arg(progDesc.options[opKey]));
        }
        break;
    case PNPMessage_Type_Close:
        res = QString("Program closing: clientIndex='%1' uuid='%2'")
                .arg(progDesc.getName())
                .arg(progDesc.uuid.toString());
        break;
    default:
        break;
    }
    return res;
}

QString PNPMessage::getXml() const
{
    switch (type) {
    case PNPMessage_Type_Search:
    {
        QDomDocument doc(DOC_TYPE);
        QDomElement root = doc.createElement(TAG_SEARCH);
        doc.appendChild(root);
        if(!searchTargets.isEmpty()){
            for (auto daqMod : searchTargets) {
                if(daqMod.isEmpty()) {
                    static bool showWarn = true;
                    if(showWarn) {
                        qWarning() << __PRETTY_FUNCTION__ << "searchTargets contains empty DaqModule";
                        showWarn = false;
                    }
                    continue;
                }
                QDomElement target = doc.createElement(TAG_SEARCH_TARGET);
                root.appendChild(target);
                QDomText text = doc.createTextNode(daqMod.toString());
                target.appendChild(text);
            }
        }
        return doc.toString();
    }
    case PNPMessage_Type_Discover:
        return progDesc.getAnnounceStr();
    case PNPMessage_Type_Close:
        return progDesc.getClosingStr();
    default:
        Q_ASSERT(false);
        return QString();
    }
}

bool PNPMessage::decode(const QByteArray &data, const QHostAddress &host, bool *ok)
{
    QDomDocument mDocument;
    if(!mDocument.setContent(data) || mDocument.doctype().name() != DOC_TYPE)
        return false;
    bool local_ok=true;
    QDomElement  rootElement;
    rootElement = mDocument.documentElement();

    if(rootElement.tagName() == TAG_PROGRAM_ANNOUNCE ||
            rootElement.tagName() == TAG_PROGRAM_CLOSE){
        QString daqTypeStr, programIndexStr;
        if(rootElement.tagName() == TAG_PROGRAM_ANNOUNCE)
            type = PNPMessage_Type_Discover;
        else
            type = PNPMessage_Type_Close;
        if(rootElement.hasAttribute(TAG_PROGRAM_HOST_NAME)) {
            progDesc.hostName = rootElement.attribute(TAG_PROGRAM_HOST_NAME);
        } else {
            progDesc.hostName = host.toString();
            qWarning() << QString("Program announce from %1 has no attribute %2")
                          .arg(progDesc.hostName)
                          .arg(TAG_PROGRAM_HOST_NAME);
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_TYPE)) {
            daqTypeStr = rootElement.attribute(TAG_PROGRAM_TYPE);
        } else {
            qWarning() << QString("Program announce from %1 has no attribute %2")
                          .arg(progDesc.hostName)
                          .arg(TAG_PROGRAM_TYPE);
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_INDEX)) {
            programIndexStr = rootElement.attribute(TAG_PROGRAM_INDEX);
        } else {
            qWarning() << QString("Program announce from %1 has no attribute %2")
                          .arg(progDesc.hostName)
                          .arg(TAG_PROGRAM_INDEX);
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_PARENT)) {
            progDesc.parentClientIndex = ClientIndex::fromString(
                        rootElement.attribute(TAG_PROGRAM_PARENT));
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_UUID)) {
            progDesc.uuid = QUuid(rootElement.attribute(TAG_PROGRAM_UUID));
        } else {
            qWarning() << QString("Program announce from '%1' has no attribute %2")
                          .arg(progDesc.getName())
                          .arg(TAG_PROGRAM_UUID);
            local_ok = false;
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_SEQ)) {
            bool seqOk;
            progDesc.seq = rootElement.attribute(TAG_PROGRAM_SEQ).toUInt(&seqOk);
            if(!seqOk){
                progDesc.seq = 0;
//                qWarning() << QString("Program announce from '%1' has wrong attribute %2: %3")
//                              .arg(progDesc.getName())
//                              .arg(TAG_PROGRAM_SEQ)
//                              .arg(rootElement.attribute(TAG_PROGRAM_SEQ));
//                local_ok = false;
            }
        } else {
//            qWarning() << QString("Program announce from '%1' has no attribute %2")
//                          .arg(progDesc.getName())
//                          .arg(TAG_PROGRAM_SEQ);
//            local_ok = false;
        }
        if(rootElement.hasAttribute(TAG_VER_HASH)){
            progDesc.ver_hash = rootElement.attribute(TAG_VER_HASH);
        }
        if(rootElement.hasAttribute(TAG_VER_DATE)){
            progDesc.ver_date = rootElement.attribute(TAG_VER_DATE);
        }

        progDesc.host = rootElement.hasAttribute(TAG_PROGRAM_HOST)
            ? QHostAddress(rootElement.attribute(TAG_PROGRAM_HOST))
            : host;

        for(int i=0; (i < rootElement.childNodes().size()) && local_ok; ++i) {
            const QDomElement domEl = rootElement.childNodes().at(i).toElement();
            if(domEl.tagName() == TAG_PROGRAM_TYPE){
                daqTypeStr = domEl.text();
            } else if(domEl.tagName() == TAG_PROGRAM_INDEX){
                programIndexStr = domEl.text();
            } else if(domEl.tagName() == TAG_PROGRAM_HOST_NAME){
                progDesc.hostName = domEl.text();
            } else if(domEl.tagName() == TAG_PROGRAM_UUID){
                progDesc.uuid = QUuid(domEl.text());
            } else if(domEl.tagName() == TAG_PROGRAM_SEQ){
                progDesc.seq = domEl.text().toUInt();
            } else if(domEl.tagName() == TAG_VER_HASH){
                progDesc.ver_hash = domEl.text();
            } else if(domEl.tagName() == TAG_VER_DATE){
                progDesc.ver_date = domEl.text();
            } else if(domEl.tagName() == TAG_OPTIONS_SECTION){
                local_ok = decodeOptions(domEl, progDesc.hostName);
            } else if(domEl.tagName() == TAG_INTERFACES_SECTION){
                local_ok = decodeInterfaces(domEl, progDesc.host);
            }  else {
                qWarning() << QString("Unknown tag:%1 from: %2")
                              .arg(domEl.tagName()).arg(host.toString());
                return false;
            }
        }
        const auto &daqModule = DaqModule::fromString(daqTypeStr);
        const auto &programIndex = ProgramIndex::fromString(programIndexStr);
        progDesc.setClientIndex({daqModule, programIndex});
        if(type == PNPMessage_Type_Discover)
            local_ok &= !progDesc.isEmpty();
    } else if(rootElement.tagName() == TAG_SEARCH) {
        type = PNPMessage_Type_Search;
        for(int i=0; (i < rootElement.childNodes().size()) && local_ok; ++i) {
            const QDomElement domEl = rootElement.childNodes().at(i).toElement();
            if(domEl.tagName() == TAG_SEARCH_TARGET) {
                const auto &searchTarg = DaqModule::fromString(domEl.text());
                if(searchTarg.isEmpty()) {
                    static bool showWarn = true;
                    if(showWarn) {
                        qWarning() << "Can't detect search target:" << domEl.text();
                        qWarning() << "Whole pnp msg is:\n" << data;
                        showWarn = false;
                    }
                } else {
                    searchTargets << searchTarg;
                }
            } else
                local_ok = false;
        }
    } else
        local_ok = false;
    if(ok)
        *ok = local_ok;
    return local_ok;
}

void PNPMessage::addInterface(const QString &type, const QHostAddress &host, const quint32 port)
{
    progDesc.interfaces.push_back(ProgramInterface(host, port, type));
}

bool PNPMessage::decodeOptions(const QDomElement &domEl, const QString &hostName)
{
    if(domEl.tagName() != TAG_OPTIONS_SECTION)
        return false;
    for(int i=0; i < domEl.childNodes().size(); ++i) {
        const QDomElement optionEl = domEl.childNodes().at(i).toElement();
        if (optionEl.tagName() != TAG_OPTION_SECTION) {
            qWarning()<<QString("Unexpected tag name:%1 from:%2")
                        .arg(optionEl.tagName()).arg(hostName);
            return false;
        }
        if (!optionEl.hasAttribute(TAG_OPTION_NAME)) {
            qWarning()<<QString("There is no attribute :%1 from:%2")
                        .arg(TAG_OPTION_NAME).arg(hostName);
            return false;
        }
        if (!optionEl.hasAttribute(TAG_OPTION_VALUE)) {
            qWarning()<<QString("There is no attribute :%1 from:%2")
                        .arg(TAG_OPTION_VALUE).arg(hostName);
            return false;
        }
//        if (optionEl.tagName() != TAG_OPTION_SECTION
//                || !optionEl.hasAttribute(TAG_OPTION_NAME)
//                || !optionEl.hasAttribute(TAG_OPTION_VALUE) )
//            return false;
        QString name = optionEl.attribute(TAG_OPTION_NAME);
        if(!name.isEmpty())
            progDesc.options[name] = optionEl.attribute(TAG_OPTION_VALUE);
    }
    return true;
}

bool PNPMessage::decodeInterfaces(const QDomElement &domEl, const QHostAddress &host)
{
    if(domEl.tagName() != TAG_INTERFACES_SECTION)
        return false;
    for(int i=0; i < domEl.childNodes().size(); ++i) {
        const QDomElement interfaceEl = domEl.childNodes().at(i).toElement();
        if (interfaceEl.tagName() != TAG_INTERFACE_SECTION){
            qWarning() << QString("Unknow tag (%1) in interface section from %2")
                          .arg(interfaceEl.tagName()).arg(progDesc.getName());
            return false;
        }
        if(!interfaceEl.hasAttribute(TAG_INTERFACE_TYPE) ){
            qWarning() << QString("There is no attribute  (%1) in interface section from %2")
                          .arg(TAG_INTERFACE_TYPE).arg(progDesc.getName());
            return false;
        }
        if(!interfaceEl.hasAttribute(TAG_INTERFACE_PORT) ){
            qWarning() << QString("There is no attribute (%1) in interface section from %2")
                          .arg(TAG_INTERFACE_PORT).arg(progDesc.getName());
            return false;
        }
        if(!interfaceEl.hasAttribute(TAG_INTERFACE_ISFREE) ){
            qWarning() << QString("There is no attribute (%1) in interface section from '%2'")
                          .arg(TAG_INTERFACE_ISFREE).arg(progDesc.getName());
            return false;
        }
        Q_ASSERT(!host.isNull());
        ProgramInterface interface;
        interface.host = host;
        interface.type = interfaceEl.attribute(TAG_INTERFACE_TYPE, interface.type);
        bool ok;
        bool enabled = interfaceEl.attribute(TAG_INTERFACE_ENABLED).toInt(&ok)==1;
        if(ok)
            interface.enabled = enabled;
        bool isFree = interfaceEl.attribute(TAG_INTERFACE_ISFREE).toInt(&ok)==1;
        if(ok)
            interface.isFree = isFree;
        int id = interfaceEl.attribute(TAG_INTERFACE_ID).toInt(&ok);
        if(ok)
            interface.id = id;
        interface.port = interfaceEl.attribute(TAG_INTERFACE_PORT).toInt();
        for(int peerId=0; peerId < interfaceEl.childNodes().size(); ++peerId) {
            const QDomElement peerEl = interfaceEl.childNodes().at(peerId).toElement();
            if(peerEl.tagName() != TAG_PEER_SECTION) {
                qWarning() << QString("Unknow tag (%1) in peer section from '%2'")
                              .arg(peerEl.tagName()).arg(progDesc.getName());
                continue;
            }
            ProgramInterfacePeer itPeer;
            if(peerEl.hasAttribute(TAG_PEER_HOST)) {
                QString conHostStr = peerEl.attribute(TAG_PEER_HOST);
                if(!conHostStr.isEmpty()) {
                    QHostAddress conHost(conHostStr);
                    if(!conHost.isNull()) {
                        itPeer.host = conHost;
                    }
                }
            }
            if(peerEl.hasAttribute(TAG_PEER_PORT)) {
                itPeer.port = peerEl.attribute(TAG_PEER_PORT).toUInt(&ok);
                if(!ok && itPeer.port==0) {
                    qWarning() << QString("Bad attribute  (%1) in peer section from '%2'")
                                  .arg(peerEl.attribute(TAG_PEER_PORT)).arg(progDesc.getName());
                    continue;
                }
            } else {
                qWarning() << QString("There is no attribute (%1) in peer section from '%2'")
                              .arg(TAG_PEER_PORT).arg(progDesc.getName());
                continue;
            }
            interface.peers << itPeer;
        }

        if(interface.port==0 && interface.enabled){
            qWarning() << QString("Empty port at enabled interface from '%1'").arg(progDesc.getName());
            return false;
        }

        progDesc.interfaces.push_back(interface);
    }
    return true;
}

ProgramDescription::ProgramDescription() :
    uuid(QUuid::createUuid()),
    hostName(QHostInfo::localHostName())
{
    discoverTimer.start();
}

bool ProgramDescription::isOutOfDate(const QElapsedTimer &other) const
{
    return !discoverTimer.isValid() || discoverTimer.msecsTo(other) > 1000;
}

ProgramInterface& ProgramDescription::getProgramInterface(ProgramInterfaceType type, bool *ok, int id)
{
    return getProgramInterface(ProgramInterface::getTypeName(type), ok, id);
}


const ProgramInterface& ProgramDescription::getProgramInterface(ProgramInterfaceType type, bool *ok) const
{
    return getProgramInterface(ProgramInterface::getTypeName(type), ok);
}


ProgramInterface& ProgramDescription::getProgramInterface(QString type, bool *ok, int id)
{
    for(int i=0; i<interfaces.size(); ++i) {
        if(id>=0 && interfaces[i].id!=id)
            continue;
        if(interfaces[i].type == type) {
            if(ok)
                *ok = true;
            return interfaces[i];
        }
    }
    if(ok)
        *ok = false;
    return dummyInterface;
}

const ProgramInterface &ProgramDescription::getProgramInterface(QString type, bool *ok) const
{
    foreach (const ProgramInterface &i, interfaces) {
        if(i.type == type){
            if(ok)
                *ok = true;
            return i;
        }
    }
    if(ok)
        *ok = false;
    return dummyInterface;
}

int ProgramDescription::getProgramInterfaceNum(ProgramInterfaceType type)
{
    return getProgramInterfaceNum(ProgramInterface::getTypeName(type));
}

int ProgramDescription::getProgramInterfaceNum(QString type)
{
    int n=0;
    for (auto i : interfaces) {
        if(type.isEmpty() || i.type == type){
            ++n;
        }
    }
    return n;
}

bool ProgramDescription::containsProgramInterface(ProgramInterfaceType type, int id)
{
    return containsProgramInterface(ProgramInterface::getTypeName(type), id);
}

bool ProgramDescription::containsProgramInterface(QString type, int id)
{
    for (auto i : interfaces) {
        if(i.type == type && i.id ==id){
            return true;
        }
    }
    return false;
}

bool ProgramDescription::updateProgramInterface(ProgramInterfaceType type, int id, int port, bool isListening, bool isFree, const QHostAddress &peerHost, int peerPort)
{
    bool hasOld;
    ProgramInterface oldInterface = getProgramInterface(type, &hasOld, id);

    if(!hasOld) {
        ProgramInterface new_interface;
        new_interface.setType(type);
        new_interface.id = id;
        interfaces.push_back(new_interface);
    }
    ProgramInterface &interface = getProgramInterface(type, nullptr, id);
    interface.port = port;
    interface.enabled = isListening;
    interface.isFree = isFree;

    if(isFree) {
        interface.peers.clear();
    } else {
        interface.setPeer(peerHost, peerPort);
    }

    return !hasOld || oldInterface!= interface;
}

QString ProgramDescription::getAnnounceStr() const
{
    QDomDocument doc(DOC_TYPE);
    QDomElement root = doc.createElement(TAG_PROGRAM_ANNOUNCE);
    root.setAttribute(TAG_PROGRAM_NAME, getName());
    root.setAttribute(TAG_PROGRAM_TYPE, getDaqModule().toString());
    root.setAttribute(TAG_PROGRAM_INDEX, getProgramIndex().toString());
    if(!getParent().getProgramType().isEmpty())
        root.setAttribute(TAG_PROGRAM_PARENT, getParent().toString());
    root.setAttribute(TAG_PROGRAM_HOST_NAME, hostName);
    if(!host.isNull())
        root.setAttribute(TAG_PROGRAM_HOST, host.toString());
    root.setAttribute(TAG_PROGRAM_UUID, uuid.toString());
    root.setAttribute(TAG_PROGRAM_SEQ, QString::number(seq));
    root.setAttribute(TAG_VER_HASH, ver_hash);
    root.setAttribute(TAG_VER_DATE, ver_date);
    doc.appendChild(root);
    if(!options.isEmpty()){
        QDomElement optionsSection = doc.createElement(TAG_OPTIONS_SECTION);
        root.appendChild(optionsSection);
        for (auto op : options.keys()) {
            QDomElement optionSection = doc.createElement(TAG_OPTION_SECTION);
            optionsSection.appendChild(optionSection);
            optionSection.setAttribute(TAG_OPTION_NAME, op);
            optionSection.setAttribute(TAG_OPTION_VALUE, options[op]);
        }
    }

    if(!interfaces.isEmpty()){
        QDomElement interfacesSection = doc.createElement(TAG_INTERFACES_SECTION);
        root.appendChild(interfacesSection);
        for (auto i : interfaces) {
            QDomElement interfaceSection = doc.createElement(TAG_INTERFACE_SECTION);
            interfacesSection.appendChild(interfaceSection);
//            interfaceSection.setAttribute(TAG_INTERFACE_HOST, i.host.toString());
            interfaceSection.setAttribute(TAG_INTERFACE_PORT, i.port);
            interfaceSection.setAttribute(TAG_INTERFACE_ENABLED, i.enabled?1:0);
            interfaceSection.setAttribute(TAG_INTERFACE_ISFREE, i.isFree?1:0);
            interfaceSection.setAttribute(TAG_INTERFACE_ID, i.id);
            for (auto itPeer : i.peers) {
                QDomElement peerSection = doc.createElement(TAG_PEER_SECTION);
                peerSection.setAttribute(TAG_PEER_HOST, itPeer.host.toString());
                peerSection.setAttribute(TAG_PEER_PORT, itPeer.port);
                interfaceSection.appendChild(peerSection);
            }

            if(!i.type.isEmpty()){
                interfaceSection.setAttribute(TAG_INTERFACE_TYPE, i.type);
            }
        }
    }

    return doc.toString();
}

QString ProgramDescription::getClosingStr() const
{
    QDomDocument doc(DOC_TYPE);
    QDomElement root = doc.createElement(TAG_PROGRAM_CLOSE);
    root.setAttribute(TAG_PROGRAM_NAME, getName());
    root.setAttribute(TAG_PROGRAM_TYPE, getDaqModule().toString());
    root.setAttribute(TAG_PROGRAM_INDEX, getProgramIndex().toString());
    if(!getParent().getProgramType().isEmpty())
        root.setAttribute(TAG_PROGRAM_PARENT, getParent().toString());
    root.setAttribute(TAG_PROGRAM_HOST_NAME, hostName);
    root.setAttribute(TAG_PROGRAM_UUID, uuid.toString());
    root.setAttribute(TAG_PROGRAM_SEQ, QString::number(seq));
    root.setAttribute(TAG_VER_HASH, ver_hash);
    root.setAttribute(TAG_VER_DATE, ver_date);
    doc.appendChild(root);
//    qDebug() << doc.toString();
    return doc.toString();
}

QString ProgramDescription::toString() const
{
    QStringList s;
    s << QString("Uuid:\t%1").arg(uuid.toString());
    s << QString("Seq:\t%1").arg(QString::number(seq));
    s << QString("Type:\t%1").arg(getDaqModule().toString());
    s << QString("Index:\t%1").arg(getProgramIndex().toString());
    if(!getParent().getProgramType().isEmpty())
        s << QString("Parent:\t%1").arg(getParent().toString());
    s << QString("host:\t%1").arg(host.toString());
    s << QString("hostName:\t%1").arg(hostName);
    s << QString("version hash:\t%1").arg(ver_hash);
    s << QString("version date:\t%1").arg(ver_date);
    if(interfaces.size()) {
        int count=1;
        for (auto i : interfaces) {
            s << QString("Interface %1:%2").arg(count++)
                 .arg(i.enabled ? (i.isFree?" (free)":" (busy)") : " (disabled)");
            s << QString("\ttype:\t%1 #%2").arg(i.type).arg(i.id);
            s << QString("\tport:\t%1").arg(i.port);
        }
    }
    if(options.size()) {
        s << QString("Options:");
        for (QString opKey : options.keys())
            s << QString("\t%1: %2").arg(opKey).arg(options[opKey]);
    }

    return s.join("\n");
}

void ProgramDescription::clear()
{
    options.clear();
    interfaces.clear();
}

quint32 ProgramDescription::getHostIp() const
{
    return host.toIPv4Address();
}

QString ProgramDescription::getHostName() const
{
    if(!hostName.isEmpty())
        return hostName;
    else
        return host.toString();
}

QString ProgramInterface::getTypeName(ProgramInterfaceType type)
{
    switch (type) {
    case ProgramInterfaceEmty:
        return "empty";
    case ProgramInterfaceDataFlow:
        return "data flow";
    case ProgramInterfaceRemoteControl:
        return "RemoteControl";
    case ProgramInterfaceMonitorData:
        return "Monitor output data flow";
    case ProgramInterfaceConfigTransfer:
        return "Configuration transfer";
    default:
        return QString();
    }
}

void ProgramInterface::setType(ProgramInterfaceType t)
{
    type = getTypeName(t);
}

void ProgramInterface::setPeer(const QHostAddress &h, quint32 p)
{
    setPeer({h,p});
}

void ProgramInterface::setPeer(const ProgramInterfacePeer &p)
{
    peers.clear();
    addPeer(p);
}

void ProgramInterface::addPeer(const QHostAddress &h, quint32 p) {
    addPeer({h,p});
}

void ProgramInterface::addPeer(const ProgramInterfacePeer &p)
{
    peers << p;
}
