#include "RemoteControlServer.h"

#include <QCoreApplication>
#include <QLoggingCategory>
#include <QJsonDocument>
#include <QJsonObject>

#include "base/RunDocument.h"
#include "base/RunDocumentConverter.h"

#include "RemoteControlProtocol.h"
#include "RemoteControlTlvServer.h"
#include "RunControlCommands.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_rcs)
Q_LOGGING_CATEGORY(lc_rcs, "afi.rcs")
} // namespace

RemoteControlServer::RemoteControlServer(QObject *parent, quint16 preferedTcpPort, QString _descr) :
    QObject(parent), server(new RemoteControlTlvServer(preferedTcpPort, this))
{
    AbstractConfigConverter<RunDocument>::registerTypes();
    AbstractConfigConverter<RcProgramState>::registerTypes();
    setObjectName("RCServer");
    connect(server,
            &RemoteControlTlvServer::connectionAccepted,
            this,
            &RemoteControlServer::connectionAccepted);

    connect(server, &RemoteControlTlvServer::tlvDataReceived, this, &RemoteControlServer::tlvDataReceived);

    connect(server,
            &RemoteControlTlvServer::connectionStatusUpdated,
            this,
            &RemoteControlServer::connectionStatusUpdated);
}

RemoteControlServer::~RemoteControlServer() {}

void RemoteControlServer::setAllowRemoteControl(bool allow)
{
    server->startListen(allow);
}

void RemoteControlServer::startListen(bool state)
{
    setAllowRemoteControl(state);
}

void RemoteControlServer::connectionAccepted()
{
    sendFsmState();
    sendRcProgramState();
}

void RemoteControlServer::tlvDataReceived(const TypeLengthHeader &hdr, const QByteArray &data)
{
    switch(hdr.type) {
    case RC_MSG_CMD_V4 :
        command_line_parse(data);
        break;
    case RC_MSG_CONFIG_V4 :
    {
        QJsonParseError error;
        auto jsonDoc = QJsonDocument::fromJson(data, &error);
        if(error.error == QJsonParseError::NoError) {
            auto rd = QVariant::fromValue(jsonDoc.object()[RC_JSON_RUN_DOC].toObject()).value<RunDocument>();
            emit gotRunDocument(rd);
        } else {
            qWarning(lc_rcs).noquote()
                    << QString("Error while convert incoming data to JSON")
                       .arg(RC_MSG_CONFIG_V4, 8, 16, QChar('0'));
        }
        break;
    }
    case RC_MSG_JSON_V5:
    {
        QJsonParseError error;
        auto jsonDoc = QJsonDocument::fromJson(data, &error);
        if(error.error == QJsonParseError::NoError) {
            json_parse(jsonDoc.object());
        } else {
            qWarning(lc_rcs).noquote()
                    << QString("Error while convert incoming data to JSON")
                       .arg(RC_MSG_JSON_V5, 8, 16, QChar('0'));
        }
        break;
    }
    default :
    {
        static bool warn = true;
        if (warn) {
            qWarning(lc_rcs).noquote()
                    << QString("Skip all packets except those with type %1, %2")
                       .arg(RC_MSG_CMD_V4, 8, 16, QChar('0'))
                       .arg(RC_MSG_CONFIG_V4, 8, 16, QChar('0'));
            warn = false;
        }
    }
    }
}

void RemoteControlServer::config_transfer(const RunDocument &rd) const
{
    auto rdValObj = QVariant::fromValue(rd).value<QJsonObject>();
    QJsonObject rdObj;
    rdObj.insert(RC_JSON_VAL, rdValObj);
    send_json({{RC_JSON_RUN_DOC, rdObj}});
}

void RemoteControlServer::send_json(const QJsonObject &obj) const
{
    if (!server->isListening())
        return;
    server->send_json(obj);
}

bool RemoteControlServer::parseCommand(const QString &command, const QStringList &args)
{
    if (command == RC_CMD_START || command == RC_CMD_STOP) {
        // start <globIndexStr> <runNumber> <uuid> [startEvNumber]
        // stop <globIndexStr> <runNumber> <uuid> [stopEvNumber]

        bool ok = args.size() >= 3;
        RcRunId runId;
        if (ok)
            runId.runNumber = args.at(1).toInt(&ok);
        if (ok) {
            runId.uuid = QUuid::fromString(args.at(2));
            ok = !runId.uuid.isNull();
        }
        QVariant evNumber;
        if (ok && args.size() > 3)
            evNumber = args.at(3).toULongLong(&ok);
        if (ok) {
            runId.runIndex = args.at(0);
            if (command == RC_CMD_START)
                emit startRequest(runId, evNumber);
            else
                emit stopRequest(runId, evNumber);
        } else {
            qWarning(lc_rcs).noquote()
                << QString("command '%1' arguments error: expected <globIndexStr> <runNumber> <uuid> [evNumber]")
                       .arg(command);
            return false;
        }
    } else if (command == RC_CMD_EV_NUM_CHECK) {
        // EvNumCheck
        bool ok = args.size() == 1;
        int checkIndex = 0;
        if (ok)
            checkIndex = args.at(0).toInt(&ok);
        if (ok) {
            emit evNumCheckRequest(checkIndex);
        } else {
            qWarning(lc_rcs).noquote()
                << QString("command '%1' arguments error: expected <checkIndex>")
                       .arg(command);
            return false;
        }
    } else {
        qWarning().noquote() << QString("unknown command '%1'").arg(command);
        return false;
    }
    return true;
}

void RemoteControlServer::command_line_parse(QByteArray data)
{
    QString line = QString::fromUtf8(data.data(), data.size());
    for (QString message : line.split(RC_CMD_END)) {
        message = message.trimmed();
#if (QT_VERSION >= QT_VERSION_CHECK(5, 14, 0))
        QStringList args = message.split(" ", Qt::SkipEmptyParts);
#else
        QStringList commandParts = command.split(" ", QString::SkipEmptyParts);
#endif
        if (args.isEmpty()) {
            qDebug(lc_rcs).noquote() << "Empty command";
            continue;
        }
        const auto &command = args.takeFirst();
        if (!parseCommand(command, args))
            qWarning(lc_rcs).noquote() << QString("Parse error: command %1, args %2").arg(command).arg(args.join(" "));
        else
            qDebug(lc_rcs).noquote() << QString("Command %1, args %2").arg(command).arg(args.join(" "));
    }
}

void RemoteControlServer::json_parse(const QJsonObject &obj)
{
    for(auto it=obj.constBegin(); it!=obj.constEnd(); ++it) {
        const auto &key = it.key();
        const auto &data = it->toObject();

        if (key == RC_CMD_START || key == RC_CMD_STOP) {
            bool ok = true;
            RcRunId runId;
            auto it = data.constFind(RC_JSON_RUN_ID);

            if (it != data.constEnd()) {
                runId = QVariant::fromValue(*it).value<RcRunId>();
            } else
                ok = false;

            QVariant evNumber;
            it = data.constFind(RC_JSON_RUN_EV_NUM);
            if (it != data.constEnd()) {
                evNumber = jsonValueToUInt64(*it);
            }
            if (ok) {
                if (key == RC_CMD_START)
                    emit startRequest(runId, evNumber);
                else
                    emit stopRequest(runId, evNumber);
            } else {
                qWarning(lc_rcs).noquote()
                    << QString("command '%1' arguments error: expected <globIndexStr> <runNumber> <uuid> [evNumber]")
                           .arg(key);
            }
            continue;
        }

        if (key == RC_CMD_EV_NUM_CHECK) {
            // EvNumCheck
            bool ok;
            auto it = data.constFind(RC_JSON_EV_NUM_CHECK_ID);
            int checkIndex;
            if (it != data.constEnd()) {
                checkIndex = QVariant::fromValue(*it).toInt(&ok);
            } else
                ok = false;
            if (ok) {
                emit evNumCheckRequest(checkIndex);
            } else {
                qWarning(lc_rcs).noquote()
                    << QString("command '%1' arguments error: expected <checkIndex>")
                           .arg(key);
            }
            continue;
        }

        if (key == RC_JSON_RUN_DOC) {
            auto rd = QVariant::fromValue(data[RC_JSON_VAL].toObject()).value<RunDocument>();
            emit gotRunDocument(rd);
            continue;
        }
        if(key == RC_CMD_TREE_STATE) {
            auto allRcProgramState = QVariant::fromValue(data).value<RcProgramState>();
            emit gotRcProgramState(allRcProgramState);
            continue;
        }

        if(unknownCommands.contains(key)) {
            qWarning() << objectName()
                       << QString("Got unknown command '%1':").arg(key) << data;
            unknownCommands << key;
        }
    }
}

void RemoteControlServer::sendState(const QString &state)
{
    send_json({{RC_CMD_STATE, state}});
}

void RemoteControlServer::onFsmStateChange(const FsmState &state)
{
    fsmState = state;
    sendFsmState();
}

void RemoteControlServer::onRcProgramStateChange(const RcProgramState &state)
{
    if(rcProgramState == state)
        return;

    rcProgramState = state;
    sendRcProgramState();
}

void RemoteControlServer::onRunError(const RcRunId &runId, QString runError)
{
    if(suppresedErrorRunId == runId)
        return;
    suppresedErrorRunId = runId;

    QJsonObject obj;
    obj.insert(RC_JSON_RUN_ID, QVariant::fromValue(runId).toJsonValue());
    obj.insert(RC_JSON_VAL, runError);
    send_json({{RC_CMD_ERROR, obj}});
}

void RemoteControlServer::onRunWarning(const RcRunId &runId, const QString &warnStr)
{
    QJsonObject obj;
    obj.insert(RC_JSON_RUN_ID, QVariant::fromValue(runId).toJsonValue());
    obj.insert(RC_JSON_VAL, warnStr);
    send_json({{RC_CMD_WARN, obj}});
}

void RemoteControlServer::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if (runId.runIndex.isEmpty())
        return;

    QJsonObject obj;
    obj.insert(RC_JSON_RUN_ID, QVariant::fromValue(runId).toJsonValue());
    obj.insert(RC_JSON_VAL, backPressure);
    send_json({{RC_CMD_BACK_PRESSURE, obj}});
}

void RemoteControlServer::onVmeFlag(bool hasVme)
{
    send_json({{RC_CMD_VME_FLAG, hasVme}});
}

void RemoteControlServer::sendFsmState() const
{
    send_json({{RC_CMD_FSM_STATE, fsmStateToString(fsmState)}});
}

void RemoteControlServer::sendRcProgramState() const
{
    send_json({{RC_CMD_TREE_STATE, QVariant::fromValue(rcProgramState).toJsonValue()}});
}

void RemoteControlServer::evNumCheckResponse(int checkIndex, const EvNumSet &evNums) const
{
    QJsonArray evNumArr;
    for (auto  evNum : evNums) {
        evNumArr.append(jsonValueFromUInt64(evNum));
    }

    QJsonObject obj;
    obj.insert(RC_JSON_EV_NUM_CHECK_ID, checkIndex);
    obj.insert(RC_JSON_VAL, evNumArr);
    send_json({{RC_CMD_EV_NUM_CHECK, obj}});
}
